#ifndef EMTIMER_H
#define EMTIMER_H
#include <linux/kernel.h>
#include <asm/ioctl.h>

#ifdef __KERNEL__
#  include <linux/time.h>
#else
#  include <time.h>
#  define __user
#endif

typedef int emtimer_t;

struct emtimer_config
{
   __u32 submit_queue;
   __u32 free_queue;
   __u32 region_dma_addr;
   __u32 region_size;
   __u32 pool_dma_addr;
   __u32 pool_size;
};

struct emtimer_create
{
   struct itimerspec its; /* optional, set to 0 to not start */
   const void __user *event;
   __u32 queue; /* for AXM, the upper 8 bits are DID */
   emtimer_t timerid; /* OUT */
};

struct emtimer_settime
{
   struct itimerspec its; /* set to 0 to cancel */
   emtimer_t timerid;
};

struct emtimer_destroy
{
   emtimer_t timerid;
   __u32 tx_count;
   __u32 overrun_count;
};

#define EMTIMER_IOC_MAGIC  't'
#define EMTIMER_CONFIG     _IOW (EMTIMER_IOC_MAGIC, 0, struct emtimer_config)
#define EMTIMER_CREATE     _IOWR(EMTIMER_IOC_MAGIC, 1, struct emtimer_create)
#define EMTIMER_SETTIME    _IOW (EMTIMER_IOC_MAGIC, 2, struct emtimer_settime)
#define EMTIMER_DESTROY    _IOWR(EMTIMER_IOC_MAGIC, 3, struct emtimer_destroy)

#ifdef __KERNEL__
#  include "emtimer_internal.h"
#endif

#endif /* EMTIMER_H */
