// *****************************************************************************
// PROJECT NAME     :   FZM
// Component name   :   mbtac_packet_parsing.c
// Filename         :   
// Author           :   Murali,Suresh
// Date created     :   09.12.2013
// Date changed     :   -
// -----------------------------------------------------------------------------
// Description      :   MBTac module
//
// -----------------------------------------------------------------------------
// Revision history :
// date       | author         | comments
// ---------- + -------------- +----------------------------------------------------
// 09.12.2013 | Suresh Y       | Initial version
//            |                |
// *****************************************************************************
// Copyright (c) Nokia 2013
// All rights reserved.
// Licensed material - property of Nokia
// *****************************************************************************
#ifndef __MBTAC_PACKET_PARSING_SRC_C__
#define __MBTAC_PACKET_PARSING_SRC_C__

#include <linux/skbuff.h>
#include <net/xfrm.h>
#include "ip_secfg_IWF_VTC.h"
#include <net/if_inet6.h>
#include <net/ip6_fib.h>
#include <linux/ipv6.h>
#include <net/addrconf.h>

#define MIN_ESP_PKT_SIZE  80 // IP(20) + ESP(51) + 8
extern uint8_t  mbtac_l2_incl_flag;
extern uint64_t         mbtac_dscp_cfg;
extern mbTacFpTmbData_t mbtacTmbCounterData;
extern bool iwf_ipsec_flag;
/*MBTAC COUNTERS */

extern uint32_t ul_uplane_pkt_cntr;
extern uint32_t ul_cplane_pkt_cntr;
extern uint32_t ul_splane_pkt_cntr;
extern uint32_t ul_mcast_cplane_pkt_cntr;

extern uint32_t dl_uplane_pkt_cntr;
extern uint32_t dl_cplane_pkt_cntr;
extern uint32_t dl_splane_pkt_cntr;
extern uint32_t dl_mcast_cplane_pkt_cntr;

#define S_PLANE_PORT1 319
#define S_PLANE_PORT2 320
#define MIN_IP_PKT_SIZE 100
#define DOWNLINK  1
#define UPLINK    0

#define ETH_HDR_LEN   14
#define VLAN_HDR_LEN  16
/*****************************************************************************
 * Method Name        : dscpIsMbtacRelevant
 *
 * Input Parameters   : dscp
 *
 * Returns            : Status - 0 or 1
 *
 * Description        : To check the DSCP on pkt is relevent to DSCP of MBTAC
 *
 ****************************************************************************/
static inline uint8_t dscpIsMbtacRelevant(uint8_t dscp)
{
    if( ((((uint64_t)0x01) << dscp) & mbtac_dscp_cfg) != ((uint64_t)0x00))
    {
        return 1;
    }

    return 0;
}
/*****************************************************************************
 * Method Name        : isIpAddrMatch
 *
 * Input Parameters   : skb,version
 *
 * Returns            : Status - 0 or 1
 *
 * Description        : To check if the ip in skb and from configuration are same
 *
 ****************************************************************************/
static inline uint8_t isIpAddrMatch(struct sk_buff *skb,mbTacFpTmbDataPerIpAddr_t *data,bool dir)
{
    struct iphdr *ip_header = (struct iphdr*)skb->network_header;
    if(ip_header->version == IWF_IPV4_VERSION)
    {
        if(dir == DOWNLINK && 
           ntohl(ip_header->daddr) == data->ip.ip4_addr32)
        {
	    return 1;
        }
        else if(dir == UPLINK && 
                ntohl(ip_header->saddr) == data->ip.ip4_addr32)
        {
            return 1;
        }
	else
        {
	    return 0;
        }
    }
    else if(ip_header->version == IWF_IPV6_VERSION)
    {
        struct ipv6hdr *ipv6_header = (struct ipv6hdr*)skb->data;
        if(dir == DOWNLINK && 
           !memcmp(&ipv6_header->daddr.s6_addr32,&data->ip.addr_u.ipv6Address.ipa_32,sizeof(struct in6_addr)))
         {
	   return 1;      
         } 
        if(dir == UPLINK && 
           !memcmp(&ipv6_header->saddr.s6_addr32,&data->ip.addr_u.ipv6Address.ipa_32,sizeof(struct in6_addr)))
	{
	   return 1;       
	}
        else
        {
           return 0;
        }
    }
    return 0;
}
/*****************************************************************************
 * Method Name        : mbTacCountSplaneBytesDownlink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : To count the pkt length of S-Plane traffic on DL
 *
 ****************************************************************************/
void mbTacCountSplaneBytesDownlink (struct sk_buff *skb)
{
    struct iphdr *ip_header = (struct iphdr*)skb->network_header;
    struct udphdr *udphdr = NULL;
    if(ip_header->version == IWF_IPV4_VERSION)
    {
        udphdr = (struct udphdr *)(skb->data + sizeof(struct iphdr));
    }
    else if(ip_header->version == IWF_IPV6_VERSION)
    {
        udphdr = (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));
    }
    if( NULL != udphdr )
    {
        if ((LTM_TOP_PORT_1 == ntohs(udphdr->source)) ||
                        (LTM_TOP_PORT_2 == ntohs(udphdr->source)))
        {
             if(isIpAddrMatch(skb,&mbtacTmbCounterData.splaneData,DOWNLINK))
            {
                uint16_t pktLen = mbTacDlPktLenCalc(skb);
		mbtacTmbCounterData.splaneData.bytesSumDl += ((skb->mac_len + L2_FCS)* mbtac_l2_incl_flag) + pktLen;
                dl_splane_pkt_cntr++;
             }
         }
     }
}
/*****************************************************************************
 * Method Name        : mbTacCountCplaneBytesDownlink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : To count the pkt length of C-Plane traffic on DL
 *
 ****************************************************************************/
void mbTacCountCplaneBytesDownlink (struct sk_buff *skb)
{
    int i = 0;
    uint16_t pktLen;
    struct iphdr *ip_header = (struct iphdr*)skb->network_header;
    if (ip_header->protocol == IPPROTO_IGMP)
    {
           pktLen = mbTacDlPktLenCalc(skb);
           mbtacTmbCounterData.mcastcplaneData.bytesSumDl += ((skb->mac_len + L2_FCS)* mbtac_l2_incl_flag) + pktLen;
           dl_mcast_cplane_pkt_cntr++;
    }
    else
    {
        for (i = 0; i < MAX_NO_CPLANE_IP_ADDR; i++)
        {
            if (isIpAddrMatch(skb,&mbtacTmbCounterData.cplaneData[i],DOWNLINK))
	    {
                pktLen = mbTacDlPktLenCalc(skb);
	        mbtacTmbCounterData.cplaneData[i].bytesSumDl += ((skb->mac_len + L2_FCS)* mbtac_l2_incl_flag) + pktLen;
	        dl_cplane_pkt_cntr++;
	        break;
	    }
        }
    }
}
/*****************************************************************************
 * Method Name        : mbTacCountUplaneBytesDownlink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : To count the pkt length of U-Plane traffic on DL
 *
 ****************************************************************************/
void mbTacCountUplaneBytesDownlink (struct sk_buff *skb)
{
    uint8_t dscp =0 , protocol = 0;
    int i = 0;
    struct iphdr *ip_header = (struct iphdr*)skb_network_header(skb);

    if(ip_header->version == IWF_IPV4_VERSION)
    {
        dscp = ip_header->tos >> 2;
        protocol = ip_header->protocol;
    }
    if(ip_header->version == IWF_IPV6_VERSION)
    {
       struct ipv6hdr *ipv6_header = (struct ipv6hdr*)skb->data;
       dscp = ipv6_header->priority << 2 | ipv6_header->flow_lbl[0] >> 6;
       protocol = ipv6_header->nexthdr;
    }

    if(dscpIsMbtacRelevant(dscp) && (MBTAC_NH_UDP == protocol))
    {
	for(i = 0; i < MAX_NO_UPLANE_IP_ADDR; i++)
         {
             if(isIpAddrMatch(skb,&mbtacTmbCounterData.uplaneData[i],DOWNLINK))
             {
                 uint16_t pktLen = mbTacDlPktLenCalc(skb);
		 mbtacTmbCounterData.uplaneData[i].bytesSumDl += ((skb->mac_len + L2_FCS)* mbtac_l2_incl_flag) + pktLen;
		 dl_uplane_pkt_cntr++;
                 break;
             }
         }
    }
}

/*****************************************************************************
 * Method Name        : mbTacValidateAndUpdateBytes
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : To count the pkt length of S/C/U-Plane traffic on UL
 *
 ****************************************************************************/
int mbTacValidateAndUpdateBytes(struct sk_buff *skb,uint16_t no_of_frags)
{
	uint8_t index = 0xff;
	uint8_t traffic_type = 0xff,protocol = 0;
	struct iphdr *ip_header = (struct iphdr*)skb_network_header(skb);
	struct net_device *dev = NULL;

	index = *((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR]));
    
        traffic_type = *((unsigned char*)(&skb->cb[MBTAC_PKT_TYPE_ADDR]));

	if(ip_header->version == IWF_IPV4_VERSION)
        {
		protocol = ip_header->protocol;
        }
        else        
	{
        	struct ipv6hdr *ipv6_header = (struct ipv6hdr*)skb->data;
		protocol = ipv6_header->nexthdr;	
	}	   	

	if((IPPROTO_UDP == protocol) || 
	   (MBTAC_NH_SCTP == protocol) || 
	   (IPPROTO_ESP == protocol) ||
            (IPPROTO_IGMP == protocol))
	{
        if(index > 0 && index <= MAX_NO_UPLANE_IP_ADDR)
        {
            uint16_t ethHdrLength = ETH_HDR_LEN;
            uint16_t pktLen = 0;
            if(ip_header->version == IWF_IPV4_VERSION)
            {
                dev = skb_dst(skb)->dev;
            }
            else if(ip_header->version == IWF_IPV6_VERSION)
            {
                struct inet6_dev *idev;
                if(skb_dst(skb))
                {
                    idev = ip6_dst_idev(skb_dst(skb));
                }
                else
                {
                    idev = __in6_dev_get(skb->dev);
                }
                if(idev)
                {
                   dev = idev->dev;
                }
            }
            if(dev)
            {
                char *ptr = dev->name;
                while(*ptr != '.' && *ptr != '\0')
                    ptr++;
                if(*ptr == '.')
                    ethHdrLength = VLAN_HDR_LEN ;
            }


            pktLen = mbTacUlPktLenCalc(skb,no_of_frags,ethHdrLength);
            if(pktLen)
            {
                switch(traffic_type)
                {
				case MBTAC_U_PLANE :
					mbtacTmbCounterData.uplaneData[index-1].bytesSumUl +=  pktLen;
					ul_uplane_pkt_cntr++;
					break;
				case MBTAC_C_PLANE:
					mbtacTmbCounterData.cplaneData[index-1].bytesSumUl += pktLen;
					ul_cplane_pkt_cntr++;
					break;
				case MBTAC_MBMS_C_PLANE:
					mbtacTmbCounterData.mcastcplaneData.bytesSumUl += pktLen;
					ul_mcast_cplane_pkt_cntr++;
					break;
				case MBTAC_S_PLANE:
					mbtacTmbCounterData.splaneData.bytesSumUl += pktLen;
					ul_splane_pkt_cntr++;
					break;
				default:
					break;
	    		}
                    }

		}
	}
    return 0;
}

/*****************************************************************************
 * Method Name        : mbTacMarkingUplaneUplink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : Marking of MBTC on U-Plane traffic in UL direction
 *
 ****************************************************************************/
 void mbTacMarkingUplaneUplink (struct sk_buff *skb)
{
    unsigned char i = 0;
    uint8_t dscp = 0,protocol = 0;
    struct iphdr *ip_header = (struct iphdr*)skb->network_header;
    
    if(ip_header->version == IWF_IPV4_VERSION)
    {
        dscp = ip_header->tos >> 2;
        protocol = ip_header->protocol;
    }
    else
    {
       struct ipv6hdr *ipv6_header = (struct ipv6hdr*)skb->data;
       dscp = ipv6_header->priority << 2 | ipv6_header->flow_lbl[0] >> 6;
       protocol = ipv6_header->nexthdr;
    }

    if(dscpIsMbtacRelevant(dscp) && MBTAC_NH_UDP == protocol)
    {
	for(i = 0; i < MAX_NO_UPLANE_IP_ADDR; i++)
	{
		if(isIpAddrMatch(skb,&mbtacTmbCounterData.uplaneData[i],UPLINK))
		{
			*((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR])) = i + 1;
	                *((unsigned char*)(&skb->cb[MBTAC_PKT_TYPE_ADDR])) = MBTAC_U_PLANE;
			break;
		}
    	}
    }
}

/*****************************************************************************
 * Method Name        : mbTacMarkingCplaneUplink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : Marking of MBTC on C-Plane traffic in UL direction
 *
 ****************************************************************************/

 void mbTacMarkingCplaneUplink (struct sk_buff *skb)
{
    unsigned char i = 0;

    struct iphdr *ip_header = (struct iphdr*)skb->network_header;
    if (ip_header->protocol == IPPROTO_IGMP)
    {
          *((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR])) = 1;
          *((unsigned char*)(&skb->cb[MBTAC_PKT_TYPE_ADDR])) = MBTAC_MBMS_C_PLANE;
    }
    else
    {
        for(i = 0; i < MAX_NO_CPLANE_IP_ADDR; i++)
        {
            if(isIpAddrMatch(skb,&mbtacTmbCounterData.cplaneData[i],UPLINK))
            {
               *((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR])) = i + 1;
               *((unsigned char*)(&skb->cb[MBTAC_PKT_TYPE_ADDR])) = MBTAC_C_PLANE;
               break;
            }
        }
    }
}

/*****************************************************************************
 * Method Name        : mbTacMarkingSplaneUplink
 *
 * Input Parameters   : struct sk_buff
 *
 * Returns            : 
 *
 * Description        : Marking of MBTC on S-Plane traffic in UL direction
 *
 ****************************************************************************/
 void mbTacMarkingSplaneUplink (struct sk_buff *skb)
{
    if(isIpAddrMatch(skb,&mbtacTmbCounterData.splaneData,UPLINK))
    {
        *((unsigned char*)(&skb->cb[MBTAC_PKT_TYPE_ADDR])) = MBTAC_S_PLANE;
        *((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR])) = 1;
    }
}



unsigned short mbTacDlPktLenCalc(struct sk_buff *skb)
{

    unsigned short pktlen = 0;
    uint16_t noOfFraments  = skb->cb[MBTAC_METADATA_FRAG_CNT];

    pktlen = (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8
              | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);

    // If the packet is not a ipsec , 
    //then the cb[] will be filled with zero
    if(pktlen < MIN_ESP_PKT_SIZE)
    {
        pktlen = skb->len;
    }//Fill the Fragmented packet L2 Header length
    pktlen += (noOfFraments * (skb->mac_len+L2_FCS))*(mbtac_l2_incl_flag);

    return pktlen;
}

unsigned short mbTacUlPktLenCalc(struct sk_buff *skb, uint16_t fragCnt, uint16_t ethHdrLength)
{
    unsigned short pktlen = 0;
    struct dst_entry *dst = skb_dst(skb);
    struct xfrm_state *x = dst->xfrm;

    if(iwf_ipsec_flag == true)
    {
        if(skb->sp != NULL)
	{
            pktlen = skb->len;
	}
	else if(x == NULL && skb->sp == NULL)
	{
            pktlen = skb->len;
	}
    } 
    else
    {
        pktlen = skb->len;
    }

    // To count the L2 header length
    pktlen =  (mbtac_l2_incl_flag ? (((ethHdrLength + L2_FCS ) * fragCnt )+
               pktlen ) : pktlen);
    return pktlen;
}

#endif
