// *****************************************************************************
// PROJECT NAME     :   FZM
// Component name   :   iwfD.c
// Filename         :
// Author           :   Abhilash.B
// Date created     :   19.02.2013
// Date changed     :   23.01.2015
// -----------------------------------------------------------------------------
// Description      :   IWF Deamon
//
// -----------------------------------------------------------------------------
// Revision history :
// date       | author         | comments
// ---------- + -------------- +----------------------------------------------------
// 19.02.2013 | Abhilash.B     | Initial version
// 20.03.2013 | Shwethambari C | Added extension header part
// 21.03.2013 | Maniraj Patri  | Added Error indication functionality
// 04.09.2013 | Abhilash.B     | Used netfilter for processing rx packet in DSP simulator
// 23.01.2015 | Drzewiecki.T   | Lower hooks priority in order to achieve traffic capturing
//            |                |
// *****************************************************************************
// Copyright (c) Nokia 2015
// All rights reserved.
// Licensed material - property of Nokia
// *****************************************************************************

#include "iwf_table_creation.h"
#include "iwfD.h"

#include <linux/skbuff.h>

#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>
#include <linux/skbuff.h>
#include <linux/udp.h>
#include <net/ip.h>
#include <net/route.h>
#include <net/ip6_route.h>
#include <net/xfrm.h>
#include <net/ip6_checksum.h>
#include <net/udp.h>
#include <net/addrconf.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <net/icmp.h>
#include <linux/netfilter_bridge.h>
#include <linux/if_vlan.h>

extern bool iwf_sim_flag;
extern bool iwf_ipsec_flag;
extern sicp_lte_ip_data_t  *sicp_lte_ip;//Holds Plane IP's
extern iwf_ip_addr_t mplaneIp;
extern mbTacFpTmbData_t mbtacTmbCounterData;

#define __IWF_VERSION_MAJOR__    2
#define __IWF_VERSION_MINOR__    0
#define __IWFD_SET_SKB_PRIORITY
#define __IWF_ONLY_ONE_CALL_BK

#define  __IWF_ENABLE_MSG_RTR
#define  __IWF_ENABLE_MSG_RTR_LB

#define  __IWF_USE_IP_FOR_UL
#define IPPROTO_HOPTOHOP 0

#ifdef  __IWF_ENABLE_MSG_RTR
#include "../include/msg_rtr_arm.h"
#endif

#define __USE_NET_FILTER__

#define _IFACE_BIT 2
#define _IFACE_VLAN_BIT 3
#define _IFACE_VLAN_ID_BIT 4

#define __IWF_UL__
#define __IWF_DL__

#define CRASH_HNDL
#define __IWFD_SKB_PRIORITY__        1
#define __IWFD_DSP_SIM_SOURCE_PORT__ 63
#define __IWF_TUPU_SICAD             0x1211
#define __IWF_FTM_SICAD              0x123E

#define __IWF_REGISTRD_SICAD         __IWF_FTM_SICAD

#define __IWF_UL_MIN_CCID            256
#define __IWF_UL_MAX_CCID   65535

#define __IWF_DL_MIN_CCID   1
#define __IWF_DL_MAX_CCID   255

#define __IWF_BIND_RETRY_MAX  3

#define PROCFS_MAX_SIZE     1024

#define ECN       0x3
#define IPv6_ECN  0x3f
#define IP_FRAG_OFF 0xffbf
#define ETH_HDR_LEN   14
#define VLAN_HDR_LEN  16

#define ADDR_LEN 4
#define IP_OPTION_LEN 5
#define IPPROTO_OSPF 89
#define NEXTHDR_IGMP 2
#define NEXTHDR_OSPF 89
#define NEXTHDR_SCTP 132
#define HOP_BY_HOP_HDR_LENGTH 8
#define ND_HOP_LIMIT 254

#define IF_NAME "br0"
#define IF_VLAN "br0."

#define NO_OF_INTF_THREE_PORT_HW 3
#define MAX_VLAN_ID 4096
#define skb_vlan_tag_present(__skb)     ((__skb)->vlan_tci & VLAN_TAG_PRESENT)
#define skb_vlan_tag_get_id(__skb)      ((__skb)->vlan_tci & VLAN_VID_MASK)
#define ETH_P_SOAM      0x8902

// k2l work around.  Check if this is requuired at system integration
struct net_device *dev_eth0;

//==> k2l debugging
#define CPP_DEBUG_TRACE_SIZE 128
#define CPP_DEBUG_BYTES_PER_LINE 40

//#define DBG_DUMP_ON
#ifdef DBG_DUMP_ON
static void dump_msg(uint8_t *msgPtr, const uint32_t msgLen, int dir) {
    char str[CPP_DEBUG_TRACE_SIZE];
    char* strPtr;
    uint32_t n;
    char tmpstr[4] ="";
    bool display = false;
    str[0] = '\0';
    strPtr = str;

    for (n = 0; n < msgLen; n++)
    {
        if (msgPtr[n] <= 0xf)
        {
            sprintf(tmpstr, " 0%x", msgPtr[n]);
        }
        else
        {
            sprintf(tmpstr, " %2x", msgPtr[n]);
        }
        strcat(strPtr,tmpstr);
        if ((n % CPP_DEBUG_BYTES_PER_LINE) == 1)
        {
            if (display)
            {
                printk("%s\n",str);
                tmpstr[0] = '\0';
                str[0] = '\0';
                strPtr = str;
            }
            else
            {
                printk("%s Len=%d\n", (dir? "Outgoing" :"Incoming"), msgLen);
                display = true;
            }
        }
    }

    printk("%s\n",str);
}

static int is_dbg_print_on = 0;
static char const *baseName(char const *path)
{
    char *s = strrchr(path, '/');
    if (!s)
        return path;
    else
        return (s + 1);
}
#define DBG_PRT_ON() { is_dbg_print_on = 1; }
#define DBG_PRT_OFF() { is_dbg_print_on = 0; }
#define DBG_PRT(format, ...) do {\
    if(is_dbg_print_on)\
    printk("%s %d: " format, baseName(__FILE__), __LINE__, ##__VA_ARGS__);\
} while(0)
#define DBG_PRT_ALL(format, ...) printk("%s %d: " format, baseName(__FILE__), __LINE__, ##__VA_ARGS__); 
#define DUMP_MSG(msg, len, direction) dump_msg(msg, len, direction)
#else
#define DBG_PRT(format, ...)
#define DBG_PRT_ALL(format, ...)
#define DBG_PRT_ON()
#define DBG_PRT_OFF()
#define DUMP_MSG(msg, len, direction)
#endif
//End of k2l debugging

extern uint8_t vlanIds[MAX_VLANID];
_ipCounters ipCounters[VLANIDS_MAX];

extern bool VlanAwareFlag;
extern uint8_t NumofPorts;
extern svfCfgInfo vlanCfgInfo[MAX_NUM_LOGICAL_INTF];
static int TaggedFramesCount[3], DiscardTaggedFramesCount[3], DiscardUnTaggedFramesCount[3];

static bool validateTTL(struct sk_buff *skb);
static bool isBTSIp(struct sk_buff *skb);
#ifdef __USE_NET_FILTER__
static struct nf_hook_ops netfilter_uplink_ops;
static struct nf_hook_ops netfilter_downlink_ops;
static struct nf_hook_ops netfilter_downlink_pre_ops;
static struct nf_hook_ops netfilter_downlink_ipv6_ops;
static struct nf_hook_ops netfilter_downlink_ipv6_pre_ops;
static struct nf_hook_ops mbtac_uplink_ops;
static struct nf_hook_ops mbtac_uplink_ipv6_ops;
static struct nf_hook_ops mbtac_uplink_host_ops;
static struct nf_hook_ops mbtac_uplink_host_ipv6_ops;
static bool bIsULHookRegistered=false;
#endif

/* extern functions */
extern  int  seCfgDrv_init(void);
extern  void seCfgDrv_exit(void);
extern  uint32_t get_ip_se_lte_ip (uint8_t index);
extern  iwf_ip_addr_t *get_up_ip_se_lte_ip (uint8_t index);

/* extern variables */
extern  bool gUplaneIpUpdated;
extern  struct mutex gUPlaneIpMutexLock;
extern  sicp_ip_data_t         *lte_sicp_ip_table;
extern iwf_ip_addr_t_n  * ipList;
extern  vtc_lte_trsw_config_table_t *Strsw_config_table;
extern  uint8_t                trsw_config_table_place[MAX_SGW_IP_ADDRESSES];

unsigned int iwf_log_level           = 0;
uint16_t ftm_nid =0;
/* globals */
struct  sicp_data_t        *sicp_hdr     = NULL;
struct  socket             *gSockDL      = NULL;
struct  socket             *gSockUL      = NULL;

//static  uint8_t ulSockAddress[16] = "12.12.12.12"; //peer address should be 12.12.12.13

static void sendPktToSgw(struct sk_buff * buffer);

#ifndef __IWF_ONLY_ONE_CALL_BK
static void sendPktToLMP(struct sk_buff * buffer);
#endif

/*IWF COUNTERS*/
uint32_t tx_total_ul_gtpu_pkt_from_tupu               = 0;
uint32_t tx_ul_gtpu_pkt_drop_no_entry                 = 0;
uint32_t tx_ul_invalid_gtpu_pkt_drop_cntr             = 0;
uint32_t tx_ul_gtpu_pkt_drop_null_pkt_rcvd            = 0;
uint32_t rx_gtpu_total_pkt                            = 0;
uint32_t rx_dl_pkt_to_tupu                            = 0;
uint32_t rx_dl_gtpu_pkt_drop_cntr                     = 0;
uint32_t rx_dl_gtpu_xfrm_drop_no_policy_ipv4          = 0;
uint32_t rx_dl_gtpu_xfrm_drop_no_policy_ipv6          = 0;
uint32_t rx_dl_gtpu_pkt_drop_wrong_sicp_hdr           = 0;
uint32_t rx_dl_gtpu_pkt_drop_invalid_gtp_hdr          = 0;

uint32_t tx_ul_gtpu_pkt_from_tupu_ipv4                = 0;
uint32_t tx_ul_gtpu_pkt_to_sgw_cntr_ipv4              = 0;
uint32_t tx_ul_gtpu_pkt_drop_cntr_ipv4                = 0;
uint32_t tx_ul_gtpu_pkt_drop_no_route_ipv4            = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4      = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv4    = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4       = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv4  = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4  = 0;
uint32_t tx_ul_gtpu_pkt_post_routing                  = 0;
uint32_t tx_ul_gtpu_pkt_ip_local_out                  = 0;

uint32_t rx_total_pkt_ipv4                            = 0;
uint32_t rx_total_pkt_ipv4_pre_route                  = 0;
uint32_t rx_total_pkt_ipv4_post_route                 = 0;
uint32_t rx_total_pkt_ipsec_post_route                = 0;
uint32_t rx_total_pkt_ipv6_pre_route                  = 0;
uint32_t rx_total_pkt_ipv6_post_route                 = 0;
uint32_t rx_udp_total_pkt_ipv4                        = 0;
uint32_t rx_dl_gtpu_mgmt_local_to_sgw_ipv4            = 0;
uint32_t rx_dl_gtpu_mgmt_local_req_ipv4               = 0;
uint32_t rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4  = 0;
uint32_t rx_dl_icmp_checksum_error_ipv4               = 0;
uint32_t rx_dl_checksum_error_ipv4                    = 0;

uint32_t tx_ul_gtpu_pkt_from_tupu_ipv6                = 0;
uint32_t tx_ul_gtpu_pkt_to_sgw_cntr_ipv6              = 0;
uint32_t tx_ul_gtpu_pkt_drop_cntr_ipv6                = 0;
uint32_t tx_ul_gtpu_pkt_drop_no_route_ipv6            = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6      = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6  = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv6    = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv6       = 0;
uint32_t tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv6  = 0;
uint32_t rx_dl_icmp_checksum_error_ipv6               = 0;

uint32_t rx_total_pkt_ipv6                            = 0;
uint32_t rx_udp_total_pkt_ipv6                        = 0;
uint32_t rx_dl_gtpu_mgmt_local_to_sgw_ipv6            = 0;
uint32_t rx_dl_gtpu_mgmt_local_req_ipv6               = 0;
uint32_t rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6  = 0;
uint32_t rx_dl_pkt_drop_multicast_src_ipv6            = 0;
uint32_t rx_udp_checksum_err_drop_ipv6                = 0;

struct   proc_dir_entry *fzm_iwf_d  = NULL;
struct   proc_dir_entry *fzm_iwf_f  = NULL;
struct   proc_dir_entry *fzm_iwf_tbl_dump_f = NULL;

/*IWF LOGLEVEL*/
unsigned int iwfLogLevelReceived               = 0;
struct   proc_dir_entry *fzm_iwf_loglevel_help_f = NULL;
struct   proc_dir_entry *fzm_iwf_loglevel_add_f = NULL;
struct   proc_dir_entry *fzm_iwf_loglevel_delete_f = NULL;
struct   proc_dir_entry *fzm_iwf_loglevel_f = NULL;

/*MBTAC COUNTERS */
uint32_t ul_uplane_pkt_cntr;
uint32_t ul_cplane_pkt_cntr;
uint32_t ul_splane_pkt_cntr;
uint32_t ul_mcast_cplane_pkt_cntr;

uint32_t dl_uplane_pkt_cntr;
uint32_t dl_cplane_pkt_cntr;
uint32_t dl_splane_pkt_cntr;
uint32_t dl_mcast_cplane_pkt_cntr;

/* IP COUNTERS */
uint64_t dl_bh_total_pkt_cntr       = 0;     //overall received packets on Backhaul interface
uint64_t dl_bh_total_bytes_cntr     = 0;     //overall received kilobytes on Backhaul interface
uint64_t dl_bh_total_drop_pkt_cntr  = 0;     //overall discarded received packets on Backhaul interface
uint64_t dl_bh_mcast_total_ip_vlan_pkt_cntr    = 0; //overall received mcast packets on Backhaul interface

uint64_t ul_bh_total_pkt_cntr       = 0;     //overall transmitted packets on Backhaul interface
uint64_t ul_bh_total_bytes_cntr     = 0;     //overall transmitted kilobytes on Backhaul interface
uint64_t ul_bh_mcast_total_ip_vlan_pkt_cntr    = 0; //overall transmitted mcast packets on Backhaul interface

/* IP COUNTERS for IPv4 */
uint64_t dl_bh_total_pkt_cntr_ipv4       = 0;     //overall received packets on Backhaul interface
uint64_t dl_bh_total_bytes_cntr_ipv4     = 0;     //overall received kilobytes on Backhaul interface
uint64_t dl_bh_total_drop_pkt_cntr_ipv4  = 0;     //overall discarded received packets on Backhaul interface

uint64_t ul_bh_total_pkt_cntr_ipv4       = 0;     //overall transmitted packets on Backhaul interface
uint64_t ul_bh_total_bytes_cntr_ipv4     = 0;     //overall transmitted kilobytes on Backhaul interface
uint64_t ul_bh_mcast_total_bytes_cntr_ipv4 = 0;      //overall transmitted mcast kilobyteson Backhaul interface
uint64_t ul_bh_mcast_total_pkt_cntr_ipv4   = 0;      //overall transmitted mcast packets on Backhaul interface

/* IP COUNTERS IPv6 */
uint64_t dl_bh_total_pkt_cntr_ipv6       = 0;     //overall received packets on Backhaul interface
uint64_t dl_bh_total_bytes_cntr_ipv6     = 0;     //overall received kilobytes on Backhaul interface
uint64_t dl_bh_total_drop_pkt_cntr_ipv6  = 0;     //overall discarded received packets on Backhaul interface

uint64_t ul_bh_total_pkt_cntr_ipv6       = 0;     //overall transmitted packets on Backhaul interface
uint64_t ul_bh_total_bytes_cntr_ipv6     = 0;     //overall transmitted kilobytes on Backhaul interface

//IPv6 mcast

uint64_t ul_mcast_pkt_cntr_ipv6        = 0;     //overall transmitted Neighbour Discovery & MLD packets on Backhaul interface
uint64_t ul_mcast_bytes_cntr_ipv6      = 0;     //overall transmitted Neighbour Discovery & MLD bytes on Backhaul interface

uint64_t dl_mcast_pkt_cntr_ipv6        = 0;     //overall received  Neighbour Discovery & MLD packets on Backhaul interface
uint64_t dl_mcast_bytes_cntr_ipv6      = 0;     //overall received  Neighbour Discovery & MLD bytes on Backhaul interface

uint64_t dl_mcast_disc_pkt_cntr_ipv6   = 0;    //overall discarded Neighbour Discovery & MLD packets on Backhaul interface
uint64_t dl_mcast_disc_bytes_cntr_ipv6 = 0;  //overall discarded Neighbour Discovery & MLD bytes on Backhaul interface

//bh mcast
uint64_t dl_bh_mcast_total_bytes_cntr = 0;      //overall transmitted mcast kilobyteson Backhaul interface
uint64_t dl_bh_mcast_total_pkt_cntr   = 0;      //overall transmitted mcast packets on Backhaul interface

uint64_t ul_bh_mcast_total_bytes_cntr = 0;      //overall transmitted mcast kilobyteson Backhaul interface
uint64_t ul_bh_mcast_total_pkt_cntr   = 0;      //overall transmitted mcast packets on Backhaul interface

uint64_t dl_bh_mcast_disc_total_bytes_cntr = 0;      //overall transmitted mcast kilobyteson Backhaul interface
uint64_t dl_bh_mcast_disc_total_pkt_cntr   = 0;      //overall transmitted mcast packets on Backhaul interface


struct   proc_dir_entry *fzm_iwf_ip_cntrs_f = NULL;


/* function prototype */
static int iwf_cntrs_proc_read(struct seq_file *m, void *v);
static ssize_t iwf_cntrs_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos);
static int iwf_table_dump(struct seq_file *m, void *v);

static int iwf_loglevel_help_proc_read(struct seq_file *m, void *v);
static int iwf_loglevel_add_proc_read(struct seq_file *m, void *v);
static ssize_t iwf_loglevel_add_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos);

static int iwf_loglevel_delete_proc_read(struct seq_file *m, void *v);
static ssize_t iwf_loglevel_delete_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos);

static int iwf_loglevel_proc_read(struct seq_file *m, void *v);

static int iwf_ip_cntrs_proc_read(struct seq_file *m, void *v);
static ssize_t iwf_ip_cntrs_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos);

static int iwf_counter_dump_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_cntrs_proc_read, NULL);
}
static int iwf_table_dump_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_table_dump, NULL);
}
static int iwf_loglevel_help_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_loglevel_help_proc_read, NULL);
}
static int iwf_loglevel_add_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_loglevel_add_proc_read, NULL);
}
static int iwf_loglevel_delete_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_loglevel_delete_proc_read, NULL);
}
static int iwf_loglevel_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_loglevel_proc_read, NULL);
}
static int iwf_ip_cntrs_open(struct inode *inode, struct file *file)
{
    return single_open(file,iwf_ip_cntrs_proc_read, NULL);
}


static const struct file_operations iwf_proc_counters_fops = {
    .owner  = THIS_MODULE,
    .open   = iwf_counter_dump_open,
    .read   = seq_read,
    .write  = iwf_cntrs_proc_write,
    .llseek = seq_lseek,
    .release= single_release
};

static const struct file_operations iwf_proc_table_dump_fops = {
    .owner = THIS_MODULE,
    .open = iwf_table_dump_open,
    .read = seq_read,
    .llseek = seq_lseek,
    .release = single_release
};

static const struct file_operations iwf_proc_loglevel_help_fops = {
    .owner = THIS_MODULE,
    .open  = iwf_loglevel_help_open,
    .read  = seq_read,
    .llseek = seq_lseek,
    .release = single_release
};

static const struct file_operations iwf_proc_loglevel_add_fops = {
    .owner = THIS_MODULE,
    .open  = iwf_loglevel_add_open,
    .read  = seq_read,
    .write = iwf_loglevel_add_proc_write,
    .llseek = seq_lseek,
    .release = single_release
};

static const struct file_operations iwf_proc_loglevel_delete_fops = {
    .owner = THIS_MODULE,
    .open  = iwf_loglevel_delete_open,
    .read  = seq_read,
    .write = iwf_loglevel_delete_proc_write,
    .llseek = seq_lseek,
    .release = seq_release,
};

static const struct file_operations iwf_proc_loglevel_fops = {
    .owner = THIS_MODULE,
    .open = iwf_loglevel_open,
    .read = seq_read,
    .llseek = seq_lseek,
    .release = single_release
};

static const struct file_operations iwf_proc_ip_cntrs_fops = {
    .owner  = THIS_MODULE,
    .open   = iwf_ip_cntrs_open,
    .read   = seq_read,
    .write  = iwf_ip_cntrs_proc_write,
    .llseek = seq_lseek,
    .release= single_release
};

static struct nf_hook_ops br_nf_ops[] __read_mostly = {
    {
        .hook = vfilter_br_nf_local,
        .owner = THIS_MODULE,
        .pf = NFPROTO_BRIDGE,
        .hooknum = NF_BR_LOCAL_IN,
        .priority = NF_BR_PRI_FILTER_BRIDGED,
    },
    {
        .hook = vfilter_br_nf_local,
        .owner = THIS_MODULE,
        .pf = NFPROTO_BRIDGE,
        .hooknum = NF_BR_FORWARD,
        .priority = NF_BR_PRI_FILTER_BRIDGED,
    },
    {
        .hook = vfilter_br_nf_local,
        .owner = THIS_MODULE,
        .pf = NFPROTO_BRIDGE,
        .hooknum = NF_BR_LOCAL_OUT,
        .priority = NF_BR_PRI_FILTER_BRIDGED,
    }
};


#if 0
DL flow
-------

    From IXIA we will send plain GTPU packet
In IWF table FTM ccid (tx_ccid) can be any range. CPID (rx_cpid) is fixed 3 values (check in requrment-vijay)
    Before sending to msg_rtr
    .   Perform IWF lookup and construct SICP
.   Iwf simulator will replace tx_sicad with registered call back SICAD ( ex: 0x01010101)
    Forward to msg_rtr
    Now Msgrtr will call IWF and IWF will send the pkt to 30000 port towards.
    -   NOTE : the RX_SICAD will always be : what we had registered (eg: 0x01010101).


    UL flow
    -------
    From IXIA configure value is SICP GTPU packet with below values
RX-sicad . predefined range and != DL CPID(DL rx_cpid)
    TX-sicad . any value
Before sending to msg_rtr tx_sicad = registered call back SICAD ( ex:0x01010101)
    Forward the packet to msg_rtr
    IWF callback called
    -   Do swap Rx<-> Tx
    Perform IWF lookup on rx_cpid
    Construct GTPU packet and forward to SGW

#endif


void printPacket(char *mesg, unsigned char *p, int len)
{

    int newline=10;
    printk("         %s         ",mesg);
    printk("\n---------------------------------- \n");
    while(len--)
    {
        printk(" %.2X", *p);
        p++;
        newline--;
        if(0 == newline)
        {
            newline=10;
            printk("\n");
        }
    }
    printk("\n---------------------------------- \n");
    printk("\n\n");
}

/* ------------------------------------------------------------------------------ */
#ifndef _IWF_EN
unsigned long ByteSwap2 (unsigned long nLongNumber)
{
    return (((nLongNumber&0x000000FF)<<24)+((nLongNumber&0x0000FF00)<<8)+((nLongNumber&0x00FF0000)>>8)+((nLongNumber&0xFF000000)>>24));
}
#endif
/* ------------------------------------------------------------------------------ */
void MsgRtrARMCallback(struct sk_buff * buffer)
{
    int                              iRxPktLen  = 0;
    iwf_sic_message_hdr_t            *sicMsgHdr       = NULL;
    iwf_gtp_hdr_t                    *gtpuHdr         = NULL;
    iwf_sicad_t                      sicad;

    if(buffer)
    {
        iRxPktLen = skb_is_nonlinear(buffer) ? skb_headlen(buffer) : buffer->len;

        /* check if in simulation node */
        if(true == iwf_sim_flag)
        {
            sicMsgHdr = (iwf_sic_message_hdr_t *)buffer->data;
            gtpuHdr   = (iwf_gtp_hdr_t *) (buffer->data + sizeof(iwf_sic_message_hdr_t));

            if(!sicMsgHdr || !gtpuHdr) {
                dev_kfree_skb_any(buffer);
                buffer = NULL;
                DBG_PRT_ALL("iwf cb enter null header\n");
                return;
            }

            DUMP_MSG((void*)sicMsgHdr, sizeof(iwf_sic_message_hdr_t)+sizeof(iwf_gtp_hdr_t), 0);
            DUMP_MSG((void*)gtpuHdr, sizeof(iwf_gtp_hdr_t), 0);

            /* get the sender sicad */
            sicad.cpid = sicMsgHdr->tx_cpid ;
            sicad.nid  = sicMsgHdr->tx_nid  ;
#ifdef _IWF_EN
            sicad.sicad_32 = ntohl(sicad.sicad_32);
#endif
            if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(" %s rx cpid =%x\n",__func__,(sicMsgHdr->rx_cpid));
                    printk(" %s rx nid  =%x\n",__func__,(sicMsgHdr->rx_nid));
                    printk(" %s tx cpid =%x\n",__func__,(sicMsgHdr->tx_cpid));
                    printk(" %s tx nid  =%x\n",__func__,(sicMsgHdr->tx_nid));

                    printk("%s ----------TEID =%x\n",__func__, ntohl(gtpuHdr->teid));
                    printk(" %s TX SICAD  from MsgRtr : =%x\n",__func__,sicad.sicad_32);
                }
            }

////==> k2l
#if 0
            /* check for the TX cpid to seggregate between UL & DL traffic  */
            if( (__IWF_UL_MIN_CCID <= sicad.cpid ) &&
                    (__IWF_UL_MAX_CCID >= sicad.cpid ))
            {
                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk("%s ---------- \n",__func__);
                        printk("%s UL Traffic \n",__func__);
                        printk("%s ---------- \n",__func__);
                    }
                }
                /* copy TX SICAD to RX before giving for classification */
                sicMsgHdr->rx_cpid = sicMsgHdr->tx_cpid;
                sicMsgHdr->rx_nid  = sicMsgHdr->tx_nid;
                //gtpuHdr->teid      = htonl(gtpuHdr->teid);
                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(" %s RX SICAD changed CPID %x, NID : %x\n",__func__,sicMsgHdr->rx_cpid,sicMsgHdr->rx_nid) ;
                    }
                }

                sendPktToSgw(buffer);

            }
            else
            {
                if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk("%s ---------- \n",__func__);
                        printk("%s DL Traffic \n",__func__);
                        printk("%s ---------- \n",__func__);
                    }
                }

                /* before sending towards LMP convert swap RX & TX SICAD */
                sicad.cpid = (sicMsgHdr->rx_cpid);
                sicad.nid  = (sicMsgHdr->rx_nid);
                //                sicad.sicad_32 = htonl(sicad.sicad_32);

                sicMsgHdr->rx_cpid = (sicMsgHdr->tx_cpid);
                sicMsgHdr->rx_nid  = (sicMsgHdr->tx_nid);
                sicMsgHdr->tx_cpid = sicad.cpid;
                sicMsgHdr->tx_nid  = sicad.nid;
#if 0
                sicad.cpid = (sicMsgHdr->rx_cpid);
                sicad.nid  = (sicMsgHdr->rx_nid);
                sicad.sicad_32 = htonl(sicad.sicad_32);
                sicMsgHdr->rx_cpid = sicad.cpid;
                sicMsgHdr->rx_nid  = sicad.nid;

                gtpuHdr->teid      = htonl(gtpuHdr->teid);
#endif

                /* send to LMP */
                sendPktToLMP(buffer);
            }
////<== k2l
#endif
            sendPktToSgw(buffer);
////<== k2l
        }
        else
        {
            /* if not in DSP SIM mode send to Sgw */
            sendPktToSgw(buffer);
        }
    }

}

/* ------------------------------------------------------------------------------ */
/* Looped back Egress packets (by Tupu) handler, forwaded to S-GW */
void sendPktToSgw(struct sk_buff * skb)
{
    bool                             isSkbToBeFreed   = true;
    int                              iRxPktLen        = 0;
    int                              iIndexToCfgTable = -1;
    sicp_ip_data_t                   *sicpIpData      = NULL;
    struct dst_entry                 *dst             = NULL;
    struct net                       *net             = NULL;
    iwf_sic_message_hdr_t            *sicMsgHdr       = NULL;
    iwf_gtp_hdr_t                    *gtpuHdr         = NULL;
    sicp_ip_data_t                   temp_trsw_entry  = {0};
    iwf_ip_addr_t                    uip_dst = {{{{0}}}};
    iwf_ip_addr_t                    uip_src = {{{{0}}}};
    _IWF_PKT_CLASSIFIER_RESULT       pktClassifier = _IWF_PKT_CLASSIFIER_INVALID_ARG;
    struct flowi                     fl;

    if(skb)
    {
        // k2l work around
        skb->dev = dev_eth0;

        net = dev_net(skb->dev);
        tx_total_ul_gtpu_pkt_from_tupu++;

        iRxPktLen = skb_is_nonlinear(skb) ? skb_headlen(skb) : skb->len;
        DBG_PRT_ALL("iRxPktLen[%d] skblen[%d] lte_sicp_ip_table[%p]\n", iRxPktLen, skb->len, lte_sicp_ip_table);

        ///// k2l debug
        if(!lte_sicp_ip_table) {
            dev_kfree_skb_any(skb);
            skb = NULL;
            tx_ul_gtpu_pkt_drop_null_pkt_rcvd++;
            return;
        }

        sicMsgHdr = (iwf_sic_message_hdr_t *)skb->data;
        gtpuHdr   = (iwf_gtp_hdr_t *) (skb->data + sizeof(iwf_sic_message_hdr_t));
        /* lookup in UL lookup-table */
        sicpIpData = get_sicp_ip_connection ((sicMsgHdr->rx_cpid), lte_sicp_ip_table);

        if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk("--------- RX CPID KMOD=%x\n",(sicMsgHdr->rx_cpid));
                printk("----------TEID =%x\n",ntohl(gtpuHdr->teid));
                printk("\n %s after  sicp =%p\n",__func__,sicpIpData);
            }
        }

        if(NULL != sicpIpData)
        {
            /* forward packet to SGW */
            pktClassifier = _IWF_FWD_PKT_TO_SGW;
            memcpy(&uip_dst.addr_u, &ipList[sicpIpData->uip_dst_index].addr_u, sizeof(UIpAddress_t));
            uip_dst.version = ipList[sicpIpData->uip_dst_index].version;

            memcpy(&uip_src.addr_u, &ipList[sicpIpData->uip_src_index].addr_u, sizeof(UIpAddress_t));
            uip_src.version = ipList[sicpIpData->uip_src_index].version;
            DBG_PRT_ALL("uip src[0x%x], dst[0x%x]\n", uip_src.addr_u.ipv4Address.ipa_32, uip_dst.addr_u.ipv4Address.ipa_32);
        }
        else /* not in UL lookup-table */
        {
            if( (0 == gtpuHdr->teid) && (_IWF_GTP_ECHO_REQ == gtpuHdr->msg_type))
            {
                /* lookup in Config table */
                for(iIndexToCfgTable = 0;iIndexToCfgTable < MAX_SGW_IP_ADDRESSES; iIndexToCfgTable++)
                {
                    /* FIXME : this is a temporary fix till PS fix is available
                       if((Strsw_config_table[iIndexToCfgTable].tx_cpid == ntohs((sicMsgHdr->rx_cpid))) &&
                       */
                    if((Strsw_config_table[iIndexToCfgTable].tx_cpid == ((sicMsgHdr->rx_cpid))) &&
                            (trsw_config_table_place[iIndexToCfgTable] == GTPU_TRSW_USED))
                    {
                        memcpy(&uip_dst, &Strsw_config_table[iIndexToCfgTable].uip_dst, sizeof(iwf_ip_addr_t));
                        memcpy(&uip_src, &Strsw_config_table[iIndexToCfgTable].uip_src, sizeof(iwf_ip_addr_t));
                        temp_trsw_entry.teid_ul  = gtpuHdr->teid;
                        temp_trsw_entry.dst_port = LTM_GTPU_PRT;
                        temp_trsw_entry.src_port = LTM_GTPU_PRT;
                        temp_trsw_entry.ip_src_index = Strsw_config_table[iIndexToCfgTable].ip_src_index;
                        temp_trsw_entry.dscp = (_IWF_GTP_MGMT_MSG_DSCP << 2);

                        sicpIpData = (sicp_ip_data_t*) &temp_trsw_entry;
                        sicpIpData->ip_ver = uip_dst.version;
                        pktClassifier = _IWF_FWD_ECHO_REQUEST_TO_SGW;
                        break;
                    }
                }/* for loop */
            }/* if( (0 == gtpuHdr->teid) && */
        }/* else  of not in lookup-table */

        /* is entry available in either UL table 'OR' config table */
        if(sicpIpData)
        {
            switch(pktClassifier)
            {
                case _IWF_FWD_PKT_TO_SGW:
                case _IWF_FWD_ECHO_REQUEST_TO_SGW:
                    {
                        struct udphdr       *udph         =  NULL;
                        struct net_device   *dev          = skb->dev;

                        /* strip off sic header */
                        skb_pull(skb,sizeof(iwf_sic_message_hdr_t));

                        if (sicpIpData->ip_ver == IWF_IPV4_VERSION)
                        {
                            struct iphdr        *ip_header    =  NULL;
                            struct rtable       *rt           =  NULL;
                            static unsigned short id = 40;
                            unsigned short dscp = 0;

                            tx_ul_gtpu_pkt_from_tupu_ipv4++;

                            /* add room for IP and udp header */
                            skb_push(skb,sizeof(struct iphdr) + sizeof(struct udphdr));
                            skb_reset_network_header(skb);

                            /* get pointers to ip and udp header */
                            ip_header  =  (struct iphdr *)skb->data;
                            udph       =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

                            ip_header->ihl = 5;
                            ip_header->version = IWF_IPV4_VERSION;
                            ip_header->id = htons (id++);
                            ip_header->frag_off = 0;
                            /* UDP protocol */
                            ip_header->protocol = IPPROTO_UDP;

                            /* update the dscp value in ip header */
                            if (pktClassifier == _IWF_FWD_PKT_TO_SGW)
                            {
                                dscp = sicpIpData->dscp;
                                ip_header->tos = (sicpIpData->dscp << 2);
                            }
                            else
                            {
                                dscp = _IWF_GTP_MGMT_MSG_DSCP;
                                ip_header->tos = _IWF_GTP_MGMT_MSG_DSCP << 2;
                            }

                            /* fill the destination ip i.e. S-GW IP and source ip */
                            ip_header->saddr = htonl(uip_src.ip4_addr32); //RANS
                            ip_header->daddr = htonl(uip_dst.ip4_addr32);
                            ip_header->tot_len = htons(skb->len);

                            /* do route lookup */
                            rt = ip_route_output(dev_net(dev), ip_header->daddr, ip_header->saddr,
                                    (ip_header->tos), 0);

                            if (!IS_ERR(rt))
                            {

                                /* calculate ttl for the packet */
                                ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

                                /*recompute ip header checksum*/
                                ip_send_check(ip_header);

                                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                                {
                                    if(0 != printk_ratelimit())
                                    {
                                        printk("\n skb->dev->name    : %s ",skb->dev->name);
                                        printk("\n rt->dst.dev->name : %s ",rt->dst.dev->name);
                                    }
                                }


                                /* fill the udp header */
                                udph->len =  htons(skb->len - sizeof(struct iphdr) );
                                udph->source  = htons(_IWFD_DL_UDP_PORT_);
                                udph->dest  = htons(_IWFD_DL_UDP_PORT_);


                                /* disable udp checksum check */
                                udph->check = 0;


                                /*  fill in route details in skb */
                                skb_dst_set(skb, &rt->dst);

#ifdef __IWFD_SET_SKB_PRIORITY
                                /* set skb->priority to 0 */
                                skb->priority = dscp;
                                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                                {
                                    if(0 != printk_ratelimit())
                                    {
                                        printk("dscp=%d", dscp);
                                    }
                                }
#endif
                                /* check if ipsec is disabled */
                                if(false == iwf_ipsec_flag)
                                {

                                    //MBTAC Marking
                                    mbTacMarkingUplaneUplink(skb);
                                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                                    dst_output(skb);

                                    /* kernel will take care of freeing it */
                                    isSkbToBeFreed = false;

                                    tx_ul_gtpu_pkt_to_sgw_cntr_ipv4++;
                                }
                                else
                                {
                                    struct flowi4 fl4 =
                                    {
                                        .flowi4_oif = inet_iif(skb),
                                        .flowi4_tos = ip_header->tos,
                                        .flowi4_proto = IPPROTO_UDP,
                                        .fl4_sport = _IWFD_DL_UDP_PORT_,
                                        .fl4_dport = _IWFD_DL_UDP_PORT_,
                                    };

                                    /* clasify the flow and update flow */
                                    security_skb_classify_flow(skb, flowi4_to_flowi(&fl4));

                                    /* check whether pkt is thru with IPsec policy check */
                                    if (xfrm_decode_session(skb, flowi4_to_flowi(&fl4), AF_INET) < 0)
                                    {
                                        tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4++;
                                    }
                                    else
                                    {
                                        /* check for route and do xfrm lookup */
                                        rt = ip_route_output_flow(net, &fl4, NULL);
                                        if (IS_ERR(rt))
                                        {
                                            tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4++;
                                        }
                                        else
                                        {
                                            skb_dst_set(skb, &rt->dst);
                                            /* Pre-Fragmentation Hook*/
                                            pre_frag_hook(skb);

                                            /* do we have a route */
                                            rt = (struct rtable *)skb_dst(skb);
                                            if (rt)
                                            {
                                                /* calculate ttl for the packet */
                                                ip_header->ttl = ip4_dst_hoplimit(&rt->dst);
                                                /*recompute ip header checksum*/
                                                ip_send_check(ip_header);
                                                /*do mbtac marking*/
                                                mbTacMarkingUplaneUplink(skb);
                                                /* send the pkt out */
                                                dst_output(skb);
                                                /* kernel will take care of freeing it */
                                                isSkbToBeFreed   = false;
                                                tx_ul_gtpu_pkt_to_sgw_cntr_ipv4++;
                                            }
                                            else
                                            {
                                                tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4++;
                                            }
                                        }
                                    }//else (!xfrm4_policy_check)
                                }//else (iwf_ipsec_flag)
                            }//if (!IS_ERR(rt))
                            else
                            {
                                tx_ul_gtpu_pkt_drop_no_route_ipv4++;
                            }
                        }//sicpIpData->ip_ver == IWF_IPV4_VERSION
                        else if (sicpIpData->ip_ver == IWF_IPV6_VERSION)
                        {
                            struct ipv6hdr     *ip6_header   =  NULL;
                            struct rt6_info     *rt           =  NULL;

                            tx_ul_gtpu_pkt_from_tupu_ipv6++;

                            /* add room for IPV6 and udp header */
                            skb_push(skb,sizeof(struct ipv6hdr) + sizeof(struct udphdr));
                            skb_reset_network_header(skb);

                            /* get pointers to ip and udp header */
                            ip6_header =  (struct ipv6hdr *)skb->data;
                            udph       =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

                            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                            {
                                if(0 != printk_ratelimit())
                                {
                                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                                }
                            }
                            ip6_header->version  = IWF_IPV6_VERSION;
                            ip6_header->priority = sicpIpData->dscp >> 2; //Set DSCP
                            ip6_header->flow_lbl[0] = sicpIpData->dscp << 6;;//Set DSCP
                            ip6_header->flow_lbl[1] = 0;;
                            ip6_header->flow_lbl[2] = 0;;
                            ip6_header->nexthdr = NEXTHDR_UDP;
                            ip6_header->hop_limit = 0xff;
                            ip6_header->payload_len = htons(skb->len - sizeof(struct ipv6hdr));
                            ip6_header->saddr.s6_addr32[0] = uip_src.addr_u.ipv6Address.ipa_32[0];
                            ip6_header->saddr.s6_addr32[1] = uip_src.addr_u.ipv6Address.ipa_32[1];
                            ip6_header->saddr.s6_addr32[2] = uip_src.addr_u.ipv6Address.ipa_32[2];
                            ip6_header->saddr.s6_addr32[3] = uip_src.addr_u.ipv6Address.ipa_32[3];

                            ip6_header->daddr.s6_addr32[0] = uip_dst.addr_u.ipv6Address.ipa_32[0];
                            ip6_header->daddr.s6_addr32[1] = uip_dst.addr_u.ipv6Address.ipa_32[1];
                            ip6_header->daddr.s6_addr32[2] = uip_dst.addr_u.ipv6Address.ipa_32[2];
                            ip6_header->daddr.s6_addr32[3] = uip_dst.addr_u.ipv6Address.ipa_32[3];
#ifdef __IWFD_SET_SKB_PRIORITY
                            /* set skb->priority  */
                            skb->priority = sicpIpData->dscp ;
                            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                            {
                                if(0 != printk_ratelimit())
                                {
                                    printk("dscp=%d,, pcp = %d", sicpIpData->dscp, sicpIpData->dscp);
                                }
                            }
#endif

                            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                            {
                                if(0 != printk_ratelimit())
                                {
                                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                                    printk(KERN_INFO "%s %d ip6_header->payload_len:%d", __FUNCTION__, __LINE__,
                                            ntohs(ip6_header->payload_len));
                                    printk(KERN_INFO "%s %d Dst Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                                            ip6_header->daddr.s6_addr32[0],
                                            ip6_header->daddr.s6_addr32[1],
                                            ip6_header->daddr.s6_addr32[2],
                                            ip6_header->daddr.s6_addr32[3]);

                                    printk(KERN_INFO "%s %d Source Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                                            ip6_header->saddr.s6_addr32[0],
                                            ip6_header->saddr.s6_addr32[1],
                                            ip6_header->saddr.s6_addr32[2],
                                            ip6_header->saddr.s6_addr32[3]);

                                    printk(KERN_INFO "\n skb->dev->name    : %s ",skb->dev->name);
                                }
                            }

                            /* fill the udp header */
                            udph->len =  htons(skb->len - sizeof(struct ipv6hdr) );
                            udph->source = htons(_IWFD_DL_UDP_PORT_);
                            udph->dest  = htons(_IWFD_DL_UDP_PORT_);

                            udph->check = 0;
                            udph->check = csum_ipv6_magic(&ip6_header->saddr,
                                    &ip6_header->daddr,ntohs(udph->len), IPPROTO_UDP,
                                    csum_partial(udph, ntohs(udph->len), 0));;
                            if (udph->check == 0)
                                udph->check = CSUM_MANGLED_0;

                            rt = rt6_lookup(dev_net(dev), &ip6_header->daddr, &ip6_header->saddr, 0, 0);
                            /*  fill in route details in skb */
                            if (rt && !IS_ERR(rt))
                            {
                                //If Getway is zero then fill destination address into getway
                                if (ipv6_addr_any(&rt->rt6i_gateway))
                                {
                                    memcpy(&rt->rt6i_gateway, &ip6_header->daddr, sizeof(struct in6_addr));
                                }

                                /*  fill in route details in skb */
                                skb_dst_set(skb, &rt->dst);

                                //Set skb->local_df to 1 to allow fragmentation.
                                skb->local_df = 1;

                                /* check if ipsec is disabled */
                                if(false == iwf_ipsec_flag)
                                {
                                    //MBTAC Marking
                                    mbTacMarkingUplaneUplink(skb);

                                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                                    dst_output(skb);

                                    /* kernel will take care of freeing it */
                                    isSkbToBeFreed = false;

                                    tx_ul_gtpu_pkt_to_sgw_cntr_ipv6++;
                                }
                                else //if ipsec is enabled
                                {
                                    if (xfrm_decode_session(skb, &fl, AF_INET6) < 0)
                                    {
                                        tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6++;

                                    }
                                    else
                                    {
                                        skb_dst_force(skb);

                                        /* fill in the xfrm details */
                                        dst = xfrm_lookup(net, skb_dst(skb), &fl, NULL, 0);
                                        if (IS_ERR(dst))
                                        {
                                            tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6++;
                                            dst = NULL;
                                        }
                                        else
                                        {
                                            skb_dst_set(skb, dst);

                                            //MBTAC Marking
                                            mbTacMarkingUplaneUplink(skb);

                                            /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                                            dst_output(skb);

                                            /* kernel will take care of freeing it */
                                            isSkbToBeFreed = false;

                                            tx_ul_gtpu_pkt_to_sgw_cntr_ipv6++;

                                        }
                                    }
                                }
                            }// if (rt)
                            else
                            {
                                tx_ul_gtpu_pkt_drop_no_route_ipv6++;
                            }
                        }// else if (sicpIpData->ip_ver == IWF_IPV6_VERSION)

                        break;
                    }

                default:
                    tx_ul_invalid_gtpu_pkt_drop_cntr++;
                    break;

            }/* switch(pktClassifier) */
        }
        else
        {
            tx_ul_gtpu_pkt_drop_no_entry++;
        }
        if((true == isSkbToBeFreed) && (NULL != skb))
        {
            /* free the skb */
            dev_kfree_skb_any(skb);
            skb = NULL;
        }

    }//if(skb)
    else
    {
        tx_ul_gtpu_pkt_drop_null_pkt_rcvd++;
    }

    return;
}

#ifndef __IWF_ONLY_ONE_CALL_BK
/* ------------------------------------------------------------------------------ */
/* Looped back Ingress packets (by Tupu) handler, will be forwarded to 30000 port */
void sendPktToLMP(struct sk_buff *skb )
{
    bool     isSkbToBeFreedLMP = true;
    uint32_t iPktLen = 0;

    if(skb)
    {
        iPktLen = skb_is_nonlinear(skb) ? skb_headlen(skb) : skb->len;

        /* for debugging send the packet to 30000 port */
        if(iPktLen > 0)
        {
            struct udphdr       *udph         =  NULL;
            struct net_device   *dev          = skb->dev;
            struct rtable       *rt           =  NULL;
            struct iphdr        *ip_header    =  NULL;

            static unsigned short id = 40;

            /* add room for IP and udp header */
            skb_push(skb,sizeof(struct iphdr) + sizeof(struct udphdr));
            skb_reset_network_header(skb);

            /* get pointers to ip and udp header */
            ip_header  =  (struct iphdr *)skb->data;
            udph       =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

            ip_header->ihl = 5;

            ip_header->version = IWF_IPV4_VERSION;
            ip_header->id = htons (id++);
            ip_header->frag_off = 0;

            /* UDP protocol */
            ip_header->protocol = IPPROTO_UDP;
            /* update the dscp value in ip header */
            ip_header->tos = 0;


            /* fill the destination ip i.e. S-GW IP and source ip */
            ip_header->saddr = ((unsigned long)in_aton("12.12.12.12"));
            ip_header->daddr = ((unsigned long)in_aton("12.12.12.13"));

            /* do route lookup */
            rt = ip_route_output(dev_net(dev), ip_header->daddr, ip_header->saddr,
                    (ip_header->tos), 0);


            if (!IS_ERR(rt))
            {
                ip_header->tot_len = htons(skb->len);


                /* calculate ttl for the packet */
                ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

                /*recompute ip header checksum*/
                ip_send_check(ip_header);

                /* fill the udp header */
                udph->len =  htons(skb->len - sizeof(struct iphdr) );
                udph->source  = htons(__IWFD_DSP_SIM_SOURCE_PORT__);
                udph->dest  = htons(_IWFD_UL_UDP_PORT_);


                /* disable udp checksum check */
                udph->check = 0;


                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk("\n skb->dev->name    : %s ",skb->dev->name);
                        printk("\n rt->dst.dev->name : %s ",rt->dst.dev->name);
                    }
                }

                /*  fill in route details in skb */
                skb_dst_set(skb, &rt->dst);

#ifdef __IWFD_SET_SKB_PRIORITY
                /* set skb->priority to 0 */
                skb->priority = __IWFD_SKB_PRIORITY__;
#endif

                /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                dst_output(skb);
                isSkbToBeFreedLMP = false;
            }
        }// if(iPktLen > 0)

        if((true == isSkbToBeFreedLMP) && (NULL != skb))
        {
            /* free the skb */
            dev_kfree_skb_any(skb);
            skb = NULL;
        }

    }
    else
    {
        if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO _IWF_MODULE_NAME": ERR!! %s  called with NULL data \n",__func__);
            }
        }
    }

}
#endif

/* ------------------------------------------------------------------------------ */
int iwfSendSicpMessage(sicp_data_t *sicpData, struct sk_buff *skb)
{
    /* add SIC header */
    int                    iPktLen  = -1;
    iwf_sicad_t            sicad;
    iwf_sic_message_hdr_t  *tempSicpBuf = NULL;
    iwf_sic_message_hdr_t  sicpMsgHdr;
    int iph_udph_len = 0;
#ifdef __IWF_ENABLE_MSG_RTR
    int msgRtrRetval = -1;
#endif

#ifndef _IWF_EN
    struct udphdr    *udph         = NULL;
    uint16_t         iDataLen      = 0;

    /* get udp start */
    udph =  (struct udphdr *)(skb->data +
            ((((struct iphdr *) skb->data)->version) == 4 ?
             sizeof(struct iphdr) : sizeof(struct ipv6hdr)));

    iDataLen = ntohs(udph->len);
#endif

    if((sicpData->rx_nid != 0) && (sicpData->rx_cpid != 0) && (sicpData->tx_nid != 0) \
            && (sicpData->tx_cpid != 0) &&(sicpData->mtype != 0))
    {
        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(" %s before rx cpid =%x\n",__func__,(sicpData->rx_cpid));
                printk(" %s before rx nid  =%x\n",__func__,(sicpData->rx_nid));
                printk(" %s before tx cpid =%x\n",__func__,(sicpData->tx_cpid));
                printk(" %s before tx nid  =%x\n",__func__,(sicpData->tx_nid));
                printk(" %s sicpData->mtype=%x\n",__func__,(sicpData->mtype));
                printk(" %s before sicad   = %x\n",__func__,sicad.sicad_32);
            }
        }

        sicpMsgHdr.hdr.version       = _IWF_SIC_VERSION;
        sicpMsgHdr.hdr.reserved      = _IWF_SIC_RESERVED;
        sicpMsgHdr.mtn               = (sicpData->mtype);
        sicpMsgHdr.rx_nid            = (sicpData->rx_nid);
        sicpMsgHdr.rx_cpid           = (sicpData->rx_cpid);
        sicpMsgHdr.tx_nid            = (sicpData->tx_nid);
        sicpMsgHdr.tx_cpid           = (sicpData->tx_cpid);

#if 0
// This is hack for testing
#ifdef DBG_DUMP_ON
        sicpMsgHdr.rx_nid = 0x123A;
#endif
#endif

#ifdef _IWF_EN
        /* convert RX SICAD to BE */
        sicad.nid                    = (sicpData->rx_nid);
        sicad.cpid                   = (sicpData->rx_cpid);
        sicad.sicad_32               = htonl(sicad.sicad_32);
        sicpMsgHdr.rx_nid            = sicad.nid;
        sicpMsgHdr.rx_cpid           = sicad.cpid;
#endif


#ifdef _IWF_EN
        /* convert TX SICAD to BE */
        sicad.nid                    = (sicpData->tx_nid);
        sicad.cpid                   = (sicpData->tx_cpid);
        sicad.sicad_32               = htonl(sicad.sicad_32);
        sicpMsgHdr.tx_nid            = sicad.nid;
        sicpMsgHdr.tx_cpid           = sicad.cpid;
#endif

        sicpMsgHdr.hdr_flags_1       = _IWF_SIC_HEADER_FLAGS_1;
        sicpMsgHdr.hdr_flags_2       = _IWF_SIC_HEADER_FLAGS_2;

        /* strip out IP and udp header  */

        iph_udph_len = (((struct iphdr *) skb->data)->version == IWF_IPV4_VERSION)?
            IPV4HDR_P_UDPHDR_LEN : IPV6HDR_P_UDPHDR_LEN;
        skb_pull(skb, iph_udph_len);

        /* create room for sic header */
        tempSicpBuf =  (iwf_sic_message_hdr_t *)skb_push(skb,sizeof(iwf_sic_message_hdr_t));

        /* fill the sicp payload length */
        iPktLen =  (skb->len);
        sicpMsgHdr.mlen = (skb->len);
        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(" %s after  rx cpid =%x\n",__func__,(sicpMsgHdr.rx_cpid));
                printk(" %s after  rx nid  =%x\n",__func__,(sicpMsgHdr.rx_nid));
                printk(" %s after  tx cpid =%x\n",__func__,(sicpMsgHdr.tx_cpid));
                printk(" %s after  tx nid  =%x\n",__func__,(sicpMsgHdr.tx_nid));
                printk(" %s sicpData->mtype=%x\n",__func__,(sicpMsgHdr.mtn));
                printk(" %s sicpMsgHdr.mlen=%x\n",__func__,(sicpMsgHdr.mlen));
            }
        }

        //TODO : check if memcpy is required
        memcpy((void *)tempSicpBuf,(void *)&sicpMsgHdr,sizeof(iwf_sic_message_hdr_t));


#ifdef __IWF_ENABLE_MSG_RTR

        /* call DMA loopback function */
        if(true == iwf_sim_flag)
        {
            if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk("\n =============%s \n sicpMsgHdr.hdr.version  : %x\n",__func__,         sicpMsgHdr.hdr.version    );
                    printk("\n sicpMsgHdr.hdr.reserved : %x\n",         sicpMsgHdr.hdr.reserved   );
                    printk("\n sicpMsgHdr.mtn          : %x\n",         sicpMsgHdr.mtn            );
                    printk("\n sicpMsgHdr.rx_nid       : %x\n",         sicpMsgHdr.rx_nid         );
                    printk("\n sicpMsgHdr.rx_cpid      : %x\n",         sicpMsgHdr.rx_cpid        );
                    printk("\n sicpMsgHdr.tx_nid       : %x\n",         sicpMsgHdr.tx_nid         );
                    printk("\n sicpMsgHdr.tx_cpid      : %x\n",         sicpMsgHdr.tx_cpid        );
                    printk("\n sicpMsgHdr.mlen         : %x\n",         sicpMsgHdr.mlen           );
                    printk("\n sicpMsgHdr.hdr_flags_1  : %x\n",         sicpMsgHdr.hdr_flags_1    );
                    printk("\n sicpMsgHdr.hdr_flags_2  : %x\n",         sicpMsgHdr.hdr_flags_2    );
                    printk(KERN_INFO _IWF_MODULE_NAME"%s: IN SIMULATION MODE : %d\n",__func__,iwf_sim_flag);
                }
            }

            msgRtrRetval = MsgRtrARMSend_loopback(skb);
            if(msgRtrRetval)
            {
                if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMSend_loopback failed : %d\n",__func__, msgRtrRetval);
                    }
                }
                dev_kfree_skb_any(skb);
                skb = NULL;
            }
            else
            {
                if(IWF_LOG_LEVEL_DL_SUCC_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(KERN_INFO _IWF_MODULE_NAME"%s: SUCCESS!! MsgRtrARMSend_loopback success : %d\n",__func__, msgRtrRetval);
                    }
                }

            }
        }
        else
        {
            msgRtrRetval = MsgRtrARMSend(skb);
            if(msgRtrRetval)
            {
                if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {

                        printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMSend failed : %d\n",__func__, msgRtrRetval);
                    }
                }
                dev_kfree_skb_any(skb);
                skb = NULL;
            }
#ifndef _IWF_EN
            else
            {
                if(IWF_LOG_LEVEL_DL_SUCC_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {

                        printk(KERN_INFO _IWF_MODULE_NAME"%s: SUCCESS!! MsgRtrARMSend success : %d\n",__func__, msgRtrRetval);
                    }
                }

            }
#endif

        }
        if(IWF_LOG_LEVEL_PKT_DUMP & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                if (NULL != skb)
                {
                    uint32_t *temp_buf_ptr = NULL;
                    int j=0;
                    int iLoop =0;
                    printk(" ******** DL Pkt DUMP start ********\n");
                    temp_buf_ptr = (uint32_t*) skb->data;
                    printk(" DL Pkt size %u\n\n",iPktLen);
                    for(iLoop =0,j=0; iLoop < skb->len/4 ; iLoop++,j++)
                    {
                        printk("0x%X ",*(temp_buf_ptr+iLoop));
                        if (!(j%4))
                            printk("\n");
                    }
                    printk(" ******** DL Pkt DUMP end ********\n");
                }
            }
        }
#else
        sendPktToLMP(skb);
        return NF_STOLEN;
#endif
    }
    else
    {
        rx_dl_gtpu_pkt_drop_wrong_sicp_hdr++;
    }

    return iPktLen;
}
/* ------------------------------------------------------------------------------ */
short iwfSendIpsecPkt(struct sk_buff * skb)
{
    short  sRetVal   = _IWF_FAILURE_;
    struct iphdr    *ip_header      =  NULL;
    struct rtable   *rt             =  NULL;
    struct dst_entry *dst         = NULL;
    struct net       *net         = dev_net(skb->dev);
    struct flowi     fl;

    /* skb->sp should be cleared, as it may be possible that the we used
       the same skb which linux had provided for ingress packet. In that case
       sp might be already popluated, and this will lead the linux to re-use
       skb->sp content as it is and when we do a policy check it will fail,
       to prohibit this skb->sp should be cleared.*/
    secpath_reset(skb);

    /* IP header */
    ip_header  =  (struct iphdr *)skb->data;

    if(4 == ip_header->version)
    {
        /* check whether pkt is thru with IPsec policy check */
        if (xfrm_decode_session(skb, &fl, AF_INET) < 0)
        {
            tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4++;
        }
        else
        {
            skb_dst_force(skb);

            /* fill in the xfrm details */
            dst = xfrm_lookup(net, skb_dst(skb), &fl, NULL, 0);
            if (IS_ERR(dst))
            {
                tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4++;
            }
            else
            {
                skb_dst_set(skb, dst);
                /* do we have a route */
                rt = (struct rtable *)skb_dst(skb);
                if (rt)
                {
                    /* calculate ttl for the packet */
                    ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

                    /*recompute ip header checksum*/
                    ip_send_check(ip_header);

                    /*do mbtac marking*/
                    mbTacMarkingUplaneUplink(skb);

                    /* send the pkt out */
                    dst_output(skb);

                    /* return SUCCESS */
                    sRetVal = _IWF_SUCCESS_;
                }
                else
                {
                    tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4++;
                }
            }//else (!xfrm4_route_forward(skb))
        }//else (!xfrm4_policy_check)
    }
    else//IPv6
    {
        if (xfrm_decode_session(skb, &fl, AF_INET6) < 0)
        {
            tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6++;
        }
        else
        {
            skb_dst_force(skb);
            /* fill in the xfrm details */
            dst = xfrm_lookup(net, skb_dst(skb), &fl, NULL, 0);
            if (IS_ERR(dst))
            {
                tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6++;
            }
            else
            {
                skb_dst_set(skb, dst);

                /*do mbtac marking*/
                mbTacMarkingUplaneUplink(skb);

                /* done with packet formatting, send pkt to Sgw */
                dst_output(skb);

                /* return SUCCESS */
                sRetVal = _IWF_SUCCESS_;
            }
        }
    }

    return sRetVal;

}

/* ------------------------------------------------------------------------------ */
int iwfSendEchoResponseToSgw(struct sk_buff * skb)
{
    unsigned char         *gtpStart       =  NULL;
    int                   iPktLen         =  -1;
    struct iphdr          *ip_header      =  NULL;
    struct udphdr         *udph           =  NULL;
    struct net_device     *dev            =  skb->dev;
    iwf_gtp_mgmt_pkt_t    gtpMgmtPkt      =  {{{0}}};
    bool                  isSkbToBeFreed  = true;

    ip_header  =  (struct iphdr *)skb->data;
    if(ip_header->version == 4)
    {
        rx_dl_gtpu_mgmt_local_req_ipv4++;
        /* calculate the packet size of EchoResponse packet */
        iPktLen = sizeof(iwf_gtp_hdr_t) + sizeof(iwf_gtp_hdr_optional_t) +\
                  sizeof(gtp_recovery_counter_ie_data_t) +\
                  sizeof(struct iphdr) + sizeof(struct udphdr);

        /* get gtp start address */

        gtpStart = skb->data + sizeof(struct iphdr) + sizeof(struct udphdr);
    }
    else//if(ip_header->version == 6)
    {
        rx_dl_gtpu_mgmt_local_req_ipv6++;
        /* calculate the packet size of EchoResponse packet */
        iPktLen = sizeof(iwf_gtp_hdr_t) + sizeof(iwf_gtp_hdr_optional_t) +\
                  sizeof(gtp_recovery_counter_ie_data_t) +\
                  sizeof(struct ipv6hdr) + sizeof(struct udphdr);

        /* get gtp start address */

        gtpStart = skb->data + sizeof(struct ipv6hdr) + sizeof(struct udphdr);

    }


    /* if Teid 0 & Msgtype is 0x01 then its Echo request
       An IP packet GTP-U: ECHO REQUEST message is identified by the below parameters:
       - destination IP address = own eNB U-plane IP address in the IP header
       -  protocol = UDP in the IP-header
       -  destination UDP port = 2152 in the UDP header
       -  TEID=0 in the GTP-U header
       */

    /** we have to send response back, request will not be forwarded to TUPu **/
    /* update gtpu header parameter as follows :
       Version=1
       PT=1
       Flag S=1
       Sequence Number = SN which was received in ECHO REQUEST
       Flag E=0
       Flag PN=0
       MessageType = 02
       TEID=0
       Next Extension Header Type IE = 0
       */
    gtpMgmtPkt.gtpuHdr.flags.version           = 1;
    gtpMgmtPkt.gtpuHdr.flags.proto_type        = 1;
    gtpMgmtPkt.gtpuHdr.flags.ext_hdr_present   = 0;
    gtpMgmtPkt.gtpuHdr.flags.seq_nr_present    = 1;
    gtpMgmtPkt.gtpuHdr.flags.npdu_num_present  = 0;
    gtpMgmtPkt.gtpuHdr.msg_type                = _IWF_GTP_ECHO_RESP;
    gtpMgmtPkt.gtpuHdr.total_len               = htons(sizeof(iwf_gtp_hdr_optional_t) + sizeof(gtp_recovery_counter_ie_data_t));;
    gtpMgmtPkt.gtpuHdr.teid                    = 0;

    if(0 == ((iwf_gtp_mgmt_pkt_t *)gtpStart)->gtpuHdr.flags.seq_nr_present)
    {
        gtpMgmtPkt.payload.optionalField.seq_num = 0;
    }
    else
    {
        gtpMgmtPkt.payload.optionalField.seq_num = ((iwf_gtp_mgmt_pkt_t *)gtpStart)->payload.optionalField.seq_num;
    }
    gtpMgmtPkt.payload.optionalField.npdu_num                = 0;
    gtpMgmtPkt.payload.optionalField.nxt_exthdr_type         = 0;

    /* update GTP-U: ECHO RESPONSE message content as follows :
       Recovery IE=0
       Private Extension IE is not present
       */
    gtpMgmtPkt.payload.extHdr.gtp_recovery_ie_data.type = _IWF_GTP_IE_TYPE_4;
    gtpMgmtPkt.payload.extHdr.gtp_recovery_ie_data.restart_counter = 0;

    /* : update IP header as follows:
       Destination IP address = S-GW IP address (copied from Source IP address of the GTP-U: ECHO REQUEST message)
       Destination UDP port = Source port from Echo Request
       Source IP address = same eNB IP address as in received packet.
       Source UDP port = 2152
       */
    if(ip_header->version == 4)
    {
        struct iphdr          ip_header_tmp   =  {0};
        struct rtable         *rt             =  NULL;

        /* make skb->tail to point at correct address */
        skb_trim(skb, iPktLen);

        /* IP and udp header */
        ip_header  =  (struct iphdr *)skb->data;
        udph       =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

        /* take a back up of destination IP which we received in packet */
        ip_header_tmp.daddr = ip_header->daddr;

        /* fill the destination ip i.e. S-GW IP and source ip -
           sgwIP needs to be updated from rx pkt */

        ip_header->daddr = ip_header->saddr;
        ip_header->saddr = ip_header_tmp.daddr; //RANS

        /* update IP pkt lenght */
        ip_header->tot_len = htons(iPktLen);

        /* update DSCP value = 46 (GTP-U management channel) */
        ip_header->tos = _IWF_GTP_MGMT_MSG_DSCP << 2;

        /* UDP protocol */
        ip_header->protocol = IPPROTO_UDP;

        /* fill the udp header */
        udph->source = htons(_IWFD_DL_UDP_PORT_);
        udph->dest   = htons(_IWFD_DL_UDP_PORT_);

        /* disable udp checksum check */
        udph->check = 0;

        /* update udp length */
        udph->len = htons(iPktLen - sizeof(struct iphdr));

        /* copy gtp header and payload */
        memcpy((void *)(skb->data +  sizeof(struct iphdr) + sizeof(struct udphdr) ),\
                (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct iphdr) + sizeof(struct udphdr)));

        /* do route lookup */
        rt = ip_route_output(dev_net(dev), ip_header->daddr, ip_header->saddr,(ip_header->tos), 0);

        if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk("\n Dest Addr         : %x ",ip_header->daddr);
                printk("\n SRC  Addr         : %x ",ip_header->saddr);
                printk("\n skb->dev->name    : %s ",skb->dev->name);
                printk("\n rt->dst.dev->name : %s ",rt->dst.dev->name);
            }
        }

        if (!IS_ERR(rt))
        {
            /* calculate ttl for the packet */
            ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

            /*recompute ip header checksum*/
            ip_send_check(ip_header);

            /*  fill in route details in skb */
            skb_dst_set(skb, &rt->dst);

            /* set skb->priority  */
            skb->priority = _IWF_GTP_MGMT_MSG_DSCP;

            /* check if ipsec is disabled */
            if(false == iwf_ipsec_flag)
            {
                /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                dst_output(skb);

                /* kernel will take care of freeing it */
                isSkbToBeFreed = false;

                rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

            }
            else
            {
                /* create and send IPSec pkt */
                if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                {
                    rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed   = false;
                }
            }//else (iwf_ipsec_flag)


        }
        else
        {
            rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4++;
        }
    }
    else//if(ip_header->version == 6)
    {
        struct ipv6hdr     *ip6_header   =  NULL;
        struct rt6_info    *rt           =  NULL;
        uint32_t            tmp_ip6add_32[4] = {0};

        /* make skb->tail to point at correct address */
        skb_trim(skb, iPktLen);

        /* IP and udp header */
        ip6_header  =  (struct ipv6hdr *)skb->data;
        udph       =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

        if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
            }
        }

        ip6_header->version  = IWF_IPV6_VERSION;
        ip6_header->priority = _IWF_GTP_MGMT_MSG_DSCP >> 2; //Set DSCP
        ip6_header->flow_lbl[0] = (__u8 )(_IWF_GTP_MGMT_MSG_DSCP << 6);;//Set DSCP
        ip6_header->flow_lbl[1] = 0;;
        ip6_header->flow_lbl[2] = 0;;
        ip6_header->nexthdr = NEXTHDR_UDP;
        ip6_header->hop_limit = 0xff;
        ip6_header->payload_len = htons(skb->len - sizeof(struct ipv6hdr));

        /*Swap Destination and source*/
        memcpy(tmp_ip6add_32, ip6_header->saddr.s6_addr32, sizeof(tmp_ip6add_32));
        memcpy(ip6_header->saddr.s6_addr32, ip6_header->daddr.s6_addr32, sizeof(tmp_ip6add_32));
        memcpy(ip6_header->daddr.s6_addr32, tmp_ip6add_32, sizeof(tmp_ip6add_32));

        if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                printk(KERN_INFO "%s %d ip6_header->payload_len:%d", __FUNCTION__, __LINE__,
                        ntohs(ip6_header->payload_len));
                printk(KERN_INFO "%s %d Dst Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                        ip6_header->daddr.s6_addr32[0],
                        ip6_header->daddr.s6_addr32[1],
                        ip6_header->daddr.s6_addr32[2],
                        ip6_header->daddr.s6_addr32[3]);

                printk(KERN_INFO "%s %d Source Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                        ip6_header->saddr.s6_addr32[0],
                        ip6_header->saddr.s6_addr32[1],
                        ip6_header->saddr.s6_addr32[2],
                        ip6_header->saddr.s6_addr32[3]);

                printk(KERN_INFO "\n skb->dev->name    : %s ",skb->dev->name);
            }
        }

        /* fill the udp header */
        udph->len =  htons(skb->len - sizeof(struct ipv6hdr) );
        udph->source = htons(_IWFD_DL_UDP_PORT_);
        udph->dest  = htons(_IWFD_DL_UDP_PORT_);
        /* copy gtp header and payload */
        memcpy((void *)(skb->data +  sizeof(struct ipv6hdr) + sizeof(struct udphdr) ),\
                (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct ipv6hdr) + sizeof(struct udphdr)));
        udph->check = 0;
        udph->check = csum_ipv6_magic(&ip6_header->saddr,
                &ip6_header->daddr,ntohs(udph->len), IPPROTO_UDP,
                csum_partial(udph, ntohs(udph->len), 0));;
        if (udph->check == 0)
            udph->check = CSUM_MANGLED_0;

        rt = rt6_lookup(dev_net(dev), &ip6_header->daddr, &ip6_header->saddr, 0, 0);
        /*  fill in route details in skb */
        if (rt && !IS_ERR(rt))
        {
            //If Getway is zero then fill destination address into getway
            if (ipv6_addr_any(&rt->rt6i_gateway))
            {
                memcpy(&rt->rt6i_gateway, &ip6_header->daddr, sizeof(struct in6_addr));
            }

            skb_dst_set(skb, &rt->dst);

            //Set skb->local_df to 1 to allow fragmentation.
            skb->local_df = 1;

            /* check if ipsec is disabled */
            if(false == iwf_ipsec_flag)
            {
                /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                dst_output(skb);

                /* kernel will take care of freeing it */
                isSkbToBeFreed = false;

                rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

            }
            else
            {
                /* create and send IPSec pkt */
                if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                {
                    rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed   = false;
                }
            }//else (iwf_ipsec_flag)

        }// if (rt)
        else
        {
            rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6++;

            /* free the skb */
            dev_kfree_skb_any(skb);
            skb = NULL;
        }
    }//else if(ip_header->version == 6)

    if((true == isSkbToBeFreed) && (NULL != skb))
    {
        /* free the skb */
        dev_kfree_skb_any(skb);
        skb = NULL;
    }

    /* return the pkt len */
    return iPktLen;
}
/* ------------------------------------------------------------------------------ */
int iwfSendSuppExthdrToSgw(struct sk_buff * skb)
{

    int                   iPktLen       = -1;
    bool                  isSkbToBeFreed   = true;
    iwf_gtp_mgmt_pkt_t    gtpMgmtPkt    = {{{0}}};
    struct iphdr          *ip_header    =  NULL;
    struct udphdr         *udph         =  NULL;
    struct net_device     *dev;
    /* we have to send response back, request will not be forwarded to TUPu **/
    /* update gtpu header parameter as follows :
       Version=1
       PT=1
       Flag S=1
       Sequence Number =0
       Flag E=0
       Flag PN=0
       MessageType = 31
       TEID=0
       Next Extension Header Type IE = 0
       */
    gtpMgmtPkt.gtpuHdr.flags.version           = 1;
    gtpMgmtPkt.gtpuHdr.flags.proto_type        = 1;
    gtpMgmtPkt.gtpuHdr.flags.ext_hdr_present   = 0;
    gtpMgmtPkt.gtpuHdr.flags.seq_nr_present    = 1;
    gtpMgmtPkt.gtpuHdr.flags.npdu_num_present  = 0;
    gtpMgmtPkt.gtpuHdr.msg_type                = _IWF_GTP_SUPPOTERD_EXT_HDR;
    gtpMgmtPkt.gtpuHdr.total_len               = htons(sizeof(iwf_gtp_hdr_optional_t) + sizeof(gtp_ext_hdr_noti_ie_data_t)) ;
    gtpMgmtPkt.gtpuHdr.teid                    = 0;

    gtpMgmtPkt.payload.optionalField.seq_num                 = 0;
    gtpMgmtPkt.payload.optionalField.npdu_num                = 0;
    gtpMgmtPkt.payload.optionalField.nxt_exthdr_type         = 0;

    gtpMgmtPkt.payload.extHdr.gtp_ext_hdr_noti_ie_data.type_3 = GTP_IE_TYPE_3;
    gtpMgmtPkt.payload.extHdr.gtp_ext_hdr_noti_ie_data.length = GTP_IE_TYPE_3_LEN;
    gtpMgmtPkt.payload.extHdr.gtp_ext_hdr_noti_ie_data.ext_hdr_type_1 = _IWF_GTP_EXT_HDR_UDP;
    gtpMgmtPkt.payload.extHdr.gtp_ext_hdr_noti_ie_data.ext_hdr_type_2 = _IWF_GTP_EXT_HDR_PDCP;

    if(skb)
    {
        dev = skb->dev;
        ip_header  =  (struct iphdr *)skb->data;

        if(ip_header->version == 4)
        {
            struct iphdr          ip_header_tmp = {0};
            struct rtable         *rt           =  NULL;

            /* calculate the packet size of SuppExthdr packet */
            iPktLen = (sizeof(iwf_gtp_hdr_t)+ sizeof(iwf_gtp_hdr_optional_t)+\
                    sizeof(gtp_ext_hdr_noti_ie_data_t))+\
                      sizeof(struct iphdr) + sizeof(struct udphdr);
            /* make skb->tail to point at correct address */
            skb_trim(skb, iPktLen);
            /* update IP header as follows:
               Destination IP address = S-GW IP address (copied from Source IP address of the GTP-U: ECHO REQUEST message)
               Destination UDP port = Source port from Echo Request
               Source IP address = same eNB IP address as in received packet.
               Source UDP port = 2152
               DSCP value = 46 (GTP-U management channel)
               */

            /* IP and udp header */
            ip_header  =  (struct iphdr *)skb->data;
            udph       =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

            /* take a back up of destination IP which we received in packet */
            ip_header_tmp.daddr = ip_header->daddr;
            /* fill the destination ip i.e. S-GW IP and source ip
             *            -  sgwIP needs to be updated from rx pkt */

            ip_header->daddr = ip_header->saddr;
            ip_header->saddr = ip_header_tmp.daddr; //RANS
            ip_header->tot_len = htons(skb->len);

            /* update DSCP value = 46 (GTP-U management channel) */
            ip_header->tos = _IWF_GTP_MGMT_MSG_DSCP << 2;

            /* UDP protocol */
            ip_header->protocol = IPPROTO_UDP;

            /* fill the udp header */
            udph->source = htons(_IWFD_DL_UDP_PORT_);
            udph->dest   = htons(_IWFD_DL_UDP_PORT_);

            /* disable udp checksum check */
            udph->check = 0;

            /* update udp length */
            udph->len = htons(iPktLen - sizeof(struct iphdr));

            /* copy gtp header and payload */
            memcpy((void *)(skb->data +  sizeof(struct iphdr) + sizeof(struct udphdr) ),\
                    (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct iphdr) + sizeof(struct udphdr)));

            /* do route lookup */
            rt = ip_route_output(dev_net(dev), ip_header->daddr, ip_header->saddr,
                    (ip_header->tos), 0);

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk("\n Dest Addr         : %x ",ip_header->daddr);
                    printk("\n SRC  Addr         : %x ",ip_header->saddr);
                    printk("\n skb->dev->name    : %s ",skb->dev->name);
                    printk("\n rt->dst.dev->name : %s ",rt->dst.dev->name);
                }
            }

            if (!IS_ERR(rt))
            {
                /* calculate ttl for the packet */
                ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

                /*recompute ip header checksum*/
                ip_send_check(ip_header);

                /*  fill in route details in skb */
                skb_dst_set(skb, &rt->dst);

                /* set skb->priority   */
                skb->priority = _IWF_GTP_MGMT_MSG_DSCP ;

                /* check if ipsec is disabled */
                if(false == iwf_ipsec_flag)
                {
                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                    dst_output(skb);

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed = false;

                    rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

                }
                else
                {
                    /* create and send IPSec pkt */
                    if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                    {
                        rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

                        /* kernel will take care of freeing it */
                        isSkbToBeFreed   = false;
                    }
                }//else (iwf_ipsec_flag)
            }
            else
            {
                rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4++;
            }
        }
        else//if(ip_header->version == 6)
        {
            struct ipv6hdr     *ip6_header   =  NULL;
            struct rt6_info    *rt           =  NULL;
            uint32_t            tmp_ip6add_32[4] = {0};

            /* calculate the packet size of SuppExthdr packet */
            iPktLen = (sizeof(iwf_gtp_hdr_t)+ sizeof(iwf_gtp_hdr_optional_t)+\
                    sizeof(gtp_ext_hdr_noti_ie_data_t))+\
                      sizeof(struct ipv6hdr) + sizeof(struct udphdr);
            /* make skb->tail to point at correct address */
            skb_trim(skb, iPktLen);

            /* IP and udp header */
            ip6_header  =  (struct ipv6hdr *)skb->data;
            udph       =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                }
            }

            ip6_header->version  = IWF_IPV6_VERSION;
            ip6_header->priority = _IWF_GTP_MGMT_MSG_DSCP >> 2; //Set DSCP
            ip6_header->flow_lbl[0] = (__u8)(_IWF_GTP_MGMT_MSG_DSCP << 6);;//Set DSCP
            ip6_header->flow_lbl[1] = 0;;
            ip6_header->flow_lbl[2] = 0;;
            ip6_header->nexthdr = NEXTHDR_UDP;
            ip6_header->hop_limit = 0xff;
            ip6_header->payload_len = htons(skb->len - sizeof(struct ipv6hdr));

            //Swap Destination and source
            memcpy(tmp_ip6add_32, ip6_header->saddr.s6_addr32, sizeof(tmp_ip6add_32));
            memcpy(ip6_header->saddr.s6_addr32, ip6_header->daddr.s6_addr32, sizeof(tmp_ip6add_32));
            memcpy(ip6_header->daddr.s6_addr32, tmp_ip6add_32, sizeof(tmp_ip6add_32));

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                    printk(KERN_INFO "%s %d ip6_header->payload_len:%d", __FUNCTION__, __LINE__,
                            ntohs(ip6_header->payload_len));
                    printk(KERN_INFO "%s %d Dst Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                            ip6_header->daddr.s6_addr32[0],
                            ip6_header->daddr.s6_addr32[1],
                            ip6_header->daddr.s6_addr32[2],
                            ip6_header->daddr.s6_addr32[3]);

                    printk(KERN_INFO "%s %d Source Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                            ip6_header->saddr.s6_addr32[0],
                            ip6_header->saddr.s6_addr32[1],
                            ip6_header->saddr.s6_addr32[2],
                            ip6_header->saddr.s6_addr32[3]);

                    printk(KERN_INFO "\n skb->dev->name    : %s ",skb->dev->name);
                }
            }

            /* fill the udp header */
            udph->len =  htons(skb->len - sizeof(struct ipv6hdr) );
            udph->source = htons(_IWFD_DL_UDP_PORT_);
            udph->dest  = htons(_IWFD_DL_UDP_PORT_);

            /* copy gtp header and payload */
            memcpy((void *)(skb->data +  sizeof(struct ipv6hdr) + sizeof(struct udphdr) ),\
                    (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct ipv6hdr) + sizeof(struct udphdr)));
            udph->check = 0;
            udph->check = csum_ipv6_magic(&ip6_header->saddr,
                    &ip6_header->daddr,ntohs(udph->len), IPPROTO_UDP,
                    csum_partial(udph, ntohs(udph->len), 0));;
            if (udph->check == 0)
                udph->check = CSUM_MANGLED_0;

            rt = rt6_lookup(dev_net(dev), &ip6_header->daddr, &ip6_header->saddr, 0, 0);
            /*  fill in route details in skb */
            if (rt && !IS_ERR(rt))
            {
                //If Getway is zero then fill destination address into getway
                if (ipv6_addr_any(&rt->rt6i_gateway))
                {
                    memcpy(&rt->rt6i_gateway, &ip6_header->daddr, sizeof(struct in6_addr));
                }

                skb_dst_set(skb, &rt->dst);

                //Set skb->local_df to 1 to allow fragmentation.
                skb->local_df = 1;


                /* check if ipsec is disabled */
                if(false == iwf_ipsec_flag)
                {
                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                    dst_output(skb);

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed = false;

                    rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

                }
                else
                {
                    /* create and send IPSec pkt */
                    if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                    {
                        rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

                        /* kernel will take care of freeing it */
                        isSkbToBeFreed   = false;
                    }
                }//else (iwf_ipsec_flag)

            }// if (rt)
            else
            {
                rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6++;
            }
        }//else if(ip_header->version == 6)
    }// if(skb)


    if((true == isSkbToBeFreed) && (NULL != skb))
    {
        /* free the skb */
        dev_kfree_skb_any(skb);
        skb = NULL;
    }

    return iPktLen;
}

/* ------------------------------------------------------------------------------ */
_IWF_PKT_CLASSIFIER_RESULT  classifyOperationOnDLPkt(uint8_t *data,int iDataLen,
        _IWF_TABLE_LOOKUP_RESULT lookupResult,sicp_data_t *outSicpConfigData, bool isFP, struct sk_buff *skb)
{
    uint8_t                           table_index = 0;
    uint32_t                          pTeid=0;
    iwf_ip_addr_t                     sgwIp = {{{{0}}}};
    iwf_ip_addr_t                     enbIp = {{{{0}}}};
    struct iphdr                      *ip_header = NULL;
    iwf_gtp_mgmt_pkt_t                *gtpMgmtPkt = NULL;
    teid_sicp_gtpu_mgmt_data_t        *sicpMgmtData=NULL;
    gtp_error_ind_no_udp_ie_data_t    *gtpuhdr_ie = NULL;
    _IWF_PKT_CLASSIFIER_RESULT        pktClassify = _IWF_PKT_CLASSIFIER_INVALID_ARG;


    if(NULL != data)
    {
        gtpMgmtPkt = (iwf_gtp_mgmt_pkt_t *)data;

        /* if teid lookup successful */
        if(_IWF_AVAILABLE_IN_DL_TABLE == lookupResult)
        {
            /* extension header not present */
            if(_IWF_EXT_HDR_NOT_PRESENT == gtpMgmtPkt->gtpuHdr.flags.ext_hdr_present)
            {
                /* fwd packet to TUPU */
                pktClassify = _IWF_FWD_PKT_TO_TUPU;
            }
            else if(_IWF_EXT_HDR_PRESENT == gtpMgmtPkt->gtpuHdr.flags.ext_hdr_present) // extension header present
            {
                /* check for supported ext hdr */
                if(((gtpMgmtPkt->payload.optionalField.nxt_exthdr_type == _IWF_GTP_EXT_HDR_UDP) &&\
                            (gtpMgmtPkt->gtpuHdr.msg_type == _IWF_GTP_EXT_HDR_MSG_TYPE_UDP))\
                        ||
                        ((gtpMgmtPkt->payload.optionalField.nxt_exthdr_type == _IWF_GTP_EXT_HDR_PDCP) &&\
                         (gtpMgmtPkt->gtpuHdr.msg_type == _IWF_GTP_EXT_HDR_MSG_TYPE_PDCP)))
                {

                    if(gtpMgmtPkt->payload.extHdr.gtp_nxt_ext_hdr.tlen == 1)
                    {
                        uint8_t *ehdr_ptr = (uint8_t *)((&(gtpMgmtPkt->payload.extHdr.gtp_nxt_ext_hdr.tlen))+\
                                (((gtpMgmtPkt->payload.extHdr.gtp_nxt_ext_hdr.tlen)*4)-1));
                        if(0 == *ehdr_ptr)
                        {
                            pktClassify =  _IWF_FWD_PKT_TO_TUPU;
                        }
                        else
                        {
                            /* more than one ext hdr, we should drop and send supported ext to Sgw */
                            pktClassify = _IWF_SEND_SUPPORTED_EXT_HDR_TO_SGW;
                        }
                    }
                    else
                    {
                        pktClassify =  _IWF_DROP_PKT;
                    }

                }
                else if (0 == gtpMgmtPkt->payload.optionalField.nxt_exthdr_type)
                {
                    pktClassify =  _IWF_FWD_PKT_TO_TUPU;
                }
                else
                {
                    pktClassify = _IWF_SEND_SUPPORTED_EXT_HDR_TO_SGW;
                }

            }/* else if(_IWF_EXT_HDR_PRESENT */

        }/* if(lookupResult == _IWF_AVAILABLE_IN_DL_TABLE) */
        else /* teid not available in table */
        {
            if(outSicpConfigData)
            {
                memset((void *)outSicpConfigData, 0, sizeof(sicp_data_t));

                if((gtpMgmtPkt->gtpuHdr.flags.ext_hdr_present == _IWF_EXT_HDR_PRESENT)&& \
                        ((gtpMgmtPkt->payload.optionalField.nxt_exthdr_type != _IWF_GTP_EXT_HDR_UDP)&& \
                         (gtpMgmtPkt->payload.optionalField.nxt_exthdr_type != _IWF_GTP_EXT_HDR_PDCP)))
                {
                    pktClassify = _IWF_SEND_SUPPORTED_EXT_HDR_TO_SGW;
                }
                else
                {
                    /* check if teid != 0 */
                    if(0 != gtpMgmtPkt->gtpuHdr.teid)
                    {
                        /* received invalid teid, drop pkt and send Err indication to S-GW*/
                        pktClassify = _IWF_SEND_ERR_IND_TO_SGW;
                    }
                    else /* 0 teid */
                    {
                        switch(gtpMgmtPkt->gtpuHdr.msg_type)
                        {
                            case _IWF_GTP_ERR_IND:
                                if ((1 == gtpMgmtPkt->gtpuHdr.flags.seq_nr_present) ||
                                        (1 == gtpMgmtPkt->gtpuHdr.flags.npdu_num_present) ||
                                        (1 == gtpMgmtPkt->gtpuHdr.flags.ext_hdr_present))

                                {
                                    if(0 == gtpMgmtPkt->payload.optionalField.nxt_exthdr_type )
                                    {
                                        gtpuhdr_ie = (gtp_error_ind_no_udp_ie_data_t *)( (uint8_t *)data + GTP_MGMT_HDR_LEN_WITH_OPTS_WO_EXT);
                                    }
                                    else
                                    {
                                        gtpuhdr_ie = (gtp_error_ind_no_udp_ie_data_t *)( (uint8_t *)data + GTP_MGMT_HDR_LEN_WITH_OPTS_AND_EXT);
                                    }
                                }
                                else
                                {
                                    gtpuhdr_ie = (gtp_error_ind_no_udp_ie_data_t *)( (uint8_t *)data + GTP_MGMT_HDR_LEN_WO_OPTS_AND_EXT);
                                }

                                /* extract pTeid & S-GW ip for lookup in Error Lookup table */
                                pTeid = (uint32_t)((gtpuhdr_ie->p_teid_b3)<< 24) + (uint32_t)(gtpuhdr_ie->p_teid_b2 << 16)
                                    + (uint32_t)(gtpuhdr_ie->p_teid_b1 << 8) + (uint32_t)(gtpuhdr_ie->p_teid_b0);

                                if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                                {
                                    if(0 != printk_ratelimit())
                                    {
                                        printk(" Error IND GSN Add length byte : %d ", gtpuhdr_ie->length_0);
                                        printk(" Error IND TEID: 0x%x ", pTeid);
                                        printk("\n pTeid_b3 : 0x%x",gtpuhdr_ie->p_teid_b3);
                                        printk("\n pTeid_b2 : 0x%x",gtpuhdr_ie->p_teid_b2);
                                        printk("\n pTeid_b1 : 0x%x",gtpuhdr_ie->p_teid_b1);
                                        printk("\n pTeid_b0 : 0x%x",gtpuhdr_ie->p_teid_b0);
                                    }
                                }

                                if(gtpuhdr_ie->length_0 == GTP_IE_TYPE_2_LEN_IPV4)
                                {
                                    sgwIp.ip4_addr32 = (uint32_t)(gtpuhdr_ie->p_ip);
                                    //sgwIp.ip4_addr32 = htonl(gtpuhdr_ie->p_ip);
                                    sgwIp.ip4_addr32 = htonl(sgwIp.ip4_addr32);
                                    sgwIp.version = IPv4;
                                    //printk(" swip IPV4 Address: %08X ", sgwIp.addr_u.ipv4Address.ipa_32);
                                }
                                else if (gtpuhdr_ie->length_0 == GTP_IE_TYPE_2_LEN_IPV6)
                                {
                                    //IPV6 address
                                    sgwIp.ip6_addr32[0] = gtpuhdr_ie->p_ip;
                                    sgwIp.ip6_addr32[1] = gtpuhdr_ie->p_ip1;
                                    sgwIp.ip6_addr32[2] = gtpuhdr_ie->p_ip2;
                                    sgwIp.ip6_addr32[3] = gtpuhdr_ie->p_ip3;
                                    sgwIp.version = IPv6;

                                    if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                                    {
                                        if(0 != printk_ratelimit())
                                        {
                                            printk(" Error IND swip IPV6 Address: %08X,%08X, %08X, %08X ",
                                                    sgwIp.ip6_addr32[0], sgwIp.ip6_addr32[1],
                                                    sgwIp.ip6_addr32[2], sgwIp.ip6_addr32[3]);
                                        }
                                    }
                                }

                                /* lookup in Error Lookup Table for the TUPU details */
                                sicpMgmtData = (teid_sicp_gtpu_mgmt_data_t *)get_dl_gtp_mgmt_error_ind_connection(pTeid,
                                        sgwIp);
                                /* is entry available in error table */
                                if (NULL == sicpMgmtData)
                                {
                                    pktClassify = _IWF_DROP_PKT;
                                }
                                else
                                {
                                    /* fill in the SIC details */
                                    memcpy((void *)outSicpConfigData,(void *)&(sicpMgmtData->sicp_data),sizeof(sicp_data_t));
                                    if(!outSicpConfigData->rx_nid)
                                    {
                                        pktClassify = _IWF_DROP_PKT;
                                    }
                                    else
                                    {
                                        pktClassify = _IWF_FWD_ERR_IND_TO_TUPU;
                                    }
                                }//else of if (NULL == sicpMgmtData)

                                break;

                            case _IWF_GTP_EXT_HDR_NOTI:
                                pktClassify =  _IWF_DROP_PKT;
                                break;

                            case _IWF_GTP_ECHO_REQ:
                                /* rx echo request from SGw send response to SGw */
                                pktClassify = _IWF_SEND_ECHO_RESPONSE_TO_SGW;
                                break;

                            case _IWF_GTP_ECHO_RESP:

                                if((NULL != skb) && (true == isFP))
                                {
                                    ip_header  =  (struct iphdr *)skb->data;

                                    /* if Teid 0 & Msgtype is 0x02 then its Echo response from Backhaul */
                                    /* get the src-ip & dest ip from the pkt received */
                                    if(ip_header->version == 4)
                                    {
                                        sgwIp.addr_u.ipv4Address.ipa_32 = ntohl(ip_header->saddr);
                                        sgwIp.version = IPv4;
                                        enbIp.addr_u.ipv4Address.ipa_32 = ntohl(ip_header->daddr); //RANS;
                                        enbIp.version = IPv4;
                                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                                        {
                                            if(0 != printk_ratelimit())
                                            {
                                                printk(" Echo Resp swip Address: %08X ", sgwIp.addr_u.ipv4Address.ipa_32);
                                                printk(" Echo Resp enb Address: %08X ", enbIp.addr_u.ipv4Address.ipa_32);
                                            }
                                        }

                                    }
                                    else if (ip_header->version == 6)
                                    {

                                        struct ipv6hdr *ip6_header =  (struct ipv6hdr *)skb->data;
                                        sgwIp.addr_u.ipv6Address.ipa_32[0]=ip6_header->saddr.s6_addr32[0];
                                        sgwIp.addr_u.ipv6Address.ipa_32[1]=ip6_header->saddr.s6_addr32[1];
                                        sgwIp.addr_u.ipv6Address.ipa_32[2]=ip6_header->saddr.s6_addr32[2];
                                        sgwIp.addr_u.ipv6Address.ipa_32[3]=ip6_header->saddr.s6_addr32[3];
                                        sgwIp.version = IPv6;
                                        enbIp.addr_u.ipv6Address.ipa_32[0] = ip6_header->daddr.s6_addr32[0]; //RANS;
                                        enbIp.addr_u.ipv6Address.ipa_32[1] = ip6_header->daddr.s6_addr32[1]; //RANS;
                                        enbIp.addr_u.ipv6Address.ipa_32[2] = ip6_header->daddr.s6_addr32[2]; //RANS;
                                        enbIp.addr_u.ipv6Address.ipa_32[3] = ip6_header->daddr.s6_addr32[3]; //RANS;
                                        enbIp.version = IPv6;
                                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                                        {
                                            if(0 != printk_ratelimit())
                                            {
                                                printk("Echo Resp sgw-ip Address: %08x %08x %08x %08x",
                                                        sgwIp.addr_u.ipv6Address.ipa_32[0], sgwIp.addr_u.ipv6Address.ipa_32[1],
                                                        sgwIp.addr_u.ipv6Address.ipa_32[2], sgwIp.addr_u.ipv6Address.ipa_32[3]);

                                                printk("Echo Resp enb Address: %08x %08x %08x %08x",
                                                        enbIp.addr_u.ipv6Address.ipa_32[0], enbIp.addr_u.ipv6Address.ipa_32[1],
                                                        enbIp.addr_u.ipv6Address.ipa_32[2], enbIp.addr_u.ipv6Address.ipa_32[3]);
                                            }
                                        }


                                    }

                                    /* lookup configTable table and get CCID by using S-GW IP as index */
                                    table_index = get_teid_0_sicp_connection(sgwIp,enbIp);

                                    /* if CCID is available : then convert IP to SIC message and fwd to Master TUPu */
                                    if (table_index < _IWF_MAX_SGW_IP_ADDRESSES)
                                    {
                                        outSicpConfigData->rx_nid   =  Strsw_config_table[table_index].rx_nid;
                                        outSicpConfigData->tx_nid   =  Strsw_config_table[table_index].tx_nid;
                                        outSicpConfigData->rx_cpid  =  Strsw_config_table[table_index].rx_cpid;
                                        outSicpConfigData->tx_cpid  =  Strsw_config_table[table_index].tx_cpid;
                                        outSicpConfigData->mtype    =  TRSW_UPLANE_RECEIVE_DATA_MSG;
                                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                                        {
                                            if(0 != printk_ratelimit())
                                            {
                                                printk("DL Echo Resp info rx_nid %x tx_nid %x rx_cpid %x tx_cpid %x mtype %x",
                                                        outSicpConfigData->rx_nid, outSicpConfigData->tx_nid,
                                                        outSicpConfigData->rx_cpid, outSicpConfigData->tx_cpid, outSicpConfigData->mtype);
                                            }
                                        }

                                        pktClassify = _IWF_FWD_ECHO_RESPONSE_TO_TUPU;

                                    }
                                    else
                                    {
                                        pktClassify = _IWF_DROP_PKT;
                                    }
                                }
                                else
                                {
                                    pktClassify = _IWF_DROP_PKT;
                                }

                                if(!outSicpConfigData->rx_nid)
                                {
                                    pktClassify = _IWF_DROP_PKT;
                                }
                                break;

                            default:
                                pktClassify = _IWF_DROP_PKT;
                                break;
                        }
                    }/* if(gtpMgmtPkt->gtpuHdr.teid == 0) */
                }
            } /* if(outSicpConfigData) */
            else
            {
                if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(KERN_INFO _IWF_MODULE_NAME": ERR!! classifyOperationOnDLPkt() sicp data not available\n");
                    }
                }
            } /* else if(outSicpConfigData) */

        }/*else  teid not available in table  */

    }/* if(data) */
    else
    {
        if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO _IWF_MODULE_NAME": ERR!! data is NULLL \n");
            }
        }
    }

    return pktClassify;
}

/* ------------------------------------------------------------------------------ */
#ifdef __USE_NET_FILTER__
bool check_payload_len(struct iphdr *ip_header, struct ipv6hdr *ip6_header)
{
    if ((ip_header->version == 6) &&
            (ntohs(ip6_header->payload_len) >= (sizeof(struct ipv6hdr) + sizeof(struct udphdr) +  sizeof(iwf_gtp_mgmt_pkt_t))) )
    {
        return true;
    }
    else if (((ip_header->version == 4) ) &&
            (ntohs(ip_header->tot_len) >= (sizeof(struct iphdr) + sizeof(struct udphdr) +  sizeof(iwf_gtp_mgmt_pkt_t))))
    {
        return true;
    }
    return false;
}

// To count the DL IP packets including fragments
void increment_DownlinkCounters(struct sk_buff *skb,uint16_t vlanid,uint16_t vlanMapid,uint8_t version)
{
    if(version == IWF_IPV6_VERSION)
    {
        struct ipv6hdr *ip6_header  =  (struct ipv6hdr *)skb->data;
        struct icmp6hdr *icmp6header =  NULL;
        struct  ipv6_opt_hdr *exthdr = NULL;

        if((ip6_header->nexthdr == NEXTHDR_HOP) || (ip6_header->nexthdr == NEXTHDR_ICMP))
        {
            if(ip6_header->nexthdr == NEXTHDR_HOP)
            {
                exthdr = (struct ipv6_opt_hdr *)(skb->data + sizeof(struct ipv6hdr));
                if(exthdr->nexthdr == NEXTHDR_ICMP)
                {
                    icmp6header = (struct icmp6hdr *)(skb->data + ((sizeof(struct ipv6hdr) )+ (HOP_BY_HOP_HDR_LENGTH) + exthdr->hdrlen));
                }
            }
            else
            {
                icmp6header = (struct icmp6hdr *)(skb->data + sizeof(struct ipv6hdr));
            }

            if ( NULL != icmp6header)
            {
                if((icmp6header->icmp6_type == ICMPV6_MGM_QUERY) || (icmp6header->icmp6_type == ICMPV6_MGM_REPORT) ||(icmp6header->icmp6_type == ICMPV6_MGM_REDUCTION ) ||
                        (icmp6header->icmp6_type == NDISC_ROUTER_SOLICITATION) || (icmp6header->icmp6_type == NDISC_ROUTER_ADVERTISEMENT)  ||
                        (icmp6header->icmp6_type == NDISC_NEIGHBOUR_SOLICITATION) ||(icmp6header->icmp6_type == NDISC_NEIGHBOUR_ADVERTISEMENT) ||
                        (icmp6header->icmp6_type == NDISC_REDIRECT))
                {
                    if(vlanid && vlanMapid)
                    {
                        dl_bh_mcast_total_ip_vlan_pkt_cntr++;
                        ipCounters[vlanMapid-1].vlanId = vlanid;
                        ipCounters[vlanMapid-1].dl_mcast_pkt_cntr_ipv6++;
                        ipCounters[vlanMapid-1].dl_mcast_bytes_cntr_ipv6+= skb->len;

                        ipCounters[vlanMapid-1].dl_bh_mcast_total_pkt_cntr++;
                        ipCounters[vlanMapid-1].dl_bh_mcast_total_bytes_cntr += skb->len;
                    }
                    else
                    {
                        dl_bh_mcast_total_ip_vlan_pkt_cntr++;
                        dl_mcast_pkt_cntr_ipv6++;
                        dl_mcast_bytes_cntr_ipv6 = dl_mcast_bytes_cntr_ipv6+skb->len;

                        dl_bh_mcast_total_pkt_cntr++;
                        dl_bh_mcast_total_bytes_cntr += skb->len;
                    }
                }
            }
        }
    }
    if(vlanid && vlanMapid)
    {
        ipCounters[vlanMapid-1].vlanId = vlanid;
        // To get the fragment count from "cb" of SKB
        if(0 == skb->cb[MBTAC_METADATA_FRAG_CNT])
        {
            ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr++;
            ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr += skb->len;

            if(version == IWF_IPV4_VERSION)
            {
                ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr_ipv4++;
                ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr_ipv4 += skb->len;
            }
            else if(version == IWF_IPV6_VERSION)
            {
                ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr_ipv6++;
                ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr_ipv6 += skb->len;
            }
        }
        else
        {
            ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr += skb->cb[MBTAC_METADATA_FRAG_CNT];
            ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);

            if(version == IWF_IPV4_VERSION)
            {
                ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr_ipv4 += skb->cb[MBTAC_METADATA_FRAG_CNT];
                ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr_ipv4 += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);
            }
            else if(version == IWF_IPV6_VERSION)
            {
                ipCounters[vlanMapid-1].dl_bh_total_pkt_cntr_ipv6 += skb->cb[MBTAC_METADATA_FRAG_CNT];
                ipCounters[vlanMapid-1].dl_bh_total_bytes_cntr_ipv6 += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);
            }
        }
    }
    else
    {
        // To get the fragment count from "cb" of SKB
        if(0 == skb->cb[MBTAC_METADATA_FRAG_CNT])
        {
            /* Increment the  total received packet count for eth0 */
            dl_bh_total_pkt_cntr++;

            /* Increment the  total received bytes count for eth0 */
            dl_bh_total_bytes_cntr =  dl_bh_total_bytes_cntr + skb->len;

            if(version == IWF_IPV4_VERSION)
            {
                dl_bh_total_pkt_cntr_ipv4 ++;
                dl_bh_total_bytes_cntr_ipv4 =  dl_bh_total_bytes_cntr + skb->len;
            }
            else if(version == IWF_IPV6_VERSION)
            {
                dl_bh_total_pkt_cntr_ipv6 ++;
                dl_bh_total_bytes_cntr_ipv6 =  dl_bh_total_bytes_cntr + skb->len;
            }
        }
        else
        {
            dl_bh_total_pkt_cntr += skb->cb[MBTAC_METADATA_FRAG_CNT]+1;
            dl_bh_total_bytes_cntr += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);

            if (version == IWF_IPV4_VERSION)
            {
                dl_bh_total_pkt_cntr_ipv4 += skb->cb[MBTAC_METADATA_FRAG_CNT]+1;
                dl_bh_total_bytes_cntr_ipv4 += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);
            }
            else if(version == IWF_IPV6_VERSION)
            {
                dl_bh_total_pkt_cntr_ipv6 += skb->cb[MBTAC_METADATA_FRAG_CNT]+1;
                dl_bh_total_bytes_cntr_ipv6 += (skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO]<< 8 | skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI]);
            }
        }
    }
}

void increamentDropCounter(struct sk_buff *skb, uint8_t version)
{
    char *devPtr = NULL;
    uint16_t vlanid=0,vlanMapid=0;
    if(skb->dev->name[_IFACE_VLAN_BIT] =='.')
    {
        devPtr = skb->dev->name;
        sscanf(devPtr+_IFACE_VLAN_ID_BIT,"%hd",&vlanid);
        if(vlanid < MAX_VLANID)
        {
            vlanMapid = vlanIds[vlanid];
        }
        if(vlanid && vlanMapid)
        {
            ipCounters[vlanMapid-1].vlanId = vlanid;
            ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr++;

            if (version == IWF_IPV4_VERSION)
            {
                ipCounters[vlanMapid-1].vlanId = vlanid;
                ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr_ipv4++;
            }
            else if(version == IWF_IPV6_VERSION)
            {
                ipCounters[vlanMapid-1].vlanId = vlanid;
                ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr_ipv6++;
            }
        }
    }
    else
    {
        dl_bh_total_drop_pkt_cntr++;

        if (version == IWF_IPV4_VERSION)
        {
            dl_bh_total_drop_pkt_cntr_ipv4++;
        }
        else if(version == IWF_IPV6_VERSION)
        {
            dl_bh_total_drop_pkt_cntr_ipv6++;
        }
    }

}

unsigned int iwfD_DownLinkHook(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*))
{
    int                              iRxPktLen;
    int                              iTxPktLen     =  0;
    struct iphdr                     *ip_header    =  NULL;
    struct ipv6hdr                   *ip6_header   =  NULL;
    struct udphdr                    *udph         =  NULL;
    iwf_gtp_mgmt_pkt_t               *gtpMgmtPkt   =  NULL;
    sicp_data_t                      *sicpData     =  NULL;
    sicp_data_t                      sicpCfgData   =  {0}; /* if not in DL table, then it will be available in config table */
    _IWF_PKT_CLASSIFIER_RESULT       iOperationOnRxPacket = _IWF_PKT_CLASSIFIER_INVALID_ARG;

    /* process the incoming packets */
    if(skb)
    {
        uint16_t vlanid=0,vlanMapid=0;
        char *devPtr = NULL;
        unsigned short frag_off;
        ip_header  =  (struct iphdr *)skb->data;
        ip6_header  =  (struct ipv6hdr *)skb->data;

        if(NULL == ip_header)
        {
            return NF_ACCEPT;
        }

        DBG_PRT_OFF();

        if(skb->dev->name[_IFACE_VLAN_BIT] =='.')
        {
            devPtr = skb->dev->name;
            sscanf(devPtr+_IFACE_VLAN_ID_BIT,"%hd",&vlanid);
            if(vlanid < MAX_VLANID)
                vlanMapid = vlanIds[vlanid];
        }

// k2l: disable pre-route hook 
        if (/*hooknum == NF_INET_PRE_ROUTING*/ 0)
        {
            if ( iwf_ipsec_flag == true)
            {
                // Fill the CB with the ESP packet size , if it's not fragmented
                if((ip_header->protocol == IPPROTO_ESP ||
                            ip6_header->nexthdr == IPPROTO_ESP ) &&
                        ((0 == ip_header->frag_off) & (IP_FRAG_OFF)))
                {
                    skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_LO] = (skb->len & 0xff00) >> 8;
                    skb->cb[MBTAC_METADATA_FRAG_PKT_SIZE_HI] = (skb->len & 0x00ff) ;

                    /* To indicate the ESP packet is not fragmented
                       This cb will be used on pre-fragmentaion case.
                       Fragemnts should not be count if this field is set */
                    skb->cb[SKB_CB_METADATA_FILLED] = 0xff ;
                }
                rx_total_pkt_ipsec_post_route++;
            }
            if (ip_header->version == 4)
            {
                if((0 == strcmp (skb->dev->name, IF_NAME)) || (0 == strncmp (skb->dev->name, IF_VLAN, 4)))
                {
                    if (checkForV4ErrorPackets(skb))
                    {
                        // Increment the drop counter
                        increamentDropCounter(skb,ip_header->version);
                        return NF_ACCEPT;
                    }

                }
                //non BTS terminated and TTL < 2 packets should be counted for drop
                if(!validateTTL(skb))
                {
                    dl_bh_total_drop_pkt_cntr++;
                    return NF_ACCEPT;
                }

                if (ip_header->protocol == IPPROTO_UDP || (ip_header->protocol == IPPROTO_ESP))
                {
                    frag_off  = (ip_header->frag_off & IP_FRAG_OFF); //DF Mask
                    if ((frag_off != 0) || (ip_header->protocol == IPPROTO_ESP))
                    {
                        okfn(skb);
                        rx_total_pkt_ipv4_post_route++;
                        return NF_STOLEN;
                    }
                    udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));
                    if (_IWFD_DL_UDP_PORT_ != ntohs(udph->dest))
                    {
                        rx_total_pkt_ipv4_post_route++;
                        return NF_ACCEPT;
                    }
                    udph = NULL;
                    rx_total_pkt_ipv4_pre_route++;
                    DBG_PRT_ON();
                }
                else
                    return NF_ACCEPT;

            }
            else if (ip_header->version == 6)
            {
                /*struct frag_hdr *fh;
                  if ( NEXTHDR_FRAGMENT  == ip6_header->nexthdr ){
                  fh =  (struct frag_hdr *)(skb->data + sizeof(struct ipv6hdr));
                  }*/
                /*ipv6 multicast check for src ip */
                if( _IPV6_MULTICAST == ip6_header->saddr.s6_addr[0] )
                {
                    /*drop the packet and free the skb*/
                    rx_dl_pkt_drop_multicast_src_ipv6++;
                    if(vlanid && vlanMapid)
                    {
                        ipCounters[vlanMapid-1].vlanId = vlanid;
                        ipCounters[vlanMapid-1].dl_mcast_disc_pkt_cntr_ipv6++;
                        ipCounters[vlanMapid-1].dl_mcast_disc_bytes_cntr_ipv6+= skb->len;
                        ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr_ipv6++;
                        ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr++;

                        ipCounters[vlanMapid-1].dl_bh_mcast_disc_total_pkt_cntr++;
                        ipCounters[vlanMapid-1].dl_bh_mcast_disc_total_bytes_cntr += skb->len;
                    }
                    else
                    {
                        dl_mcast_disc_pkt_cntr_ipv6++;
                        dl_mcast_disc_bytes_cntr_ipv6 = dl_mcast_bytes_cntr_ipv6 + skb->len;
                        dl_bh_total_drop_pkt_cntr_ipv6++;
                        dl_bh_total_drop_pkt_cntr++;

                        dl_bh_mcast_disc_total_pkt_cntr++;
                        dl_bh_mcast_disc_total_bytes_cntr += skb->len;
                    }
                    dev_kfree_skb_any(skb);
                    skb = NULL;
                    return NF_STOLEN;
                }

                /* packet drop cases ipv6 start  for eth0*/
                /* drop the packet when unspecified destination address */

                /* packet drop cases ipv6 end for eth0 */

                /* fix fot the MLD issues*/
                /*1.DUT is processing MLDv2 query when hop limit is not 1
                 **2.PR 141960ESPE02 DUT is processing MLDv2 query which doesn't have Hop-By-Hop Option header */

                if((ip6_header->nexthdr == NEXTHDR_HOP ) || (ip6_header->nexthdr == NEXTHDR_ICMP))
                {
                    struct icmp6hdr             *icmp6header =  NULL;
                    struct ipv6_opt_hdr         *exthdr      =  NULL;
                    bool mld_pkt_drop = false;

                    if (ip6_header->nexthdr == NEXTHDR_ICMP)
                    {
                        icmp6header = (struct icmp6hdr *)(skb->data + sizeof(struct ipv6hdr));
                        if(icmp6header->icmp6_type == ICMPV6_MGM_QUERY )
                        {
                            mld_pkt_drop = true;
                        }
                        else if ((icmp6header->icmp6_type == NDISC_NEIGHBOUR_SOLICITATION) && ((icmp6header->icmp6_code != 0 ) || (ip6_header->hop_limit == ND_HOP_LIMIT ) || (ip6_header->hop_limit == 0 )))
                        {
                            mld_pkt_drop = true;
                        }
                    }
                    else
                    {
                        exthdr = (struct ipv6_opt_hdr *)(skb->data + sizeof(struct ipv6hdr));
                        if((ip6_header->hop_limit != 1 ) && (exthdr->nexthdr == NEXTHDR_ICMP))
                        {
                            mld_pkt_drop = true;
                        }
                    }

                    if(mld_pkt_drop == true)
                    {
                        if(vlanid && vlanMapid)
                        {
                            ipCounters[vlanMapid-1].vlanId = vlanid;
                            ipCounters[vlanMapid-1].dl_mcast_disc_pkt_cntr_ipv6++;
                            ipCounters[vlanMapid-1].dl_mcast_disc_bytes_cntr_ipv6+= skb->len;
                            ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr_ipv6++;
                            ipCounters[vlanMapid-1].dl_bh_total_drop_pkt_cntr++;

                            ipCounters[vlanMapid-1].dl_bh_mcast_disc_total_pkt_cntr++;
                            ipCounters[vlanMapid-1].dl_bh_mcast_disc_total_bytes_cntr += skb->len;
                        }
                        else
                        {
                            dl_mcast_disc_pkt_cntr_ipv6++;
                            dl_mcast_disc_bytes_cntr_ipv6 = dl_mcast_bytes_cntr_ipv6+skb->len;
                            dl_bh_total_drop_pkt_cntr_ipv6++;
                            dl_bh_total_drop_pkt_cntr++;

                            dl_bh_mcast_disc_total_pkt_cntr++;
                            dl_bh_mcast_disc_total_bytes_cntr += skb->len;
                        }
                        /*drop the packet and free the skb*/
                        dev_kfree_skb_any(skb);
                        skb = NULL;
                        return NF_STOLEN;
                    }
                }

                if((0 == strcmp (skb->dev->name, IF_NAME)) || (0 == strncmp (skb->dev->name, IF_VLAN, 4)))
                {
                    if(checkForV6ErrorPackets(skb))
                    {
                        // Increment the drop counter
                        increamentDropCounter(skb,ip_header->version);

                        return NF_ACCEPT;
                    }
                }

                if ( NEXTHDR_UDP == ip6_header->nexthdr )
                {
                    udph =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));
                    if (_IWFD_DL_UDP_PORT_ != ntohs(udph->dest))
                    {
                        rx_total_pkt_ipv6_post_route++;
                        return NF_ACCEPT;
                    }
                    udph = NULL;
                    rx_total_pkt_ipv6_pre_route++;
                }
                else
                {
                    rx_total_pkt_ipv6_post_route++;
                    return NF_ACCEPT;
                }
            }
        } //Pre_routing

        if(ip_header->version == 4)
        {

            /* packet drop cases ipv4 start for eth0 */

            // if br0 interface
            if(skb->dev->name[_IFACE_BIT] == '0')
            {

                if(iwf_ipsec_flag == true)
                {
                    if(!__xfrm_policy_check(NULL,XFRM_POLICY_IN,skb,AF_INET))
                        increment_DownlinkCounters(skb,vlanid,vlanMapid,ip_header->version);
                }
                else
                {
                    increment_DownlinkCounters(skb,vlanid,vlanMapid,ip_header->version);
                }

            } // close if br0

            /* packet drop cases ipv4 end for eth0 */
            DBG_PRT("dlhook: pkt_type %d\n", skb->pkt_type);
            if( PACKET_HOST == skb->pkt_type || PACKET_MULTICAST == skb->pkt_type )
            {
                if(MBTAC_NH_SCTP == ip_header->protocol || IPPROTO_IGMP == ip_header->protocol)
                {
                    mbTacCountCplaneBytesDownlink(skb);
                }
                else if ( IPPROTO_UDP == ip_header->protocol )
                {
                    mbTacCountSplaneBytesDownlink(skb);
                }
            }

#if 1 //Not required as U plane traffic
            /* check if the IP is u-plane IP */
            //////==> Need uplane IP set
            if( /*(get_ip_se_lte_ip(0) == ntohl(ip_header->daddr) || get_ip_se_lte_ip(1) == ntohl(ip_header->daddr))&&*/ (0 != ip_header->daddr)) //RANS
            {

                rx_total_pkt_ipv4++;

                /* do we have required payload to work on */
                if(ntohs(ip_header->tot_len) >= (sizeof(struct iphdr) + sizeof(struct udphdr) +  sizeof(iwf_gtp_hdr_t)))
                {
                    /* if its a udp packet */
                    if((IPPROTO_UDP == ip_header->protocol))
                    {
                        rx_udp_total_pkt_ipv4++;

                        udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

                        /* Make check-sum of all udp packet to 0.

NOTE :
As per requirement : WBTS_TRS_3179 UDP checksum for UDP-over-IPv4
The BTS shall not use UDP checksum.
The BTS shall ignore the checksum of received UDP datagrams.
The BTS shall set the checksum for transmitted UDP datagrams to 0 to mark that the checksum
is not used.

*/
                        udph->check = 0;
                    }
                }
            }
#endif
        }
        else if (ip_header->version == 6)
        {
            /* packet drop cases ipv6 start  for eth0*/
            /* drop the packet when unspecified destination address */

            /* packet drop cases ipv6 end for eth0 */


            if(skb->dev->name[_IFACE_BIT] == '0')
            {
                if (iwf_ipsec_flag == true)
                {
                    if(!__xfrm_policy_check(NULL,XFRM_POLICY_IN,skb,AF_INET6))
                        increment_DownlinkCounters(skb,vlanid,vlanMapid,ip_header->version);
                }
                else
                    increment_DownlinkCounters(skb,vlanid,vlanMapid,ip_header->version);
            }

            if( PACKET_HOST == skb->pkt_type )
            {
                if(MBTAC_NH_SCTP == ip6_header->nexthdr)
                {
                    mbTacCountCplaneBytesDownlink(skb);
                }
                else if ( NEXTHDR_UDP == ip6_header->nexthdr )
                {
                    mbTacCountSplaneBytesDownlink(skb);
                }
            }


            /* check if the IP is u-plane IP */
#if 1 //Not required as U plane traffic
            if (/*is_ip_U_plane_ipv6(&ip6_header->daddr) == true*/ 1)
            {
                rx_total_pkt_ipv6++;
                /* do we have required payload to work on */
                if (ntohs(ip6_header->payload_len) >= (sizeof(struct udphdr) +  sizeof(iwf_gtp_hdr_t)))
                {
                    if ( NEXTHDR_UDP == ip6_header->nexthdr )
                    {
                        rx_udp_total_pkt_ipv6++;
                        udph =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

                        if (udp6_csum_init(skb, udph, IPPROTO_UDP))
                        {
                            //If checksum is calculated is not Zero
                            //then drop the packet in IWF
                            rx_udp_checksum_err_drop_ipv6 ++;

                            if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                            {
                                if(0 != printk_ratelimit())
                                {
                                    printk(" DL Pkt size skb:%u, iphl:%u, skb csum:0x%x, udp->csum:0x%x, ip_summed:%u\n\n",
                                            skb->len, ntohs(ip6_header->payload_len),
                                            skb->csum, udph->check, skb->ip_summed);
                                }

                            }
                            /* free the skb */
                            dev_kfree_skb_any(skb);
                            skb = NULL;

                            return NF_STOLEN;
                        }

                    }
                }
            }
#endif
        }
        /*If its a udp packet &  u-plane DL packet */
        if(udph &&(_IWFD_DL_UDP_PORT_ == ntohs(udph->dest)))
        {
            rx_gtpu_total_pkt++;

            /* check for IPSec policy, if no rule drop the packet */
            if (ip_header->version == 4)
            {
                if (!__xfrm_policy_check(NULL,XFRM_POLICY_IN,skb,AF_INET))
                {
                    if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                    {
                        if(0 != printk_ratelimit())
                        {
                            printk(KERN_INFO _IWF_MODULE_NAME": ERR!! %s NO IPSEC RULES... DL Packet dropped !!!! \n",__func__);
                        }
                    }

                    /* increment drop counter */
                    rx_dl_gtpu_xfrm_drop_no_policy_ipv4++;

                    /* free the skb */
                    dev_kfree_skb_any(skb);
                    skb = NULL;

                    /* no ipsec rules, let the packet be handled by the IP stack */
                    return NF_STOLEN;
                }
            }
            else if (ip_header->version == 6)
            {
                if (!__xfrm_policy_check(NULL,XFRM_POLICY_IN,skb,AF_INET6))
                {
                    if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                    {
                        if(0 != printk_ratelimit())
                        {
                            printk(KERN_INFO _IWF_MODULE_NAME": ERR!! %s NO IPSEC RULES... DL Packet dropped !!!! \n",__func__);
                        }
                    }

                    /* increment drop counter */
                    rx_dl_gtpu_xfrm_drop_no_policy_ipv6++;

                    /* free the skb */
                    dev_kfree_skb_any(skb);
                    skb = NULL;

                    /* no ipsec rules, let the packet be handled by the IP stack */
                    return NF_STOLEN;
                }
            }

            iRxPktLen = udph->len - sizeof(struct udphdr);

            /* retrieve gtpu data */
            gtpMgmtPkt = (iwf_gtp_mgmt_pkt_t *)((char *)udph + sizeof(struct udphdr));
            /* validate GTPU header */
            if((gtpMgmtPkt->gtpuHdr.flags.version == 1) && (gtpMgmtPkt->gtpuHdr.flags.proto_type == 1))
            {
                unsigned char  *gtpStart     =  (unsigned char *) ( (char *)udph + sizeof(struct udphdr));

                /* lookup downlink table for the teid */
                sicpData = get_teid_sicp_connection(ntohl(gtpMgmtPkt->gtpuHdr.teid));

                if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(" %s teid : 0x%x, %d\n",__func__, ntohl(gtpMgmtPkt->gtpuHdr.teid),ntohl(gtpMgmtPkt->gtpuHdr.teid));
                        printk(" %s sicpData %p\n", __func__ , sicpData);
                    }
                }

                /* connection available */
                if(NULL != sicpData)
                {
                    iOperationOnRxPacket = classifyOperationOnDLPkt(gtpStart,iRxPktLen,_IWF_AVAILABLE_IN_DL_TABLE,NULL, true, skb);
                }
                else
                {
                    iOperationOnRxPacket = classifyOperationOnDLPkt(gtpStart,iRxPktLen,_IWF_NOT_AVAILABLE_IN_DL_TABLE,&sicpCfgData, true, skb);
                    sicpData = &sicpCfgData;
                }

                if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(KERN_INFO _IWF_MODULE_NAME": classifyOperationOnDLPkt : %d\n",iOperationOnRxPacket);
                    }
                }

                switch(iOperationOnRxPacket)
                {
                    /* this is slow path, so better deal in normal flow */
                    case _IWF_SEND_ECHO_RESPONSE_TO_SGW:

                        /*  if Teid 0 & Msgtype is 0x01 then its Echo request from SGw, so send back the response */
                        iTxPktLen = iwfSendEchoResponseToSgw(skb);

                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME": _IWF_SEND_ECHO_RESPONSE_TO_SGW pkt sent size : %d",iTxPktLen);
                            }
                        }
                        return NF_STOLEN;

                    case _IWF_SEND_SUPPORTED_EXT_HDR_TO_SGW:

                        iTxPktLen = iwfSendSuppExthdrToSgw(skb);

                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME": _IWF_SEND_SUPPORTED_EXT_HDR_TO_SGW pkt sent size : %d",iTxPktLen);
                            }
                        }
                        return NF_STOLEN;

                    case _IWF_SEND_ERR_IND_TO_SGW:

                        iTxPktLen = iwfSendErrorIndMsgToSgw(skb);
                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME": _IWF_SEND_ERR_IND_TO_SGW pkt sent size : %d",iTxPktLen);
                            }
                        }
                        return NF_STOLEN;

                    case _IWF_FWD_ECHO_RESPONSE_TO_TUPU:
                    case _IWF_FWD_ERR_IND_TO_TUPU:
                    case _IWF_FWD_PKT_TO_TUPU:

                        rx_dl_pkt_to_tupu++;
                        mbTacCountUplaneBytesDownlink(skb);
                        iTxPktLen = iwfSendSicpMessage(sicpData, skb);

                        if(IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME": Message sent to Tupu pkt sent size : %d",iTxPktLen);
                            }
                        }

                        return NF_STOLEN;

                    case _IWF_DROP_PKT:
                        rx_dl_gtpu_pkt_drop_cntr++;
                        if(skb)
                        {
                            dev_kfree_skb_any(skb);
                            skb = NULL;
                        }

                        /* packets should drop in case of invalid GTPu version and PT types */
                        return NF_STOLEN;

                    case _IWF_PKT_CLASSIFIER_INVALID_ARG:
                    default:
                        rx_dl_gtpu_pkt_drop_cntr++;
                        if(IWF_LOG_LEVEL_DL_ERR_CRT_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                /* Drop the pkt */
                                printk(KERN_INFO _IWF_MODULE_NAME" %s : PKT DROPPED .. NOT Classified \n",__func__);
                            }
                        }
                        break;
                }
            }//if((gtpMgmtPkt->gtpuHdr.flags.version == 1)
            else
            {
                rx_dl_gtpu_pkt_drop_invalid_gtp_hdr++;
                if(skb)
                {
                    dev_kfree_skb_any(skb);
                    skb = NULL;
                }

                /* packets should drop in case of invalid GTPu version and PT types */
                return NF_STOLEN;
            }
        }//if (udph && (_IWFD_DL_UDP_PORT_ == ntohs(udph->dest)))
    }/*  if(skb) */

    return NF_ACCEPT;
}


static bool isBTSIp(struct sk_buff *skb)
{
    int i;
    struct iphdr *ip_header = (struct iphdr *)skb->data;
    //check U plane IP's
    for ( i = 0 ; i < LTE_MAX_IP_SRC_IDX ; i++)
    {
        if(( sicp_lte_ip[i].uip_src.version == IPv4 ) &&
                (sicp_lte_ip[i].uip_src.addr_u.ipv4Address.ipa_32 == ntohl(ip_header->daddr)))
        {
            return true;
        }
    }
    //check S Plane IP
    if ( mbtacTmbCounterData.splaneData.ip.ip4_addr32 == ntohl(ip_header->daddr))
    {
        return true;
    }

    //check M Plane
    if(  mplaneIp.ip4_addr32 == ntohl(ip_header->daddr))
    {
        return true;
    }

    //check C Plane
    for( i = 0 ; i < MAX_NO_CPLANE_IP_ADDR ; i++)
    {
        if((mbtacTmbCounterData.cplaneData[i].ip.version == IPv4) &&
                (mbtacTmbCounterData.cplaneData[i].ip.ip4_addr32 == ntohl(ip_header->daddr)))
        {
            return true;
        }
    }
    return false;
}

static bool validateTTL(struct sk_buff *skb)
{
    struct iphdr *ip_header  =  (struct iphdr *)skb->data;
    if(!isBTSIp(skb))
    {
        // TTL < 2 then increment the drop count and send it for upper layer where it shoudl get droppped.
        if(ip_header->ttl < 2)
        {
            return false;
        }
    }

    return true;
}

bool checkForValidProtocol(struct iphdr *ip_header)
{
    switch (ip_header->protocol)
    {
        case IPPROTO_ICMP:
        case IPPROTO_IGMP:
        case IPPROTO_TCP:
        case IPPROTO_UDP:
        case IPPROTO_ESP:
        case IPPROTO_OSPF:
        case IPPROTO_SCTP:
            {
                return true;
            }
    }
    return false;
}

bool isValidIpv4HeaderChecksum(struct sk_buff *skb)
{
    bool errorInPacket = true;
    uint16_t prev_len = skb->len;
    skb->len = sizeof(struct iphdr);
    if (__skb_checksum_complete(skb))
    {
        rx_dl_checksum_error_ipv4++;
        errorInPacket = false;
    }
    skb->len = prev_len;
    return errorInPacket;
}


bool isValidIcmpv4Checksum(struct sk_buff *skb)
{
    bool errorInPacket = true;
    skb_pull(skb, sizeof(struct iphdr));
    if (__skb_checksum_complete(skb))
    {
        rx_dl_icmp_checksum_error_ipv4++;
        errorInPacket = false;
    }
    skb_push(skb, sizeof(struct iphdr));
    return errorInPacket;
}

bool isValidIcmpv6Checksum(struct sk_buff *skb)
{
    const struct in6_addr *saddr, *daddr;
    bool errorInPacket = true;
    saddr = &ipv6_hdr(skb)->saddr;
    daddr = &ipv6_hdr(skb)->daddr;
    skb_pull(skb, sizeof(struct ipv6hdr));
    skb->csum = ~csum_unfold(csum_ipv6_magic(saddr, daddr, skb->len, IPPROTO_ICMPV6, 0));
    if (__skb_checksum_complete(skb))
    {
        rx_dl_icmp_checksum_error_ipv6++;
        errorInPacket = false;
    }
    skb_push(skb, sizeof(struct ipv6hdr));
    return errorInPacket;
}

bool isValidUdp6Checksum(struct sk_buff *skb)
{
    bool errorInPacket = true;
    uint16_t check = 0;
    uint16_t csum = 0;
    struct udphdr *udph =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));
    check = udph->check;
    udph->check = 0;
    csum = csum_ipv6_magic(&ipv6_hdr(skb)->saddr, &ipv6_hdr(skb)->daddr, ntohs(udph->len), IPPROTO_UDP, csum_partial(udph, ntohs(udph->len), 0));
    if(csum != check)
    {
        rx_udp_checksum_err_drop_ipv6++;
        errorInPacket = false;
    }
    udph->check = check;
    return errorInPacket;
}

bool checkForV4ErrorPackets (struct sk_buff *skb)
{
    struct iphdr *ip_header = NULL;
    bool errorInPacket = false;
    ip_header = (struct iphdr *)skb->data;

    if(NULL == ip_header)
    {
        return NF_ACCEPT;
    }

    // Reason : Unknown protocol. BTS will support only the termination of the following next level protocols,
    // packets with other protocols will be dropped.

    //TODO IPVersion mismatch

    if(!checkForValidProtocol (ip_header))
    {
        errorInPacket = true;
    }

    // Reason : IPv4 header options present.
    else if (ip_header->ihl > IP_OPTION_LEN)
    {
        errorInPacket = true;
    }

    //IPv4 Checksum
    else if(!isValidIpv4HeaderChecksum(skb))
    {
        errorInPacket = true;
    }
    //ICMPv4 Checksum
    else if(ip_header->protocol == IPPROTO_ICMP)
    {
        errorInPacket = (!isValidIcmpv4Checksum(skb));
    }
    return errorInPacket;
}

int checkHdrType(uint16_t hdr)
{
    int retVal = _IWF_ERR_HDR;
    switch (hdr)
    {
        case NEXTHDR_HOP:
        case NEXTHDR_DEST:
        case NEXTHDR_ROUTING:
        case NEXTHDR_FRAGMENT:
        case NEXTHDR_AUTH:
        case NEXTHDR_ESP:
        case NEXTHDR_MOBILITY:
            {
                retVal = _IWF_EXT_HDR;
                break;
            }
        case NEXTHDR_ICMP:
        case NEXTHDR_IGMP:
        case NEXTHDR_TCP:
        case NEXTHDR_UDP:
        case NEXTHDR_OSPF:
        case NEXTHDR_SCTP:
            {
                retVal = _IWF_UL_HDR;
                break;
            }
        case NEXTHDR_NONE:
            {
                retVal = _IWF_NOHDR;
                break;
            }
    }
    return retVal;
}

bool isValidProtoAndExtnHdr(struct sk_buff *skb)
{
    struct ipv6hdr *ip6_header = NULL;
    int hdrType = 0;

    ip6_header = (struct ipv6hdr *)skb->data;
    hdrType = checkHdrType(ip6_header->nexthdr);

    if(_IWF_EXT_HDR == hdrType)
    {
        // Check Next Upper Layer header
        uint32_t start = skb_network_offset(skb) + sizeof(struct ipv6hdr);
        struct ipv6_opt_hdr exHdr,*hp;
        hp = skb_header_pointer(skb, start, sizeof(exHdr), &exHdr);

        if(!hp)
        {
            // No packet drop
            return false;
        }
        hdrType = checkHdrType(hp->nexthdr);

        if(_IWF_EXT_HDR == hdrType)
        {
            // Drop the packet
            return true;
        }
        else if((NEXTHDR_HOP == ip6_header->nexthdr) &&
                (NEXTHDR_ICMP != hp->nexthdr))
        {
            // Drop the packet
            return true;
        }
    }
    else if(_IWF_ERR_HDR == hdrType)
    {
        // Drop the packet
        return true;
    }
    return false;
}

bool checkForV6ErrorPackets (struct sk_buff *skb)
{
    struct ipv6hdr *ip6_header   =  NULL;
    struct udphdr *udph         =  NULL;
    bool errorInPacket = false;
    struct in6_addr localaddr[ADDR_LEN];

    ip6_header = (struct ipv6hdr *)skb->data;
    udph = (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

    memset (localaddr, 0, sizeof (localaddr));

    // TODO IPVersion mismatch

    // Reason : Unspecified address (::/128) as destination.
    if(memcmp(ip6_header->daddr.s6_addr32, localaddr, ADDR_LEN) == 0)
    {
        errorInPacket = true;
    }

    // Reason : Multicast address is used as source.
    else if( ipv6_addr_is_multicast(&(ip6_header->saddr)))
    {
        errorInPacket = true;
    }

    // Reason : Multicast as destination and scope=0.
    else if ((ipv6_addr_is_multicast(&(ip6_header->daddr))) &&
            ((((ip6_header->daddr.s6_addr32[0]) & 0x0F00) >> 8) == 0))
    {
        errorInPacket = true;
    }

    // Reason : Unknown protocol/Unsupported Extension Header
    else if (isValidProtoAndExtnHdr (skb))
    {
        errorInPacket = true;
    }

    //ICMP checksum failure
    else if(ip6_header->nexthdr == NEXTHDR_ICMP)
    {
        errorInPacket = (!isValidIcmpv6Checksum(skb));
    }
    //UDP checksum failure
    else if(ip6_header->nexthdr == NEXTHDR_UDP)
    {
        errorInPacket = (!isValidUdp6Checksum(skb));
    }
    return errorInPacket;
}
#endif
/* ------------------------------------------------------------------------------ */
#ifdef __USE_NET_FILTER__

/* To do the Pre-framentation of the packte */
void pre_frag_hook(struct sk_buff *skb)
{
    struct dst_entry *dst;
    struct xfrm_state *x;
    const struct iphdr *iph;
    unsigned int pmtu, ihl;


    dst = skb_dst(skb);
    x = dst->xfrm;

    if (!x) {
        //printk("pre_frag_hook: XFRM data not present !");
        return ;
    }

    pmtu = dst_mtu(dst);
    //If the SKB MTU is less than the Path MTU give the packe to IP-Stack
    if (skb->len <= pmtu) {
        return ;
    }
    skb->local_df = 0;

    iph = ip_hdr(skb);
    // Pre-Fragmentation support is only for UDP packets
    if (iph->protocol != 17) {
        //printk("pre_frag_hook:Pre-Fragmentation supported only for UDP !");
        return ;
    }

    ihl = iph->ihl * 4;
    skb->ip_summed = CHECKSUM_PARTIAL;
    skb->csum = 0;
    skb_shinfo(skb)->gso_size = ((pmtu - ihl) & ~7);
    skb_shinfo(skb)->gso_type = SKB_GSO_UDP;
    return;
}

unsigned int iwfD_UpLinkHook(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*))
{
    struct iphdr                     *ip_header    =  NULL;
    struct udphdr                    *udph         =  NULL;
    iwf_sic_message_hdr_t            *sicMsgHdr    =  NULL;
    iwf_gtp_hdr_t                    *gtpuHdr      =  NULL;
    iwf_sicad_t                      sicad;
#ifdef __IWF_ENABLE_MSG_RTR
    int msgRtrRetval = -1;
#endif

    sicad.sicad_32 = 1;

    /* if not NULL */
    if(skb)
    {

        ip_header  =  (struct iphdr *)skb->data;

        *((unsigned char*)(&skb->cb[MBTAC_PKT_ID_ADDR])) = 0xff; //non mbtac traffic
        /* if not fragment packet & if its a udp packet */
        //if( (!ip_is_fragment(ip_header)) && (IPPROTO_UDP == ip_header->protocol) )
        if( (IPPROTO_UDP == ip_header->protocol) )
        {

            udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

            /* if uplink packet. In DSP sim mode the UL packets comes on 30,000 port */
            if(_IWFD_UL_UDP_PORT_ == ntohs(udph->dest))
            {
                /* move the skb->data to point from sicp header as the IP header
                   is not used for Tuppu and FTM communication */
                skb_pull(skb,sizeof(struct iphdr) + sizeof(struct udphdr));

                /* retrieve the sic and gtpu header */
                sicMsgHdr = (iwf_sic_message_hdr_t *)(skb->data);
                gtpuHdr   = (iwf_gtp_hdr_t *)(skb->data + sizeof(iwf_sic_message_hdr_t));

                sicad.cpid = (sicMsgHdr->rx_cpid);
                sicad.nid  = (sicMsgHdr->rx_nid);


                sicad.sicad_32 = ntohl(sicad.sicad_32);
                sicMsgHdr->rx_cpid = sicad.cpid;
                sicMsgHdr->rx_nid  = sicad.nid;

                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(" %s RX NID         =%x\n",__func__,(sicad.nid));
                        printk(" %s RX CPID        =%x\n",__func__,(sicad.cpid));
                    }
                }

                sicad.cpid = (sicMsgHdr->tx_cpid);
                sicad.nid  = (sicMsgHdr->tx_nid);

#ifndef _IWF_EN
                sicad.sicad_32 = ntohl(sicad.sicad_32);
                sicMsgHdr->tx_cpid = sicad.cpid;
                /* swap the tx nid with rx nid */
                sicMsgHdr->rx_nid  = sicad.nid;
#ifdef __IWF_ONLY_ONE_CALL_BK
                sicMsgHdr->tx_nid =  ftm_nid;
#endif
#endif

                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(" %s TX NID         =%x\n",__func__,(sicad.nid));
                        printk(" %s TX CPID        =%x\n",__func__,(sicad.cpid));
                    }
                }

#ifdef __IWF_ENABLE_MSG_RTR
                /* call DMA loopback function */

#ifdef __IWF_ONLY_ONE_CALL_BK
                sicMsgHdr->tx_nid = ftm_nid;
#endif

                if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk(KERN_INFO _IWF_MODULE_NAME"%s: Calling MsgRtrARMSend_loopback \n",__func__);
                    }
                }

                if(true == iwf_sim_flag)
                {
                    if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                    {
                        if(0 != printk_ratelimit())
                        {
                            printk(KERN_INFO _IWF_MODULE_NAME"%s: ##### IN SIMULATION MODE : %d\n",__func__,iwf_sim_flag);
                        }
                    }
                    msgRtrRetval = MsgRtrARMSend_loopback(skb);
                    if(msgRtrRetval)
                    {
                        if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMSend_loopback failed : %d\n",__func__, msgRtrRetval);
                            }
                        }
                        dev_kfree_skb_any(skb);
                        skb = NULL;
                    }
                    else
                    {
                        if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME"%s: SUCCESS!! MsgRtrARMSend_loopback success : %d\n",__func__, msgRtrRetval);
                            }
                        }
                    }
                }
                else
                {
                    if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
                    {
                        if(0 != printk_ratelimit())
                        {
                            printk(KERN_INFO _IWF_MODULE_NAME"%s: ##### IN DSP MODE : %d\n",__func__,iwf_sim_flag);
                        }
                    }
                    msgRtrRetval = MsgRtrARMSend(skb);
                    if(msgRtrRetval)
                    {
                        if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMSend failed : %d\n",__func__, msgRtrRetval);
                            }
                        }
                        dev_kfree_skb_any(skb);
                        skb = NULL;
                    }
                    else
                    {
                        if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
                        {
                            if(0 != printk_ratelimit())
                            {
                                printk(KERN_INFO _IWF_MODULE_NAME"%s: SUCCESS!! MsgRtrARMSend success : %d\n",__func__, msgRtrRetval);
                            }
                        }
                    }
                }//else, if(iwf_sim_flag)

#else
                sendPktToSgw(skb);
#endif
                return NF_STOLEN;

            }//if(_IWFD_UL_UDP_PORT_ == ntohs(udph->dest))

        }//if( (!ip_is_fragment(ip_header)) && (IPPROTO_UDP == ip_header->protocol) )

    }//if(skb)

    return NF_ACCEPT;
}

unsigned int mbTac_UpLinkHook(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*)){

    struct iphdr *ip_header;
    uint8_t t_dscp=0;

    uint16_t vlanid=0,vlanMapid=0;
    char *devPtr=NULL;
    struct net_device *dev = NULL;
    struct inet6_dev *idev = NULL;
    struct udphdr    *udph = NULL;
    unsigned int total_pkt_size = 0;
    uint16_t no_of_frags=1;
    bool ltm_gtp_traffic_flag = 0;


    if (skb !=NULL)
    {
        ip_header =(struct iphdr*)skb_network_header(skb);
        if (ip_header != NULL)
        {
            if (ip_header->version == IWF_IPV4_VERSION)
            {
                if(IPPROTO_HOPTOHOP != ip_header->protocol)
                {
                    dev = skb_dst(skb)->dev;
                }
                if(MBTAC_NH_TCP == ip_header->protocol)
                {
                    if(ip_header->saddr == htonl(mp_IpAddr))
                    {
                        skb->priority = mp_dscp ;

                        t_dscp = ((mp_dscp << 2) | ((ip_header->tos) & ECN ));  //ECN bits are set based on ipheader->tos
                        ip_header->tos = t_dscp;

                        /*recompute ip header checksum*/
                        ip_send_check(ip_header);
                    }
                }

                if (MBTAC_NH_SCTP == ip_header->protocol)
                {
                    /*set skb->prio */
                    skb->priority = cp_dscp;

                    t_dscp = ((cp_dscp << 2) | ((ip_header->tos) & ECN ));  //ECN bits are set based on ipheader->tos
                    ip_header->tos = t_dscp;

                    /*recompute ip header checksum*/
                    ip_send_check(ip_header);

                }
                else if(IPPROTO_UDP == ip_header->protocol)
                {
                    udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));
                    /* if uplink packet. In DSP sim mode the UL packets comes on 30,000 port */
                    t_dscp = ip_header->tos >> 2;
                    /*set skb prio */
                    skb->priority = t_dscp ;
                }
                total_pkt_size += sizeof(struct iphdr);
            }
            else if(ip_header->version == IWF_IPV6_VERSION)
            {
                struct ipv6hdr     *ip6_header   =  NULL;
                ip6_header =  (struct ipv6hdr *)skb->data;

                if(ip6_header != NULL)
                {
                    if((ip6_header->nexthdr == NEXTHDR_ICMP) || (ip6_header->nexthdr == NEXTHDR_HOP) )
                    {
                        struct icmp6hdr *icmp6header =  NULL;
                        if((ip6_header->nexthdr == NEXTHDR_HOP))
                        {
                            struct ipv6_opt_hdr *exthdr;
                            exthdr = (struct ipv6_opt_hdr *)(skb->data + sizeof(struct ipv6hdr));
                            if(exthdr->nexthdr == NEXTHDR_ICMP)
                            {
                                icmp6header = (struct icmp6hdr *)(skb->data + ((sizeof(struct ipv6hdr) )+ (HOP_BY_HOP_HDR_LENGTH) + exthdr->hdrlen));
                            }
                        }
                        else
                        {
                            icmp6header = (struct icmp6hdr *)(skb->data + sizeof(struct ipv6hdr));
                        }
                        if ( NULL != icmp6header)
                        {
                            if ((icmp6header->icmp6_type == ICMPV6_MGM_QUERY) || (icmp6header->icmp6_type == ICMPV6_MGM_REPORT) ||(icmp6header->icmp6_type == ICMPV6_MGM_REDUCTION ) ||
                                    (icmp6header->icmp6_type == NDISC_ROUTER_SOLICITATION) || (icmp6header->icmp6_type == NDISC_ROUTER_ADVERTISEMENT)  ||
                                    (icmp6header->icmp6_type == NDISC_NEIGHBOUR_SOLICITATION) ||(icmp6header->icmp6_type == NDISC_NEIGHBOUR_ADVERTISEMENT) ||
                                    (icmp6header->icmp6_type == NDISC_REDIRECT) ||(icmp6header->icmp6_type == ICMPV6_MLD2_REPORT))
                            {
                                if(skb->dev->name[_IFACE_VLAN_BIT] =='.')
                                {
                                    devPtr = skb->dev->name;
                                    sscanf(devPtr+_IFACE_VLAN_ID_BIT,"%hd",&vlanid);
                                    if(vlanid < MAX_VLANID)
                                        vlanMapid = vlanIds[vlanid];
                                }

                                ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                if(vlanid && vlanMapid)
                                {
                                    ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                    ipCounters[vlanMapid-1].vlanId = vlanid;
                                    ipCounters[vlanMapid-1].ul_mcast_pkt_cntr_ipv6++;
                                    ipCounters[vlanMapid-1].ul_mcast_bytes_cntr_ipv6+= skb->len;

                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr++;
                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr += skb->len;
                                }
                                else
                                {
                                    ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                    ul_mcast_pkt_cntr_ipv6++;
                                    ul_mcast_bytes_cntr_ipv6 = ul_mcast_bytes_cntr_ipv6+skb->len;

                                    ul_bh_mcast_total_pkt_cntr++;
                                    ul_bh_mcast_total_bytes_cntr = ul_bh_mcast_total_bytes_cntr + skb->len;
                                }
                            }
                            if ((icmp6header->icmp6_type == ICMPV6_MGM_QUERY) || (icmp6header->icmp6_type == ICMPV6_MGM_REPORT) ||
                                (icmp6header->icmp6_type == ICMPV6_MGM_REDUCTION) ||(icmp6header->icmp6_type == ICMPV6_MLD2_REPORT))
                            {
                                if(vlanid && vlanMapid)
                                {
                                    ipCounters[vlanMapid-1].vlanId = vlanid;
                                    ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv6++;
                                    ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv6+= skb->len;
                                    ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr++;
                                    ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr+= skb->len;
                                }
                                else
                                {
                                    ul_bh_total_pkt_cntr_ipv6++;
                                    ul_bh_total_bytes_cntr_ipv6 = ul_bh_total_bytes_cntr_ipv6 +skb->len;
                                    ul_bh_total_pkt_cntr++;
                                    ul_bh_total_bytes_cntr+= skb->len;
                                }
                            }

                        }
                    }
                    if((IPPROTO_HOPTOHOP != ip6_header->nexthdr) && (skb->pkt_type != PACKET_MULTICAST))
                    {
                        idev = ip6_dst_idev(skb_dst(skb));
                        if(idev)
                            dev = idev->dev;
                    }
                    if(MBTAC_NH_SCTP == ip6_header->nexthdr)
                    {
                        /*set skb prio */
                        skb->priority = cp_dscp;
                        //Set 4bits of DSCP
                        ip6_header->priority = cp_dscp >> 2;
                        //Set 2 bits DSCP and retain remaining 6 bits in flow_lbl[0]
                        ip6_header->flow_lbl[0] = ((cp_dscp << 6) | (ip6_header->flow_lbl[0] & IPv6_ECN));
                    }
                    else if(IPPROTO_UDP == ip6_header->nexthdr)
                    {
                        udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));
                        /* if uplink packet. In DSP sim mode the UL packets comes on 30,000 port */
                        t_dscp = ((ip6_header->flow_lbl[0] >>6) | (ip6_header->priority << 2));
                        /*set skb prio */
                        skb->priority = t_dscp ;
                    }
                }
                total_pkt_size += sizeof(struct ipv6hdr);
            }
            // if br0 interface
            if(dev && dev->name[_IFACE_BIT] == '0')
            {
                struct dst_entry *dst;
                unsigned int pmtu=0;

                dst = skb_dst(skb);
                pmtu = dst_mtu(dst);
                if(dev->name[_IFACE_VLAN_BIT] =='.')
                {
                    devPtr = dev->name;
                    sscanf(devPtr+_IFACE_VLAN_ID_BIT,"%hd",&vlanid);
                    if(vlanid < MAX_VLANID)
                        vlanMapid = vlanIds[vlanid];
                }
                if(vlanid && vlanMapid)
                {
                    total_pkt_size += (skb->len + VLAN_HDR_LEN);
                    if(total_pkt_size > pmtu)
                        no_of_frags = total_pkt_size/pmtu + ((total_pkt_size%pmtu)?1:0);
                    ipCounters[vlanMapid-1].vlanId = vlanid;
                    if(iwf_ipsec_flag == true)
                    {
                        if(skb->sp != NULL)
                        {
                            ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr += no_of_frags;
                            ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr += skb->len;

                            // V4 & V6 specific counters
                            //
                            if (ip_header->version == IWF_IPV4_VERSION)
                            {
                                ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv4 += no_of_frags;
                                ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv4 += skb->len;
                                if (MBTAC_NH_IGMP == ip_header->protocol)
                                {
                                    ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr_ipv4++;
                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr++;
                                    ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                                }
                            }
                            else if(ip_header->version == IWF_IPV6_VERSION)
                            {
                                ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv6 += no_of_frags;
                                ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv6 += skb->len;
                            }

                        }
                        else
                        {
                            struct xfrm_state *x = dst->xfrm;
                            if(!x)
                            {
                                ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr += no_of_frags;
                                ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr += skb->len;

                                // V4 & V6 specific counters

                                if (ip_header->version == IWF_IPV4_VERSION)
                                {
                                    ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv4 += no_of_frags;
                                    ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv4 += skb->len;
                                    if (MBTAC_NH_IGMP == ip_header->protocol)
                                    {
                                        ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                        ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr_ipv4++;
                                        ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                        ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr++;
                                        ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                                    }
                                }
                                else if (ip_header->version == IWF_IPV6_VERSION)
                                {
                                    ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv6 += no_of_frags;
                                    ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv6 += skb->len;
                                }
                            }
                        }
                    }
                    else
                    {
                        ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr += no_of_frags;
                        ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr += skb->len;

                        if (ip_header->version == IWF_IPV4_VERSION)
                        {
                            ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv4 += no_of_frags;
                            ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv4 += skb->len;
                            if (MBTAC_NH_IGMP == ip_header->protocol)
                            {
                                ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr_ipv4++;
                                ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                ipCounters[vlanMapid-1].ul_bh_mcast_total_pkt_cntr++;
                                ipCounters[vlanMapid-1].ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                            }
                        }
                        else if (ip_header->version == IWF_IPV6_VERSION)
                        {
                            ipCounters[vlanMapid-1].ul_bh_total_pkt_cntr_ipv6 += no_of_frags;
                            ipCounters[vlanMapid-1].ul_bh_total_bytes_cntr_ipv6 += skb->len;
                        }
                    }
                }
                else
                {
                    total_pkt_size += (skb->len + ETH_HDR_LEN);
                    if(total_pkt_size > pmtu)
                        no_of_frags = total_pkt_size/pmtu + ((total_pkt_size%pmtu)?1:0);
                    if(iwf_ipsec_flag == true)
                    {
                        if(skb->sp != NULL)
                        {
                            /* Increment the  total transmitted packet count for eth0 */
                            ul_bh_total_pkt_cntr   += no_of_frags;
                            /* Increment the  total transmitted packet count for eth0 */
                            ul_bh_total_bytes_cntr = ul_bh_total_bytes_cntr + skb->len;

                            // V4 & V6 specific counters

                            if (ip_header->version == IWF_IPV4_VERSION)
                            {
                                ul_bh_total_pkt_cntr_ipv4   += no_of_frags;
                                ul_bh_total_bytes_cntr_ipv4 = ul_bh_total_bytes_cntr_ipv4 + skb->len;
                                if (MBTAC_NH_IGMP == ip_header->protocol)
                                {
                                    ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                    ul_bh_mcast_total_pkt_cntr_ipv4++;
                                    ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                    ul_bh_mcast_total_pkt_cntr++;
                                    ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                                }
                            }
                            else if (ip_header->version == IWF_IPV6_VERSION)
                            {
                                ul_bh_total_pkt_cntr_ipv6   += no_of_frags;
                                ul_bh_total_bytes_cntr_ipv6 = ul_bh_total_bytes_cntr_ipv6 + skb->len;
                            }
                        }
                        else
                        {
                            struct xfrm_state *x = dst->xfrm;
                            if(!x)
                            {
                                /* Increment the  total transmitted packet count for eth0 */
                                ul_bh_total_pkt_cntr   += no_of_frags;
                                /* Increment the  total transmitted packet count for eth0 */
                                ul_bh_total_bytes_cntr = ul_bh_total_bytes_cntr + skb->len;
                            }

                            // V4 & V6 specific counters
                            if (ip_header->version == IWF_IPV4_VERSION)
                            {
                                ul_bh_total_pkt_cntr_ipv4   += no_of_frags;
                                ul_bh_total_bytes_cntr_ipv4 = ul_bh_total_bytes_cntr_ipv4 + skb->len;
                                if (MBTAC_NH_IGMP == ip_header->protocol)
                                {
                                    ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                    ul_bh_mcast_total_pkt_cntr_ipv4++;
                                    ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                    ul_bh_mcast_total_pkt_cntr++;
                                    ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                                }
                            }
                            else if (ip_header->version == IWF_IPV6_VERSION)
                            {
                                ul_bh_total_pkt_cntr_ipv6   += no_of_frags;
                                ul_bh_total_bytes_cntr_ipv6 = ul_bh_total_bytes_cntr_ipv6 + skb->len;
                            }
                        }
                    }
                    else
                    {
                        /* Increment the  total transmitted packet count for eth0 */
                        ul_bh_total_pkt_cntr   += no_of_frags;
                        /* Increment the  total transmitted packet count for eth0 */
                        ul_bh_total_bytes_cntr = ul_bh_total_bytes_cntr + skb->len;

                        // V4 & V6 specific counters

                        if (ip_header->version == IWF_IPV4_VERSION)
                        {
                            ul_bh_total_pkt_cntr_ipv4   += no_of_frags;
                            ul_bh_total_bytes_cntr_ipv4 = ul_bh_total_bytes_cntr_ipv4 + skb->len;
                            if (MBTAC_NH_IGMP == ip_header->protocol)
                            {
                                ul_bh_mcast_total_ip_vlan_pkt_cntr++;
                                ul_bh_mcast_total_pkt_cntr_ipv4++;
                                ul_bh_mcast_total_bytes_cntr_ipv4 += ip_header->tot_len;
                                ul_bh_mcast_total_pkt_cntr++;
                                ul_bh_mcast_total_bytes_cntr += ip_header->tot_len;
                            }
                        }
                        else if (ip_header->version == IWF_IPV6_VERSION)
                        {
                            ul_bh_total_pkt_cntr_ipv6   += no_of_frags;
                            ul_bh_total_bytes_cntr_ipv6 = ul_bh_total_bytes_cntr_ipv6 + skb->len;
                        }

                    }
                }
            }

        }

        mbTacValidateAndUpdateBytes(skb,no_of_frags);
        //PR 154599ESPE02: FZM:RL80: LTE1401: MBTAC: UL traffic is not counted for MBTAC calculations
        //Moved this down to fix the above issue.
        if (ltm_gtp_traffic_flag == 1)
        {
            okfn(skb);
            return NF_STOLEN;
        }
    }
    return NF_ACCEPT;
}

unsigned int mbTac_UpLinkHostTrafficHook(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*)){

    if(skb !=NULL)
    {

        struct iphdr                     *ip_header    =  NULL;
        struct udphdr                    *udph         = NULL;
        uint8_t protocol = 0;

        ip_header  =  (struct iphdr *)skb->network_header;

        if(ip_header->version == IWF_IPV4_VERSION)
        {
            protocol = ip_header->protocol;
            udph =  (struct udphdr *)(skb->data + sizeof(struct iphdr));
        }
        else if(ip_header->version == IWF_IPV6_VERSION)
        {
            struct ipv6hdr *ipv6_header    =  NULL;
            ipv6_header = (struct ipv6hdr *)skb->data;
            protocol = ipv6_header->nexthdr;
            udph  = (struct udphdr*)(skb->data + sizeof(struct ipv6hdr));
        }
        if(IPPROTO_UDP == protocol)
        {
            /* if uplink packet. In DSP sim mode the UL packets comes on 30,000 port */
            if(udph!=NULL)                    {
                if((LTM_TOP_PORT_1 == ntohs(udph->source)) || (LTM_TOP_PORT_2 == ntohs(udph->source)))
                {
                    mbTacMarkingSplaneUplink(skb);
                }
                else if ((LTM_GTPU_PRT == ntohs(udph->dest)))
                {
                    tx_ul_gtpu_pkt_ip_local_out++;
                    okfn(skb);
                    return NF_STOLEN;
                }

            }
            else
            {
                if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
                {
                    if(0 != printk_ratelimit())
                    {
                        printk("mbTac_UpLinkHook  UDP Headre NULL\r\n");
                    }
                }
            }
        }
        else if (MBTAC_NH_SCTP == protocol || IPPROTO_IGMP == protocol)
        {
            mbTacMarkingCplaneUplink(skb);
        }
    }
    return NF_ACCEPT;
}
#endif


/* ------------------------------------------------------------------------------ */
int iwfSendErrorIndMsgToSgw(struct sk_buff *skb)
{
    int                   iPktLen           =  -1;
    iwf_gtp_mgmt_pkt_t    gtpMgmtPkt        =  {{{0}}};
    iwf_gtp_mgmt_pkt_t    *p_gtp_mgmt_pkt   =  NULL;
    struct iphdr          *ip_header        =  NULL;
    struct udphdr         *udph             =  NULL;
    struct net_device     *dev;
    bool                  isSkbToBeFreed    = true;
    uint32_t              teid;

    if(skb)
    {

        dev = skb->dev;

        // k2l: if carrier is bad, discard
        if(!netif_carrier_ok(dev)) {
            dev_kfree_skb_any(skb);
            skb = NULL;
            return iPktLen;
        }

        /* IP header */
        ip_header  =  (struct iphdr *)skb->data;

        if(ip_header->version == 4)
        {
            p_gtp_mgmt_pkt = (iwf_gtp_mgmt_pkt_t *)(skb->data + sizeof(struct iphdr) + sizeof(struct udphdr));
            gtpMgmtPkt.gtpuHdr.total_len = htons(sizeof(gtp_error_ind_ext_hdr_ie_ipv4_data_t) + sizeof(iwf_gtp_hdr_optional_t));
        }
        else//if(ip_header->version == 6)
        {
            p_gtp_mgmt_pkt = (iwf_gtp_mgmt_pkt_t *)(skb->data + sizeof(struct ipv6hdr) + sizeof(struct udphdr));
            gtpMgmtPkt.gtpuHdr.total_len = htons(sizeof(gtp_error_ind_ext_hdr_ie_data_t) + sizeof(iwf_gtp_hdr_optional_t));
        }

        gtpMgmtPkt.gtpuHdr.flags.version          = 1;
        gtpMgmtPkt.gtpuHdr.flags.proto_type       = 1;
        gtpMgmtPkt.gtpuHdr.flags.ext_hdr_present  = 1;
        gtpMgmtPkt.gtpuHdr.flags.seq_nr_present   = 1;
        gtpMgmtPkt.gtpuHdr.flags.npdu_num_present = 0;
        gtpMgmtPkt.gtpuHdr.msg_type = GTPU_MSG_ERR_IND;
        gtpMgmtPkt.payload.optionalField.seq_num  = 0;
        gtpMgmtPkt.payload.optionalField.npdu_num = 0;
        gtpMgmtPkt.payload.optionalField.nxt_exthdr_type = (GTPU_MGMT_EXT_HDR_UDPPORT);
        // create EXT HDR UDP
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.ext_hdr_len = 1;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.ext_hdr_udp_prt_1 = (LTM_GTPU_PRT>>8) & 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.ext_hdr_udp_prt_0 = LTM_GTPU_PRT & 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.nxt_ext_hdr_type = 0;
        // create GTP-U IE.s
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.type_1 = GTP_IE_TYPE_1;

        teid = ntohl(p_gtp_mgmt_pkt->gtpuHdr.teid);

        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_teid_b3 = ((teid)>>24) & 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_teid_b2 = ((teid)>>16) & 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_teid_b1 = ((teid)>>8) & 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_teid_b0 = (teid)& 0xff;
        gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.type_2 = GTP_IE_TYPE_2;
        gtpMgmtPkt.gtpuHdr.teid=0;

        if(ip_header->version == 4)
        {
            struct iphdr          ip_header_tmp = {0};
            struct rtable         *rt           =  NULL;

            /* calculate the packet size of ErrorIndMsg packet */
            iPktLen = (sizeof(iwf_gtp_hdr_t)+\
                    sizeof(iwf_gtp_hdr_optional_t)+\
                    sizeof(gtp_error_ind_ext_hdr_ie_ipv4_data_t))+\
                      sizeof(struct iphdr) + sizeof(struct udphdr);
            /* make skb->tail to point at correct address */
            pskb_trim(skb, iPktLen);

            udph       =  (struct udphdr *)(skb->data + sizeof(struct iphdr));

            /* take a back up of destination IP which we received in packet */
            ip_header_tmp.daddr = ip_header->daddr;

            /* fill the destination ip i.e. S-GW IP and source ip
               -  sgwIP needs to be updated from rx pkt */
            ip_header->daddr = ip_header->saddr;
            ip_header->saddr = ip_header_tmp.daddr; //RANS
            ip_header->tot_len = htons(skb->len);

            /* update DSCP value = 46 (GTP-U management channel) */
            ip_header->tos = _IWF_GTP_MGMT_MSG_DSCP << 2;

            /* UDP protocol */
            ip_header->protocol = IPPROTO_UDP;

            /* fill the udp header */
            udph->source = htons(_IWFD_DL_UDP_PORT_);
            udph->dest   = htons(_IWFD_DL_UDP_PORT_);

            /* disable udp checksum check */
            udph->check = 0;

            /* update udp length */
            udph->len = htons(iPktLen - sizeof(struct iphdr));

            //NOTE: Since IPs in Skb are swapped above we are coping saddr below
            //      which is one of U-Plan IPs for eNb
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_ip     = ip_header->saddr;
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.length_1 = 0;
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.length_0 = GTP_IE_TYPE_2_LEN_IPV4;

            /* copy gtp header and payload */
            memcpy((void *)(skb->data +  sizeof(struct iphdr) + sizeof(struct udphdr) ),\
                    (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct iphdr) + sizeof(struct udphdr)));

            /* do route lookup */
            rt = ip_route_output(dev_net(dev), ip_header->daddr, ip_header->saddr,
                    (ip_header->tos), 0);

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk("\n Dest Addr         : %x ",ip_header->daddr);
                    printk("\n SRC  Addr         : %x ",ip_header->saddr);
                    printk("\n skb->dev->name    : %s ",skb->dev->name);
                    printk("\n rt->dst.dev->name : %s ",rt->dst.dev->name);
                }
            }

            if (!IS_ERR(rt))
            {
                /* calculate ttl for the packet */
                ip_header->ttl = ip4_dst_hoplimit(&rt->dst);

                /* recompute ip header checksum*/
                ip_send_check(ip_header);

                /* fill in route details in skb */
                skb_dst_set(skb, &rt->dst);

                /* set skb->priority */
                skb->priority = _IWF_GTP_MGMT_MSG_DSCP;
                /* check if ipsec is disabled */
                if(false == iwf_ipsec_flag)
                {
                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                    dst_output(skb);

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed = false;

                    rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

                }
                else
                {
                    /* create and send IPSec pkt */
                    if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                    {
                        rx_dl_gtpu_mgmt_local_to_sgw_ipv4++;

                        /* kernel will take care of freeing it */
                        isSkbToBeFreed   = false;
                    }
                }//else (iwf_ipsec_flag)

            }
            else
            {
                rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4++;
            }
        }
        else
        {
            struct ipv6hdr     *ip6_header   =  NULL;
            struct rt6_info    *rt           =  NULL;
            uint32_t            tmp_ip6add_32[4] = {0};

            /* calculate the packet size of ErrorIndMsg packet */
            iPktLen = (sizeof(iwf_gtp_hdr_t)+\
                    sizeof(iwf_gtp_hdr_optional_t)+\
                    sizeof(gtp_error_ind_ext_hdr_ie_data_t))+\
                      sizeof(struct ipv6hdr) + sizeof(struct udphdr);
            /* make skb->tail to point at correct address */
            pskb_trim(skb, iPktLen);

            /* IPV6 and udp header */
            ip6_header  =  (struct ipv6hdr *)skb->data;
            udph       =  (struct udphdr *)(skb->data + sizeof(struct ipv6hdr));

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                }
            }
            ip6_header->version  = IWF_IPV6_VERSION;
            ip6_header->priority = _IWF_GTP_MGMT_MSG_DSCP >> 2; //Set DSCP
            ip6_header->flow_lbl[0] = (__u8)(_IWF_GTP_MGMT_MSG_DSCP << 6);//Set DSCP
            ip6_header->flow_lbl[1] = 0;
            ip6_header->flow_lbl[2] = 0;
            ip6_header->nexthdr = NEXTHDR_UDP;
            ip6_header->hop_limit = 0xff;
            ip6_header->payload_len = htons(skb->len - sizeof(struct ipv6hdr));

            //Swap Destination and source
            memcpy(tmp_ip6add_32, ip6_header->saddr.s6_addr32, sizeof(tmp_ip6add_32));
            memcpy(ip6_header->saddr.s6_addr32, ip6_header->daddr.s6_addr32, sizeof(tmp_ip6add_32));
            memcpy(ip6_header->daddr.s6_addr32, tmp_ip6add_32, sizeof(tmp_ip6add_32));

            if(IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO "%s %d skb->len:%d", __FUNCTION__, __LINE__, skb->len);
                    printk(KERN_INFO "%s %d ip6_header->payload_len:%d", __FUNCTION__, __LINE__,
                            ntohs(ip6_header->payload_len));
                    printk(KERN_INFO "%s %d Dst Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                            ip6_header->daddr.s6_addr32[0],
                            ip6_header->daddr.s6_addr32[1],
                            ip6_header->daddr.s6_addr32[2],
                            ip6_header->daddr.s6_addr32[3]);

                    printk(KERN_INFO "%s %d Source Address: %08X %08X %08X %08X", __FUNCTION__, __LINE__,
                            ip6_header->saddr.s6_addr32[0],
                            ip6_header->saddr.s6_addr32[1],
                            ip6_header->saddr.s6_addr32[2],
                            ip6_header->saddr.s6_addr32[3]);

                    printk(KERN_INFO "\n skb->dev->name    : %s ",skb->dev->name);
                }
            }

            /* fill the udp header */
            udph->len =  htons(skb->len - sizeof(struct ipv6hdr) );
            udph->source = htons(_IWFD_DL_UDP_PORT_);
            udph->dest  = htons(_IWFD_DL_UDP_PORT_);


            //NOTE: Since IPs in Skb are swapped above we are coping saddr below
            //      which is one of U-Plan IPs for eNb
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_ip      = ip6_header->saddr.s6_addr32[0];
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_ip1     = ip6_header->saddr.s6_addr32[1];
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_ip2     = ip6_header->saddr.s6_addr32[2];
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.p_ip3     = ip6_header->saddr.s6_addr32[3];

            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.length_1 = 0;
            gtpMgmtPkt.payload.extHdr.gtp_error_ind_ext_hdr_ie_data.length_0 = GTP_IE_TYPE_2_LEN_IPV6;

            /* copy gtp header and payload */
            memcpy((void *)(skb->data +  sizeof(struct ipv6hdr) + sizeof(struct udphdr) ),
                    (void *)&gtpMgmtPkt,iPktLen -(sizeof(struct ipv6hdr) + sizeof(struct udphdr)));
            udph->check = 0;
            udph->check = csum_ipv6_magic(&ip6_header->saddr,
                    &ip6_header->daddr,ntohs(udph->len), IPPROTO_UDP,
                    csum_partial(udph, ntohs(udph->len), 0));
            if (udph->check == 0)
                udph->check = CSUM_MANGLED_0;

            rt = rt6_lookup(dev_net(dev), &ip6_header->daddr, &ip6_header->saddr, 0, 0);
            /*  fill in route details in skb */
            if (rt && !IS_ERR(rt))
            {
                //If Getway is zero then fill destination address into getway
                if (ipv6_addr_any(&rt->rt6i_gateway))
                {
                    memcpy(&rt->rt6i_gateway, &ip6_header->daddr, sizeof(struct in6_addr));
                }

                skb_dst_set(skb, &rt->dst);

                //Set skb->local_df to 1 to allow fragmentation.
                skb->local_df = 1;

                /* check if ipsec is disabled */
                if(false == iwf_ipsec_flag)
                {
                    /* done with packet formatting, send pkt to Sgw on udp port 2152 */
                    dst_output(skb);

                    /* kernel will take care of freeing it */
                    isSkbToBeFreed = false;

                    rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

                }
                else
                {
                    /* create and send IPSec pkt */
                    if( _IWF_SUCCESS_ == iwfSendIpsecPkt(skb))
                    {
                        rx_dl_gtpu_mgmt_local_to_sgw_ipv6++;

                        /* kernel will take care of freeing it */
                        isSkbToBeFreed   = false;
                    }
                }//else (iwf_ipsec_flag)
            }// if (rt)
            else
            {
                rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6++;
            }
        }// else if (ip_version == 6)
    }//if (skb)

    if((true == isSkbToBeFreed) && (NULL != skb))
    {
        /* free the skb */
        dev_kfree_skb_any(skb);
        skb = NULL;
    }

    return iPktLen;
}
static int iwf_table_dump(struct seq_file *m, void *v)
{

    ip_se_lte_up_proc_dump(m);
    ip_se_uplink_proc_dump (m);
    ip_se_downlink_proc_dump (m);
    proc_dump_trswConfigTable (m);
    proc_dump_errIndTable(m);

    return 0;
}


/* ------------------------------------------------------------------------------ */
static int iwf_cntrs_proc_read(struct seq_file *m, void *v)
{
    int portidx;

    seq_printf(m, "============================================== \n");
    seq_printf(m, "               FZM IWF COUNTERS \n");
    seq_printf(m, "-------------------- UL ---------------------- \n");
    seq_printf(m, "IPSec enabled   : %d \n",iwf_ipsec_flag);
    seq_printf(m, "tx_total_ul_gtpu_pkt_from_tupu              : %u \n", tx_total_ul_gtpu_pkt_from_tupu);
    seq_printf(m, "tx_ul_gtpu_pkt_from_tupu_ipv4               : %u \n", tx_ul_gtpu_pkt_from_tupu_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_to_sgw_cntr_ipv4             : %u \n", tx_ul_gtpu_pkt_to_sgw_cntr_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_no_route_ipv4           : %u \n", tx_ul_gtpu_pkt_drop_no_route_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_cntr_ipv4               : %u \n", tx_ul_gtpu_pkt_drop_cntr_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4     : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv4   : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4      : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv4 : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_from_tupu_ipv6               : %u \n", tx_ul_gtpu_pkt_from_tupu_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_to_sgw_cntr_ipv6             : %u \n", tx_ul_gtpu_pkt_to_sgw_cntr_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_no_route_ipv6           : %u \n", tx_ul_gtpu_pkt_drop_no_route_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_cntr_ipv6               : %u \n", tx_ul_gtpu_pkt_drop_cntr_ipv6);
    seq_printf(m, "tx_ul_invalid_gtpu_pkt_drop_cntr            : %u \n", tx_ul_invalid_gtpu_pkt_drop_cntr);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_no_entry                : %u \n", tx_ul_gtpu_pkt_drop_no_entry);
    seq_printf(m, "tx_ul_gtpu_pkt_drop_null_pkt_rcvd           : %u \n", tx_ul_gtpu_pkt_drop_null_pkt_rcvd);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6     : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6 : %u \n", tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv6   : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv6      : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv6 : %u \n", tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv6);
    seq_printf(m, "tx_ul_gtpu_pkt_post_routing                 : %u \n", tx_ul_gtpu_pkt_post_routing);
    seq_printf(m, "tx_ul_gtpu_pkt_ip_local_out                 : %u \n", tx_ul_gtpu_pkt_ip_local_out);

    seq_printf(m, "-------------------- DL ---------------------- \n");
    seq_printf(m, "rx_total_pkt_ipv4                           : %u \n", rx_total_pkt_ipv4);
    seq_printf(m, "rx_total_pkt_ipv4_pre_route                 : %u \n", rx_total_pkt_ipv4_pre_route);
    seq_printf(m, "rx_total_pkt_ipv4_post_route                : %u \n", rx_total_pkt_ipv4_post_route);
    seq_printf(m, "rx_total_pkt_ipsec_post_route               : %u \n", rx_total_pkt_ipsec_post_route);
    seq_printf(m, "rx_total_pkt_ipv6_pre_route                 : %u \n", rx_total_pkt_ipv6_pre_route);
    seq_printf(m, "rx_total_pkt_ipv6_post_route                : %u \n", rx_total_pkt_ipv6_post_route);
    seq_printf(m, "rx_udp_total_pkt_ipv4                       : %u \n", rx_udp_total_pkt_ipv4);
    seq_printf(m, "rx_dl_gtpu_mgmt_local_req_ipv4              : %u \n", rx_dl_gtpu_mgmt_local_req_ipv4);
    seq_printf(m, "rx_dl_gtpu_mgmt_local_to_sgw_ipv4           : %u \n", rx_dl_gtpu_mgmt_local_to_sgw_ipv4);
    seq_printf(m, "rx_gtpu_total_pkt                           : %u \n", rx_gtpu_total_pkt);
    seq_printf(m, "rx_total_pkt_ipv6                           : %u \n", rx_total_pkt_ipv6);
    seq_printf(m, "rx_udp_total_pkt_ipv6                       : %u \n", rx_udp_total_pkt_ipv6);
    seq_printf(m, "rx_dl_pkt_to_tupu                           : %u \n", rx_dl_pkt_to_tupu);
    seq_printf(m, "rx_dl_gtpu_mgmt_local_req_ipv6              : %u \n", rx_dl_gtpu_mgmt_local_req_ipv6);
    seq_printf(m, "rx_dl_gtpu_mgmt_local_to_sgw_ipv6           : %u \n", rx_dl_gtpu_mgmt_local_to_sgw_ipv6);
    seq_printf(m, "rx_dl_gtpu_pkt_drop_cntr                    : %u \n", rx_dl_gtpu_pkt_drop_cntr);
    seq_printf(m, "rx_dl_gtpu_xfrm_drop_no_policy_ipv4         : %u \n", rx_dl_gtpu_xfrm_drop_no_policy_ipv4);
    seq_printf(m, "tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4 : %u \n", tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4);
    seq_printf(m, "rx_dl_gtpu_xfrm_drop_no_policy_ipv6         : %u \n", rx_dl_gtpu_xfrm_drop_no_policy_ipv6);
    seq_printf(m, "rx_udp_checksum_err_drop_ipv6               : %u \n", rx_udp_checksum_err_drop_ipv6);
    seq_printf(m, "rx_dl_gtpu_pkt_drop_wrong_sicp_hdr          : %u \n", rx_dl_gtpu_pkt_drop_wrong_sicp_hdr);
    seq_printf(m, "rx_dl_gtpu_pkt_drop_invalid_gtp_hdr         : %u \n", rx_dl_gtpu_pkt_drop_invalid_gtp_hdr);
    seq_printf(m, "rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4 : %u \n", rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4);
    seq_printf(m, "rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6 : %u \n", rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6);
    seq_printf(m, "rx_dl_pkt_drop_multicast_src_ipv6           : %u \n", rx_dl_pkt_drop_multicast_src_ipv6);
    seq_printf(m, "rx_dl_checksum_error_ipv4                   : %u \n", rx_dl_checksum_error_ipv4);
    seq_printf(m, "rx_dl_icmp_checksum_error_ipv4              : %u \n", rx_dl_icmp_checksum_error_ipv4);
    seq_printf(m, "rx_dl_icmp_checksum_error_ipv6              : %u \n", rx_dl_icmp_checksum_error_ipv6);
    seq_printf(m, "rx_udp_checksum_err_drop_ipv6              : %u \n", rx_udp_checksum_err_drop_ipv6);
    seq_printf(m, "============================================== \n");
    seq_printf(m, "    FZM MBTAC COUNTERS  \n");
    seq_printf(m, "ul_splane_pkt_cntr : %u \n",ul_splane_pkt_cntr);
    seq_printf(m, "ul_uplane_pkt_cntr : %u \n",ul_uplane_pkt_cntr);
    seq_printf(m, "ul_cplane_pkt_cntr : %u \n",ul_cplane_pkt_cntr);
    seq_printf(m, "ul_mcast_cplane_pkt_cntr : %u \n",ul_mcast_cplane_pkt_cntr);
    seq_printf(m, "-------------------------- \n");
    seq_printf(m, "dl_splane_pkt_cntr : %u \n",dl_splane_pkt_cntr);
    seq_printf(m, "dl_uplane_pkt_cntr : %u \n",dl_uplane_pkt_cntr);
    seq_printf(m, "dl_cplane_pkt_cntr : %u \n",dl_cplane_pkt_cntr);
    seq_printf(m, "dl_mcast_cplane_pkt_cntr : %u \n",dl_mcast_cplane_pkt_cntr);
    seq_printf(m, "============================================== \n");
    seq_printf(m, "    FZM L2 VLAN COUNTERS  \n");
    for(portidx = 0;portidx < NumofPorts; portidx++)
    {
        switch(portidx)
        {
            case 0:
                seq_printf(m, "eth0\n");
                break;
            case 1:
                seq_printf(m, "eth1\n");
                break;
            case 2:
                seq_printf(m, "eth3\n");
                break;
        }
        seq_printf(m, "(%d)DiscardTaggedFramesCount		: %u \n",portidx,DiscardTaggedFramesCount[portidx]);
        seq_printf(m, "(%d)DiscardUnTaggedFramesCount	: %u \n",portidx,DiscardUnTaggedFramesCount[portidx]);
        seq_printf(m, "(%d)TaggedFramesCount  			: %u \n",portidx,TaggedFramesCount[portidx]);
    }
    seq_printf(m, "============================================== \n");

    return 0;
}

/* ------------------------------------------------------------------------------ */
static ssize_t iwf_cntrs_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos)
{
    char * kernel_buf;
    if (count > PROCFS_MAX_SIZE)
    {
        return -EFAULT;
    }

    kernel_buf = kmalloc(count + 1, GFP_KERNEL);
    if(NULL == kernel_buf)
    {
        return -ENOMEM;
    }
    if(copy_from_user(kernel_buf, buffer, count))
    {
        kfree(kernel_buf);
        return -EFAULT;
    }
    kernel_buf[count] = '\0';


    if(0 == strncmp("0", kernel_buf, 1))
    {
        tx_total_ul_gtpu_pkt_from_tupu    = 0;
        tx_ul_gtpu_pkt_from_tupu_ipv4     = 0;
        tx_ul_gtpu_pkt_from_tupu_ipv6     = 0;
        tx_ul_gtpu_pkt_drop_no_entry      = 0;
        tx_ul_gtpu_pkt_to_sgw_cntr_ipv4   = 0;
        tx_ul_gtpu_pkt_to_sgw_cntr_ipv6   = 0;
        tx_ul_gtpu_pkt_drop_cntr_ipv4     = 0;
        tx_ul_gtpu_pkt_drop_cntr_ipv6     = 0;
        tx_ul_invalid_gtpu_pkt_drop_cntr  = 0;
        tx_ul_gtpu_pkt_drop_no_route_ipv4 = 0;
        tx_ul_gtpu_pkt_drop_no_route_ipv6 = 0;
        tx_ul_gtpu_pkt_drop_null_pkt_rcvd = 0;
        tx_ul_gtpu_pkt_post_routing       = 0;
        tx_ul_gtpu_pkt_ip_local_out       = 0;


        rx_total_pkt_ipv4                 = 0;
        rx_total_pkt_ipv4_pre_route       = 0;
        rx_total_pkt_ipv4_post_route      = 0;
        rx_total_pkt_ipsec_post_route     = 0;
        rx_total_pkt_ipv6_pre_route       = 0;
        rx_total_pkt_ipv6_post_route      = 0;
        rx_total_pkt_ipv6                 = 0;
        rx_udp_total_pkt_ipv4             = 0;
        rx_udp_total_pkt_ipv6             = 0;
        rx_udp_checksum_err_drop_ipv6     = 0;
        rx_gtpu_total_pkt                 = 0;
        rx_dl_pkt_to_tupu                 = 0;
        rx_dl_gtpu_pkt_drop_cntr          = 0;
        rx_dl_gtpu_mgmt_local_to_sgw_ipv4 = 0;
        rx_dl_gtpu_mgmt_local_to_sgw_ipv6 = 0;
        rx_dl_gtpu_mgmt_local_req_ipv4    = 0;
        rx_dl_gtpu_mgmt_local_req_ipv6    = 0;
        rx_dl_gtpu_pkt_drop_wrong_sicp_hdr = 0;
        rx_dl_gtpu_pkt_drop_invalid_gtp_hdr = 0;
        rx_dl_gtpu_xfrm_drop_no_policy_ipv4 = 0;
        rx_dl_gtpu_xfrm_drop_no_policy_ipv6 = 0;
        rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv4  = 0;
        rx_dl_gtpu_pkt_drop_mgmt_back_no_route_ipv6  = 0;
        rx_dl_pkt_drop_multicast_src_ipv6 = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv4      = 0;
        tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv4  = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv4    = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv4       = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv4  = 0;
        tx_ul_gtpu_pkt_xfrm_drop_lookup_failed_ipv6  = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_policy_ipv6      = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_routefwd_ipv6    = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_route_ipv6       = 0;
        tx_ul_gtpu_pkt_xfrm_drop_no_xfrm_state_ipv6  = 0;
        ul_mcast_pkt_cntr_ipv6        = 0;
        ul_mcast_bytes_cntr_ipv6      = 0;
        dl_mcast_pkt_cntr_ipv6        = 0;
        dl_mcast_bytes_cntr_ipv6      = 0;
        dl_mcast_disc_pkt_cntr_ipv6   = 0;
        dl_mcast_disc_bytes_cntr_ipv6 = 0;
        ul_bh_mcast_total_pkt_cntr        = 0;
        ul_bh_mcast_total_bytes_cntr      = 0;
        dl_bh_mcast_total_pkt_cntr        = 0;
        dl_bh_mcast_total_bytes_cntr      = 0;
        dl_bh_mcast_disc_total_pkt_cntr   = 0;
        dl_bh_mcast_disc_total_bytes_cntr = 0;
    }

    kfree(kernel_buf);
    return count;
}

static int iwf_loglevel_help_proc_read(struct seq_file *m, void *v)
{

    seq_printf(m,"================================================ \n");
    seq_printf(m,"                   IWF LOGLEVELS \n");
    seq_printf(m,"------------------------------------------------- \n");
    seq_printf(m,"     User Input                     LOGLEVEL  \n");
    seq_printf(m,"------------------------------------------------- \n");
    seq_printf(m,"         0                   IWF_LOG_LEVEL_NO_LOG \n");
    seq_printf(m,"         1                   IWF_LOG_LEVEL_PKT_DUMP \n");
    seq_printf(m,"         2                   IWF_LOG_LEVEL_DL_ERR_CRT_LOG \n");
    seq_printf(m,"         4                   IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG \n");
    seq_printf(m,"         8                   IWF_LOG_LEVEL_DL_SUCC_CRT_LOG \n");
    seq_printf(m,"        16                   IWF_LOG_LEVEL_DL_SUCC_NORMAL_LOG \n");
    seq_printf(m,"        32                   IWF_LOG_LEVEL_UL_ERR_CRT_LOG \n");
    seq_printf(m,"        64                   IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG \n");
    seq_printf(m,"       128                   IWF_LOG_LEVEL_UL_SUCC_CRT_LOG \n");
    seq_printf(m,"       256                   IWF_LOG_LEVEL_UL_SUCC_NORMAL_LOG \n");
    seq_printf(m,"       512                   IWF_LOG_LEVEL_TABLES_CONFIG\n");


    seq_printf(m,"       96                    IWF_LOG_LEVEL_UL_ERR_ALL_LOG (IWF_LOG_LEVEL_UL_ERR_CRT_LOG + IWF_LOG_LEVEL_UL_ERR_NORMAL_LOG) \n");
    seq_printf(m,"       384                   IWF_LOG_LEVEL_UL_SUCC_ALL_LOG (IWF_LOG_LEVEL_UL_SUCC_CRT_LOG + IWF_LOG_LEVEL_UL_SUCC_NORMAL_LOG)\n");
    seq_printf(m,"       6                     IWF_LOG_LEVEL_DL_ERR_ALL_LOG (IWF_LOG_LEVEL_DL_ERR_CRT_LOG + IWF_LOG_LEVEL_DL_ERR_NORMAL_LOG)  \n");
    seq_printf(m,"       24                    IWF_LOG_LEVEL_DL_SUCC_ALL_LOG (IWF_LOG_LEVEL_DL_SUCC_CRT_LOG + IWF_LOG_LEVEL_DL_SUCC_NORMAL_LOG)\n");
    seq_printf(m,"================================================= \n");


    return 0;
}

/* ------------------------------------------------------------------------------ */
static int iwf_loglevel_add_proc_read(struct seq_file *m, void *v)
{
    seq_printf(m,"\n iwf_log_level Added : %d \n ",iwfLogLevelReceived);
    return 0;
}

/* ------------------------------------------------------------------------------ */
static ssize_t iwf_loglevel_add_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos)
{

    int retVal;

    char *kernel_buf = NULL;

    if (count > PROCFS_MAX_SIZE)
    {
        return -EFAULT;
    }

    kernel_buf = kmalloc(count + 1, GFP_KERNEL);
    if(NULL == kernel_buf)
    {
        return -ENOMEM;
    }
    if(copy_from_user(kernel_buf, buffer, count))
    {
        kfree(kernel_buf);
        return -EFAULT;
    }
    kernel_buf[count] = '\0';

    retVal = kstrtouint(kernel_buf, 0, &iwfLogLevelReceived);

    if(!retVal)
    {

//
// Testing hack: It assume bits 14 and 15 are not used. bit 15 for iwf_sim_flag=true, 14 for false
// Currently it only uses update bit 10 for setting log level
//
        // Test enable sim
        if(iwfLogLevelReceived & 0x80000000) {
            iwf_sim_flag = true;
            goto checkbuf;
        }

        // Test disable sim
        if(iwfLogLevelReceived & 0x40000000) {
            iwf_sim_flag = false;
            goto checkbuf;
        }

        if(0 == iwfLogLevelReceived)
        {
            iwf_log_level = 0;
        }
        else if(IWF_LOG_LEVEL_MAX > iwfLogLevelReceived)
        {
            iwf_log_level = iwf_log_level | iwfLogLevelReceived;
        }
        else
        {
            printk(KERN_INFO _IWF_MODULE_NAME": %s: ERR!! Invalid log level entered : %d\n",__func__,iwfLogLevelReceived);
        }

    }

    checkbuf: if(kernel_buf)
        kfree(kernel_buf);

    return count;

}

/* ------------------------------------------------------------------------------ */
static int iwf_loglevel_delete_proc_read(struct seq_file *m, void *v)
{
    seq_printf(m,"\n iwf_log_level Deleted : %d \n ",iwfLogLevelReceived);
    return 0;
}

/* ------------------------------------------------------------------------------ */
static ssize_t iwf_loglevel_delete_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos)
{
    int retVal;
    char *kernel_buf = NULL;

    if (count > PROCFS_MAX_SIZE)
    {
        return -EFAULT;
    }

    kernel_buf = kmalloc(count + 1, GFP_KERNEL);

    if (!kernel_buf)
    {
        printk(KERN_INFO _IWF_MODULE_NAME": %s: ERR!! kmalloc() failed : %d\n",__func__,iwfLogLevelReceived);
        return -EFAULT;
    }

    if(copy_from_user(kernel_buf, buffer, count))
    {
        kfree(kernel_buf);
        return -EFAULT;
    }

    kernel_buf[count] = '\0';
    retVal = kstrtouint(kernel_buf, 0, &iwfLogLevelReceived);
    if(!retVal)
    {
        if(0 == iwfLogLevelReceived)
        {
            iwf_log_level = 0;
        }

        else if(IWF_LOG_LEVEL_MAX > iwfLogLevelReceived)
        {
            iwf_log_level = iwf_log_level & (~(iwfLogLevelReceived));
        }

        else
        {
            printk(KERN_INFO _IWF_MODULE_NAME": %s: ERR!! Invalid log level entered : %d\n",__func__,iwfLogLevelReceived);
        }
    }
    if(kernel_buf)
        kfree(kernel_buf);

    return count;
}


/* ------------------------------------------------------------------------------ */

static int iwf_loglevel_proc_read(struct seq_file *m, void *v)
{
    seq_printf(m,"\n iwf_log_level : %d \n ",iwf_log_level);
    return 0;
}

/* ------------------------------------------------------------------------------ */

static int iwf_ip_cntrs_proc_read(struct seq_file *m, void *v)
{
    int i;
    seq_printf(m, "============================================== \n");
    seq_printf(m, "               FZM IWF IP COUNTERS \n");
    seq_printf(m, "-------------------- DL ---------------------- \n");
    seq_printf(m, "dl_bh_total_pkt_cntr:%llu \n", dl_bh_total_pkt_cntr);
    seq_printf(m, " dl_bh_total_bytes_cntr:%llu \n",  dl_bh_total_bytes_cntr);
    seq_printf(m, "dl_bh_total_drop_pkt_cntr:%llu \n", dl_bh_total_drop_pkt_cntr);
    seq_printf(m, "dl_bh_mcast_total_ip_vlan_pkt_cntr:%llu \n", dl_bh_mcast_total_ip_vlan_pkt_cntr);

    seq_printf(m, "-------------------- UL ---------------------- \n");
    seq_printf(m, "ul_bh_total_pkt_cntr:%llu \n", ul_bh_total_pkt_cntr);
    seq_printf(m, "ul_bh_total_bytes_cntr:%llu \n", ul_bh_total_bytes_cntr);
    seq_printf(m, "ul_bh_mcast_total_ip_vlan_pkt_cntr:%llu \n", ul_bh_mcast_total_ip_vlan_pkt_cntr);
    seq_printf(m, "============================================== \n");


    seq_printf(m, "--------------------  IpV4 DL ---------------------- \n");
    seq_printf(m, "dl_bh_total_pkt_cntr_ipv4:%llu \n", dl_bh_total_pkt_cntr_ipv4);
    seq_printf(m, "dl_bh_total_bytes_cntr_ipv4:%llu \n",  dl_bh_total_bytes_cntr_ipv4);
    seq_printf(m, "dl_bh_total_drop_pkt_cntr_ipv4:%llu \n", dl_bh_total_drop_pkt_cntr);

    seq_printf(m, "-------------------- IpV4 UL ---------------------- \n");
    seq_printf(m, "ul_bh_total_pkt_cntr_ipv4:%llu \n", ul_bh_total_pkt_cntr_ipv4);
    seq_printf(m, "ul_bh_total_bytes_cntr_ipv4:%llu \n", ul_bh_total_bytes_cntr_ipv4);
    seq_printf(m, "ul_bh_mcast_total_pkt_cntr_ipv4:%llu \n", ul_bh_mcast_total_pkt_cntr_ipv4);
    seq_printf(m, "ul_bh_mcast_total_bytes_cntr_ipv4:%llu \n", ul_bh_mcast_total_bytes_cntr_ipv4);
    seq_printf(m, "============================================== \n");

    seq_printf(m, "--------------------  IpV6 DL ---------------------- \n");
    seq_printf(m, "dl_bh_total_pkt_cntr_ipv6:%llu \n", dl_bh_total_pkt_cntr_ipv6);
    seq_printf(m, "dl_bh_total_bytes_cntr_ipv6:%llu \n",  dl_bh_total_bytes_cntr_ipv6);
    seq_printf(m, "dl_bh_total_drop_pkt_cntr_ipv6:%llu \n", dl_bh_total_drop_pkt_cntr);

    seq_printf(m, "-------------------- IpV6 UL ---------------------- \n");
    seq_printf(m, "ul_bh_total_pkt_cntr_ipv6:%llu \n", ul_bh_total_pkt_cntr_ipv6);
    seq_printf(m, "ul_bh_total_bytes_cntr_ipv6:%llu \n", ul_bh_total_bytes_cntr_ipv6);
    seq_printf(m, "============================================== \n");

    seq_printf(m, "-------------------- IPV6 MCAST MLD & ND DL ---------------------- \n");
    seq_printf(m, "dl_mcast_pkt_cntr_ipv6:%llu \n", dl_mcast_pkt_cntr_ipv6);
    seq_printf(m, "dl_mcast_bytes_cntr_ipv6:%llu \n",  dl_mcast_bytes_cntr_ipv6);
    seq_printf(m, "dl_mcast_disc_pkt_cntr_ipv6:%llu \n", dl_mcast_disc_pkt_cntr_ipv6);
    seq_printf(m, "dl_mcast_disc_bytes_cntr_ipv6:%llu \n",  dl_mcast_disc_bytes_cntr_ipv6);


    seq_printf(m, "-------------------- IPV6 MCAST MLD & ND UL ---------------------- \n");
    seq_printf(m, "ul_mcast_pkt_cntr_ipv6:%llu \n",ul_mcast_pkt_cntr_ipv6);
    seq_printf(m, "ul_mcast_bytes_cntr_ipv6:%llu \n", ul_mcast_bytes_cntr_ipv6);
    seq_printf(m, "============================================== \n");

    seq_printf(m, "-------------------- BH MCAST TOTAL DL ---------------------- \n");
    seq_printf(m, "dl_bh_mcast_total_pkt_cntr:%llu \n", dl_bh_mcast_total_pkt_cntr);
    seq_printf(m, "dl_bh_mcast_total_bytes_cntr:%llu \n", dl_bh_mcast_total_bytes_cntr);
    seq_printf(m, "dl_bh_mcast_disc_total_pkt_cntr:%llu \n", dl_bh_mcast_disc_total_pkt_cntr);
    seq_printf(m, "dl_bh_mcast_disc_total_bytes_cntr:%llu \n", dl_bh_mcast_disc_total_bytes_cntr);

    seq_printf(m, "-------------------- BH MCAST TOTAL UL ---------------------- \n");
    seq_printf(m, "ul_bh_mcast_total_pkt_cntr:%llu \n", ul_bh_mcast_total_pkt_cntr);
    seq_printf(m, "ul_bh_mcast_total_bytes_cntr:%llu \n", ul_bh_mcast_total_bytes_cntr);
    seq_printf(m, "============================================== \n");

    for(i=0;i< VLANIDS_MAX;i++)
    {
        if(ipCounters[i].vlanId)
        {
            seq_printf(m, "-------------------- DL ---------------------- \n");
            seq_printf(m, "dl_bh_total_pkt_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_pkt_cntr);
            seq_printf(m, " dl_bh_total_bytes_cntr_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_bh_total_bytes_cntr);
            seq_printf(m, "dl_bh_total_drop_pkt_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_drop_pkt_cntr);

            seq_printf(m, "-------------------- UL ---------------------- \n");
            seq_printf(m, "ul_bh_total_pkt_cntr_%hd:%llu \n",ipCounters[i].vlanId,ipCounters[i].ul_bh_total_pkt_cntr);
            seq_printf(m, "ul_bh_total_bytes_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_total_bytes_cntr);
            seq_printf(m, "ul_bh_mcast_total_pkt_cntr_ipv4_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_mcast_total_pkt_cntr_ipv4);
            seq_printf(m, "ul_bh_mcast_total_bytes_cntr_ipv4_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_mcast_total_bytes_cntr_ipv4);

            seq_printf(m, "-------------------- IPV4DL ---------------------- \n");
            seq_printf(m, "dl_bh_total_pkt_cntr_ipv4_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_pkt_cntr_ipv4);
            seq_printf(m, "dl_bh_total_bytes_cntr_ipv4_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_bh_total_bytes_cntr_ipv4);
            seq_printf(m, "dl_bh_total_drop_pkt_cntr_ipv4_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_drop_pkt_cntr_ipv4);

            seq_printf(m, "-------------------- IPV4UL ---------------------- \n");
            seq_printf(m, "ul_bh_total_pkt_cntr_ipv4_%hd:%llu \n",ipCounters[i].vlanId,ipCounters[i].ul_bh_total_pkt_cntr_ipv4);
            seq_printf(m, "ul_bh_total_bytes_cntr_ipv4_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_total_bytes_cntr_ipv4);



            seq_printf(m, "-------------------- IPV6DL ---------------------- \n");
            seq_printf(m, "dl_bh_total_pkt_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_pkt_cntr_ipv6);
            seq_printf(m, "dl_bh_total_bytes_cntr_ipv6_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_bh_total_bytes_cntr_ipv6);
            seq_printf(m, "dl_bh_total_drop_pkt_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_total_drop_pkt_cntr_ipv6);

            seq_printf(m, "-------------------- IPV6UL ---------------------- \n");
            seq_printf(m, "ul_bh_total_pkt_cntr_ipv6_%hd:%llu \n",ipCounters[i].vlanId,ipCounters[i].ul_bh_total_pkt_cntr_ipv6);
            seq_printf(m, "ul_bh_total_bytes_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_total_bytes_cntr_ipv6);


            seq_printf(m, "-------------------- IPV6 MCAST MLD & ND DL ---------------------- \n");
            seq_printf(m, "dl_mcast_pkt_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_mcast_pkt_cntr_ipv6);
            seq_printf(m, "dl_mcast_bytes_cntr_ipv6_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_mcast_bytes_cntr_ipv6);
            seq_printf(m, "dl_mcast_disc_pkt_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_mcast_disc_pkt_cntr_ipv6);
            seq_printf(m, "dl_mcast_disc_bytes_cntr_ipv6_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_mcast_disc_bytes_cntr_ipv6);


            seq_printf(m, "-------------------- IPV6 MCAST MLD & ND UL ---------------------- \n");
            seq_printf(m, "ul_mcast_pkt_cntr_ipv6_%hd:%llu \n",ipCounters[i].vlanId,ipCounters[i].ul_mcast_pkt_cntr_ipv6);
            seq_printf(m, "ul_mcast_bytes_cntr_ipv6_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_mcast_bytes_cntr_ipv6);

            seq_printf(m, "-------------------- BH MCAST TOTAL DL ---------------------- \n");
            seq_printf(m, "dl_bh_mcast_total_pkt_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_mcast_total_pkt_cntr);
            seq_printf(m, "dl_bh_mcast_total_bytes_cntr_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_bh_mcast_total_bytes_cntr);
            seq_printf(m, "dl_bh_mcast_disc_total_pkt_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].dl_bh_mcast_disc_total_pkt_cntr);
            seq_printf(m, "dl_bh_mcast_disc_total_bytes_cntr_%hd:%llu \n",  ipCounters[i].vlanId,ipCounters[i].dl_bh_mcast_disc_total_bytes_cntr);


            seq_printf(m, "-------------------- BH MCAST TOTAL UL ---------------------- \n");
            seq_printf(m, "ul_bh_mcast_total_pkt_cntr_%hd:%llu \n",ipCounters[i].vlanId,ipCounters[i].ul_bh_mcast_total_pkt_cntr);
            seq_printf(m, "ul_bh_mcast_total_bytes_cntr_%hd:%llu \n", ipCounters[i].vlanId,ipCounters[i].ul_bh_mcast_total_bytes_cntr);

        }
    }
    return 0;

}
/* ------------------------------------------------------------------------------ */

static ssize_t iwf_ip_cntrs_proc_write(struct file *file, const char __user *buffer,size_t count, loff_t *ppos)
{
    char * kernel_buf;
    if (count > PROCFS_MAX_SIZE)
    {
        return -EFAULT;
    }

    kernel_buf = kmalloc(count + 1, GFP_KERNEL);
    if(NULL == kernel_buf)
    {
        return -ENOMEM;
    }
    if(copy_from_user(kernel_buf, buffer, count))
    {
        kfree(kernel_buf);
        return -EFAULT;
    }
    kernel_buf[count] = '\0';


    if(0 == strncmp("0", kernel_buf, 1))
    {
        dl_bh_total_pkt_cntr        = 0;
        dl_bh_total_bytes_cntr      = 0;
        dl_bh_total_drop_pkt_cntr   = 0;

        ul_bh_total_pkt_cntr        = 0;
        ul_bh_total_bytes_cntr      = 0;
    }

    kfree(kernel_buf);
    return count;

}

/* ------------------------------------------------------------------------------ */

#ifdef __USE_NET_FILTER__
void iwfD_uplink_hook_register(bool bRegisterFlag)
{
    /* register the uplink hook, this will be done only in simulation mode */
    if(true == bRegisterFlag)
    {
        /* if not already registered */
        if(false == bIsULHookRegistered)
        {
            netfilter_uplink_ops.hook              =       iwfD_UpLinkHook;
            netfilter_uplink_ops.pf                =       PF_INET;
            netfilter_uplink_ops.hooknum           =       NF_INET_LOCAL_IN;
            netfilter_uplink_ops.priority          =       NF_IP_PRI_FIRST + 1;
            nf_register_hook(&netfilter_uplink_ops);
            bIsULHookRegistered                    = true;
            if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO _IWF_MODULE_NAME": %s Uplink hook registered \n",__func__);
                }
            }
        }
    }
    else if(true == bIsULHookRegistered )/* unregister the hook */
    {
        nf_unregister_hook(&netfilter_uplink_ops);
        bIsULHookRegistered = false;
        if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO _IWF_MODULE_NAME": %s Uplink hook UN-REGISTERED \n",__func__);
            }
        }
    }
}

void mbtac_uplink_hook_register(void)
{

    // Traffic generated for other host (C / S / U Plane traffic)
    mbtac_uplink_ops.hook              =       mbTac_UpLinkHook;
    mbtac_uplink_ops.pf                =       PF_INET;
    mbtac_uplink_ops.hooknum           =       NF_INET_POST_ROUTING;
    mbtac_uplink_ops.priority          =       NF_IP_PRI_FIRST + 1;
    nf_register_hook(&mbtac_uplink_ops);

    mbtac_uplink_ipv6_ops.hook              =       mbTac_UpLinkHook;
    mbtac_uplink_ipv6_ops.pf                =       PF_INET6;
    mbtac_uplink_ipv6_ops.hooknum           =       NF_INET_POST_ROUTING;
    mbtac_uplink_ipv6_ops.priority          =       NF_IP6_PRI_FIRST + 1;
    nf_register_hook(&mbtac_uplink_ipv6_ops);

    if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
    {
        if(0 != printk_ratelimit())
        {
            printk(KERN_INFO _IWF_MODULE_NAME": %s Uplink hook for mbtac is registered \n",__func__);
        }
    }

    // Traffic generated from host (C & S Plane traffic)
    mbtac_uplink_host_ops.hook              =       mbTac_UpLinkHostTrafficHook;
    mbtac_uplink_host_ops.pf                =       PF_INET;
    mbtac_uplink_host_ops.hooknum           =       NF_INET_LOCAL_OUT;
    mbtac_uplink_host_ops.priority          =       NF_IP_PRI_FIRST + 1;
    nf_register_hook(&mbtac_uplink_host_ops);

    mbtac_uplink_host_ipv6_ops.hook              =       mbTac_UpLinkHostTrafficHook;
    mbtac_uplink_host_ipv6_ops.pf                =       PF_INET6;
    mbtac_uplink_host_ipv6_ops.hooknum           =       NF_INET_LOCAL_OUT;
    mbtac_uplink_host_ipv6_ops.priority          =       NF_IP6_PRI_FIRST + 1;
    nf_register_hook(&mbtac_uplink_host_ipv6_ops);

    if(IWF_LOG_LEVEL_UL_SUCC_CRT_LOG & iwf_log_level)
    {
        if(0 != printk_ratelimit())
        {
            printk(KERN_INFO _IWF_MODULE_NAME": %s Uplink HostTraffic hook for mbtac is registered \n",__func__);
        }
    }
}

#endif
uint32_t set_vtc_nid( sIpSicpApiMsg *msg )
{
#ifdef __IWF_ENABLE_MSG_RTR
    int msgRtrRetval = -1;
    ip_sicp_api_t * cfg       = (ip_sicp_api_t *) getCfgMsgPayload(msg);
    sicp_data_t   * conf_data = (sicp_data_t *)  &cfg->u.teid_sicp_data;
    ftm_nid = conf_data->tx_nid;
    /* for packet give the sic address of eNB */
    msgRtrRetval = MsgRtrARMRegisterSysComID(conf_data->tx_nid, MsgRtrARMCallback);
    if(msgRtrRetval)
    {
        if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
        {
            if(0 != printk_ratelimit())
            {
                printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMRegisterSysComID failed for Single Callback registration: %d\n",__func__, msgRtrRetval);
            }
        }
        return VTC_ERR;
    }
#endif
    return VTC_OK;

}

/* ------------------------------------------------------------------------------ */
uint16_t applyVlanFilter(uint16_t vlan_id, uint16_t port)
{
    uint16_t index;
    if(vlan_id > MAX_VLAN_ID || port == INTF_INVALID)
    {
        return 0;
    }
    if(vlan_id == 0) //For priority tagged packets we need to consider accepted frame types value
    {
        if (vlanCfgInfo[port].AcceptFrameType == ADMIT_ALL)
        {
            return 1;
        }
    }
    for(index = 0; index < MAX_VLAN_FILTERS; index++)
    {
        if(vlan_id <= vlanCfgInfo[port].vidfilter[index].high)
        {
            if(vlan_id >= vlanCfgInfo[port].vidfilter[index].low)
            {
                return 1;
            }
            break;
        }
    }
    return 0;
}

void vfilter_hook_register(void)
{
    nf_register_hooks(br_nf_ops, ARRAY_SIZE(br_nf_ops));
    printk(KERN_ALERT ": %s hook registered \n",__func__);
}

unsigned int vfilter_br_nf_local(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*))
{
    int retVal = NF_ACCEPT;
    if(skb)
    {
        uint16_t vid = INVALID_VLAN_ID;
        uint8_t in_port = INTF_INVALID;
        uint8_t out_port = INTF_INVALID;
        unsigned int ethType;
        struct ethhdr *ethHeader = eth_hdr(skb);

        if(ethHeader && VlanAwareFlag)
        {
            //WBTS_TRS_20787: Exclude VLAN Filtering for
            //L2 Management Protocols e.g. TOP, LOAM, SOAM
            ethType = ntohs(ethHeader->h_proto);
            if(ETH_P_8021Q  == ethType)
            {
                struct vlan_ethhdr* vlan_hdr = vlan_eth_hdr(skb);
                ethType = ntohs(vlan_hdr->h_vlan_encapsulated_proto);
            }

            switch (ethType)
            {
                case ETH_P_SLOW: //LOAM
                case ETH_P_1588:  //TOP
                case ETH_P_SOAM:  //SOAM
                    {
                        return NF_ACCEPT;
                    }
                default:
                    break;
            }

            if(in->name)
            {
                in_port = getPortNumPerInterface(in->name);
            }
            if(out->name)
            {
                out_port = getPortNumPerInterface(out->name);
            }

            //Exempt VLAN filtering on LMP port, for 3 port hardware
            if((NumofPorts == NO_OF_INTF_THREE_PORT_HW)&&(in_port == INTF_ETH1))
            {
                retVal = NF_ACCEPT;
            }
            else
            {
                if(skb_vlan_tag_present(skb))
                {
                    //VLAN Tagged Frames
                    vid = skb_vlan_tag_get_id(skb);

                    if(NF_BR_LOCAL_IN == hooknum)
                    {
                        if(!applyVlanFilter(vid,in_port))
                        {
                            DiscardTaggedFramesCount[in_port]++;
                            retVal = NF_DROP;
                        }
                        else
                        {
                            TaggedFramesCount[in_port]++;
                        }
                    }
                    else if(NF_BR_FORWARD == hooknum)
                    {
                        if(!applyVlanFilter(vid, in_port))
                        {
                            /* Kernel makes copies of any incoming Broadcast/Multicast packets,
                               sends to both FORWARD/INPUT br chain.Forward hook (NF_BR_FORWARD)
                               gets triggered first followed by the Local In hook (NF_BR_LOCAL_IN).
                               hence to avoid duplicate packet count, Packet count is done at
                               Local hook only.*/

                            if(!(is_multicast_ether_addr(ethHeader->h_dest) || is_broadcast_ether_addr(ethHeader->h_dest)))
                            {
                                DiscardTaggedFramesCount[in_port]++;
                            }
                            retVal = NF_DROP;
                        }
                        else
                        {
                            if(!(is_multicast_ether_addr(ethHeader->h_dest) || is_broadcast_ether_addr(ethHeader->h_dest)))
                            {
                                TaggedFramesCount[in_port]++;
                            }
                        }

                        if(retVal != NF_DROP)
                        {
                            if(!applyVlanFilter(vid, out_port))
                            {
                                DiscardTaggedFramesCount[out_port]++;
                                retVal = NF_DROP;
                            }
                            else
                            {
                                TaggedFramesCount[out_port]++;
                            }
                        }
                    }
                    else if(NF_BR_LOCAL_OUT == hooknum)
                    {
                        if(!applyVlanFilter(vid,out_port))
                        {
                            DiscardTaggedFramesCount[out_port]++;
                            retVal = NF_DROP;
                        }
                        else
                        {
                            TaggedFramesCount[out_port]++;
                        }
                    }
                }
                else
                {
                    if((vlanCfgInfo[in_port].AcceptFrameType == ADMIT_TAGGED) && (NF_BR_LOCAL_IN == hooknum))
                    {
                        DiscardUnTaggedFramesCount[in_port]++;
                        retVal = NF_DROP;
                    }
                    else if ((NF_BR_FORWARD == hooknum))
                    {
                        if(vlanCfgInfo[in_port].AcceptFrameType == ADMIT_TAGGED)
                        {
                            /* Kernel makes copies of any incoming Broadcast/Multicast packets,
                               sends to both FORWARD/INPUT br chain.Forward hook (NF_BR_FORWARD)
                               gets triggered first followed by the Local In hook (NF_BR_LOCAL_IN).
                               hence to avoid duplicate packet count, Packet count is done at
                               Local hook only.*/

                            if(!(is_multicast_ether_addr(ethHeader->h_dest) || is_broadcast_ether_addr(ethHeader->h_dest)))
                            {
                                DiscardUnTaggedFramesCount[in_port]++;
                            }
                            retVal = NF_DROP;
                        }
                        else
                        {
                            if(vlanCfgInfo[out_port].AcceptFrameType == ADMIT_TAGGED)
                            {
                                DiscardUnTaggedFramesCount[out_port]++;
                                retVal = NF_DROP;
                            }
                        }
                    }
                }
            }
        }
    }
    else
    {
        if(0 != printk_ratelimit())
        {
            printk(KERN_INFO _IWF_MODULE_NAME": ERR!! %s  called with NULL data \n",__func__);
        }
    }
    return retVal;
}

unsigned int vfilter_br_nf_local_out(unsigned int hooknum,
        struct sk_buff *skb,
        const struct net_device *in,
        const struct net_device *out,
        int (*okfn)(struct sk_buff*))
{
    return NF_ACCEPT;
}

/* ------------------------------------------------------------------------------ */

int __init iwf_deamon_init(void)
{
#ifdef __IWF_ENABLE_MSG_RTR
    int msgRtrRetval = -1;
#endif

    gUplaneIpUpdated = false;

    seCfgDrv_init();

    /* initialize u-plane ip mutex */
    mutex_init(&gUPlaneIpMutexLock);

    /* initialize IWF tables */
    init_lte_iwf_table_dl_ul ();

    /* Initialize VLan ipcounters array */
    memset(ipCounters,0,sizeof(ipCounters));

    initVlanConfig();

#ifdef __USE_NET_FILTER__
    netfilter_downlink_ops.hook                   =       iwfD_DownLinkHook;
    netfilter_downlink_ops.pf                     =       PF_INET;
    netfilter_downlink_ops.hooknum                =       NF_INET_LOCAL_IN;
    netfilter_downlink_ops.priority               =       NF_IP_PRI_FIRST + 1;
    nf_register_hook(&netfilter_downlink_ops);

// k2l: disable pre-route hook
#if 0
    netfilter_downlink_pre_ops.hook                   =       iwfD_DownLinkHook;
    netfilter_downlink_pre_ops.pf                     =       PF_INET;
    netfilter_downlink_pre_ops.hooknum                =       NF_INET_PRE_ROUTING;;
    netfilter_downlink_pre_ops.priority               =       NF_IP_PRI_FIRST + 1;
    nf_register_hook(&netfilter_downlink_pre_ops);
#endif

    netfilter_downlink_ipv6_ops.hook              =       iwfD_DownLinkHook;
    netfilter_downlink_ipv6_ops.pf                =       NFPROTO_IPV6;
    netfilter_downlink_ipv6_ops.hooknum           =       NF_INET_LOCAL_IN;
    netfilter_downlink_ipv6_ops.priority          =       NF_IP6_PRI_FIRST + 1;
    nf_register_hook(&netfilter_downlink_ipv6_ops);


// k2l: disable pre-route hook
#if 0
    netfilter_downlink_ipv6_pre_ops.hook                   =       iwfD_DownLinkHook;
    netfilter_downlink_ipv6_pre_ops.pf                     =       NFPROTO_IPV6;
    netfilter_downlink_ipv6_pre_ops.hooknum                =       NF_INET_PRE_ROUTING;;
    netfilter_downlink_ipv6_pre_ops.priority               =       NF_IP6_PRI_FIRST + 1;
    nf_register_hook(&netfilter_downlink_ipv6_pre_ops);
#endif


    mbtac_uplink_hook_register();
    vfilter_hook_register();
#endif

    // k2l work-around
    dev_eth0 = dev_get_by_name(&init_net, "eth0");

#ifdef __IWF_ENABLE_MSG_RTR
#ifndef __IWF_ONLY_ONE_CALL_BK

    /* for DL packet give the sic address of Tupu */
    msgRtrRetval =  MsgRtrARMRegisterSysComID(__IWF_TUPU_SICAD, sendPktToLMP);
    if(msgRtrRetval)
    {
        printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMRegisterSysComID for DL failed : %d\n",__func__, msgRtrRetval);
    }
#endif
    msgRtrRetval =  MsgRtrARMRegisterSysComID(__IWF_REGISTRD_SICAD, MsgRtrARMCallback);
    if(msgRtrRetval)
    {
        printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMRegisterSysComID for 0x123E failed : %d\n",__func__, msgRtrRetval);
    }
    DBG_PRT_ALL("MsgRtrARMRegisterSysComID: rc %d\n", msgRtrRetval);
#endif
    fzm_iwf_d = proc_mkdir("fzm_iwf", NULL);
    if(NULL == fzm_iwf_d)
    {
        return -ENOMEM;
    }

    fzm_iwf_f = proc_create("iwf_ko", 0644, fzm_iwf_d, &iwf_proc_counters_fops);
    if(NULL == fzm_iwf_f)
    {
        remove_proc_entry("fzm_iwf", NULL);
        return -ENOMEM;
    }

    fzm_iwf_tbl_dump_f =  proc_create("iwf_ko_tbl_dump", 0644, fzm_iwf_d,&iwf_proc_table_dump_fops);
    if(NULL == fzm_iwf_tbl_dump_f)
    {
        remove_proc_entry("fzm_iwf", NULL);
        return -ENOMEM;
    }

    fzm_iwf_loglevel_help_f = proc_create("iwf_ko_loglevel_help", 0644 , fzm_iwf_d , &iwf_proc_loglevel_help_fops);
    if(NULL == fzm_iwf_loglevel_help_f)
    {
        remove_proc_entry("iwf_ko_loglevel_help", NULL);
        return -ENOMEM;
    }

    fzm_iwf_loglevel_add_f = proc_create("iwf_ko_loglevel_add", 0644 , fzm_iwf_d, &iwf_proc_loglevel_add_fops);
    if(NULL == fzm_iwf_loglevel_add_f)
    {
        remove_proc_entry("iwf_ko_loglevel_add", NULL);
        return -ENOMEM;
    }

    fzm_iwf_loglevel_delete_f = proc_create("iwf_ko_loglevel_delete", 0644 , fzm_iwf_d, &iwf_proc_loglevel_delete_fops);

    if(NULL == fzm_iwf_loglevel_delete_f)
    {
        remove_proc_entry("iwf_ko_loglevel_delete", NULL);
        return -ENOMEM;
    }

    fzm_iwf_loglevel_f = proc_create("iwf_ko_loglevel", 0644 , fzm_iwf_d , &iwf_proc_loglevel_fops);
    if(NULL == fzm_iwf_loglevel_f)
    {
        remove_proc_entry("iwf_ko_loglevel", NULL);
        return -ENOMEM;
    }

    fzm_iwf_ip_cntrs_f = proc_create("iwf_ko_ip_cntrs", 0666 , fzm_iwf_d, &iwf_proc_ip_cntrs_fops);
    if(NULL == fzm_iwf_ip_cntrs_f)
    {
        remove_proc_entry("iwf_ko_ip_cntrs", NULL);
        return -ENOMEM;
    }


    printk(KERN_INFO _IWF_MODULE_NAME"Version : %d.%d ... Loaded...\n\n",__IWF_VERSION_MAJOR__,__IWF_VERSION_MINOR__);

    return 0;
}


/* ------------------------------------------------------------------------------ */
void __exit iwf_deamon_exit(void)
{
#ifdef __IWF_ENABLE_MSG_RTR
    int msgRtrRetval = -1;
#endif

    // k2l
    dev_put(dev_eth0);

#ifdef CRASH_HNDL

    /* the up-link hook will be registered only in case of simulation mode, so check its registration */
    if(true == bIsULHookRegistered )
    {
        nf_unregister_hook(&netfilter_uplink_ops);
        bIsULHookRegistered = false;
    }
    nf_unregister_hook(&netfilter_downlink_ops);
    nf_unregister_hook(&netfilter_downlink_pre_ops);
    nf_unregister_hook(&mbtac_uplink_ops);
    nf_unregister_hook(&mbtac_uplink_ipv6_ops);
    nf_unregister_hook(&mbtac_uplink_host_ops);
    nf_unregister_hook(&mbtac_uplink_host_ipv6_ops);

    nf_unregister_hook(&netfilter_downlink_ipv6_ops);
    nf_unregister_hook(&netfilter_downlink_ipv6_pre_ops);
    nf_unregister_hooks(br_nf_ops, ARRAY_SIZE(br_nf_ops));
#ifdef __IWF_ENABLE_MSG_RTR
    MsgRtrARMUnregister(__IWF_REGISTRD_SICAD);
#ifndef __IWF_ONLY_ONE_CALL_BK
    /* unregister callbacks */
    if (ftm_nid)
    {
        /* unregister callbacks */
        msgRtrRetval = MsgRtrARMUnregister (ftm_nid);
        if(msgRtrRetval)
        {
            if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMUnregister for FTM failed : %d\n",__func__, msgRtrRetval);
                }
            }
        }
    }

    msgRtrRetval = MsgRtrARMUnregister (__IWF_TUPU_SICAD);
    if(msgRtrRetval)
    {
        printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMUnregister for TUPU failed : %d\n",__func__, msgRtrRetval);
    }
#else
    /* unregister callbacks */
    if (ftm_nid)
    {
        msgRtrRetval = MsgRtrARMUnregister (ftm_nid);
        if(msgRtrRetval)
        {
            if(IWF_LOG_LEVEL_UL_ERR_CRT_LOG & iwf_log_level)
            {
                if(0 != printk_ratelimit())
                {
                    printk(KERN_INFO _IWF_MODULE_NAME"%s: ERR!! MsgRtrARMUnregister for FTM failed : %d\n",__func__, msgRtrRetval);
                }
            }
        }
    }
#endif
#endif

    seCfgDrv_exit();

#endif

    if(NULL != fzm_iwf_d)
    {
        remove_proc_entry("iwf_ko", fzm_iwf_d);
        remove_proc_entry("iwf_ko_tbl_dump", fzm_iwf_d);
        remove_proc_entry("iwf_ko_loglevel_help", fzm_iwf_d);
        remove_proc_entry("iwf_ko_loglevel_add", fzm_iwf_d);
        remove_proc_entry("iwf_ko_loglevel_delete", fzm_iwf_d);
        remove_proc_entry("iwf_ko_loglevel", fzm_iwf_d);
        remove_proc_entry("iwf_ko_ip_cntrs", fzm_iwf_d);

        remove_proc_entry("fzm_iwf", NULL);

    }

    printk(KERN_INFO _IWF_MODULE_NAME"%s: UNLOADED ..... \n",__func__);
}
/* ------------------------------------------------------------------------------ */
module_init(iwf_deamon_init);
module_exit(iwf_deamon_exit);

/* ----------------------------- EOF -------------------------------------------- */
